
static private inherit "_Charset";

object decoder(string name)
{
  if(!name || (<
    "iso_8859-1:1987", "iso-8859-1", "iso_8859-1", "iso-ir-100", "latin1",
    "l1", "ansi_x3.4-1968", "iso_646.irv:1991", "iso646-us", "iso-ir-6",
    "us", "us-ascii", "ascii", "cp367", "ibm367", "cp819", "ibm819"
  >)[name=lower_case(name)])
    return class {
      static private string s = "";
      object feed(string ss)
      {
	s += ss;
	return this_object();
      }
      string drain()
      {
	string ss = s;
	s = "";
	return ss;
      }
      object clear()
      {
	s = "";
	return this_object();
      }
    }();

  if(name[..7]=="iso-2022")
    return ISO2022();

  program p = ([
    "utf-7": UTF7dec,
    "utf-8": UTF8dec
  ])[name];

  if(p)
    return p();

  object o = rfc1345(name);

  if(o)
    return o;

  throw(({"Unknown character encoding "+name+"\n", backtrace()}));
}

object encoder(string name, string|void replacement)
{
  if(!name || (<
    "iso_8859-1:1987", "iso-8859-1", "iso_8859-1", "iso-ir-100", "latin1",
    "l1", "ansi_x3.4-1968", "iso_646.irv:1991", "iso646-us", "iso-ir-6",
    "us", "us-ascii", "ascii", "cp367", "ibm367", "cp819", "ibm819"
  >)[name=lower_case(name)])
    return class {
      static private string s = "";
      static private string|void replacement;
      static private string low_convert(string s, string|void r)
      {
	int i = strlen(s);
	while(--i>=0)
	  if(s[i]>255)
	    if(r)
	      s=s[..i-1]+low_convert(r)+s[i+1..];
	    else
	      throw(({"Character unsupported by encoding.\n", backtrace()}));
	return s;
      }
      object feed(string ss)
      {
	s += low_convert(ss, replacement);
	return this_object();
      }
      string drain()
      {
	string ss = s;
	s = "";
	return ss;
      }
      object clear()
      {
	s = "";
	return this_object();
      }
      void create(string|void r)
      {
	replacement = r;
      }
    }(replacement);

  program p = ([
    "utf-7": UTF7enc,
    "utf-8": UTF8enc
  ])[name];

  if(p)
    return p(replacement);

  object o = rfc1345(name, 1, replacement);

  if(o)
    return o;

  throw(({"Unknown character encoding "+name+"\n", backtrace()}));
}
