#!NOMODULE
//
// This source-code is a part of the Roxen IntraSeek search engine package.
//  1996-1998 Idonex AB
// Email: info@idonex.se 
//

#if constant(Process.create_process)
import ".";
#endif

#include "crawler.h"

#define FINISHED 1
#define FREEZING 5
#define FROZEN 6
#define ERROR 7
#define DUMPING 8

#define XDUMP 1
#define REST  2
#define COLL  0
#define FINI  4
#define ERRR 5
#define XREFE 9

import Simulate;
import Stdio;

object scheduler, indexer, global_profile_handler, global_fetcher;
int safety, status_info, status_info2, gensym, j, i, start_time;
program fetcher_program, scheduler_program, indexer_program;
string neuurl, global_profile;

string robot_home;

void crawlerlog(string txt)
{
  write_file(robot_home+"logs/crawler.log", txt+"\n");
}

void plog(string data)
{
  if(data && global_profile_handler)
    global_profile_handler->log(global_profile, "Control: "+data);
}

int q_safety() {   return safety;  }
int q_safety2() { return ((global_profile_handler->safety_save(global_profile))); }
int q_status_info2() {  return status_info2; }
int q_status_info() { return status_info; }

void debugdump(string dumptext)
{
  crawlerlog("Control: "+dumptext+"\n");
}

void touch_file(string file_name)
{
  object filen;
  filen=Stdio.File();
  filen->open(file_name, "wtc");
  filen->write("flag");
  filen->close();
}


void create_status_file(string profil, int extrainfo, int|string message)
{
  object filen;
  mixed *status;
  status=scheduler->query_status();

  if (message)
  {
    filen=Stdio.File();
    filen->open(global_profile_handler->working_at(profil)+profil+".error.html", "wtc");
    filen->write(message);
    filen->close();
  }
  
  filen=Stdio.File();
  filen->open(global_profile_handler->working_at(profil)+profil+".status.html", "wtc");
  switch(extrainfo)
  {
   case FINI :  filen->write("FINISHED!\n");
                 break;
   case XREFE :  filen->write("Currently writing the site structure log!\n");
                 break;
   case REST :  filen->write("Currently restoring frozen info from last freeze-point. \n");
                 break;
   case DUMPING :  filen->write("Currently saving databases to disc. (This can take a while!)\n");
                 break;
   case COLL :  filen->write("Currently running the crawler to fetch pages. ("+
			     ((global_profile_handler->safety_save(global_profile))-safety)+
			     " pages left to next safety save.)");
                break;
   case ERRR :  filen->write("<font color=red> Error! <br>"+
			     (message?message:"View log to see reason!")+"</font>");
                break;
  }
  filen->write("<p><ul>\n<li>I have visited "+status[0]+" web pages.</li>\n"
	       "<li>Right now, I know of "+status[1]+" more pages to visit.</li>\n");

  if (extrainfo!=FINI)
    filen->write("<li>I've been running for "+global_profile_handler->describe_time(time()-start_time)+".</li>\n");
  filen->write("</ul><p>This information was last updated on "+ctime(time())+"." );
  filen->close();
}

void shutdown()
{
  string from;
  string to;

  global_profile_handler->modify_robot_id(global_profile, time(), XREFE);
  create_status_file(global_profile, XREFE, 0);
  
  scheduler->write_references_log(global_profile);

  plog("Stop. Shutting down business! Totally downloaded "+status_info+" bytes!");
  plog("All fetchers shut down! Saving Database!");
  
  create_status_file(global_profile, DUMPING, 0);
  global_profile_handler->modify_robot_id(global_profile, time(), XDUMP);
  // 1=new database is beeing created.
  indexer->dump_to_disc(global_profile, global_profile_handler);  
  create_status_file(global_profile, FINI, 0);
  crawlerlog("Finished "+global_profile+" crawler at "+ctime(time()));

  plog("Finished!\n");
  plog("Creating flag "+(global_profile_handler->store_at(global_profile))
       +global_profile+".new.flag, to tell engine to swap to the new database!");

  plog("->"+global_profile+".new.flag");
  plog("=>"+(global_profile_handler->store_at(global_profile))+global_profile+".new.flag");
  touch_file((global_profile_handler->store_at(global_profile))+global_profile+".new.flag");

  global_profile_handler->modify_robot_id(global_profile, time(), FINISHED);

  _exit(1);
}

void safe_freeze()
{
  plog("Safe-Freeze! Calling indexer and scheduler!");
  create_status_file(global_profile, DUMPING,0);
  indexer->dump_to_disc(global_profile, global_profile_handler);   // 1=new database is beeing created.
  scheduler->freeze();
}


void real_alarmet(int thefetcherid)
{
  plog("Ok, Freezing everything as it is right now.");
  create_status_file(global_profile, FREEZING, 0);
  this_object()->safe_freeze();
  global_profile_handler->modify_robot_id(global_profile, time(), FROZEN);
  create_status_file(global_profile, FINI, 0);
  _exit(1);
}

void crawl_around(object rpc)
{
  string neuurl;
  int noshutdownflag=1;
  int xpause;
  mixed fetcher_result=({});
  
  // The mixed fetcher result consist of: int fetcher_id, string textbody,
  // string texttitle, string *urls_visited, string *new_possible_urls, string metainfo,
  // int sizeofdoc, string summary, string charset)
  
  create_status_file(global_profile, COLL, 0);

  object filters=Crawler.filters();

  do
  {
    neuurl=scheduler->give_me_url();
    
    if(neuurl!="")
    {
      gensym++;
      
//      werror("Fetching doc #"+gensym+" = "+neuurl+".\n");
      
      create_status_file(global_profile, COLL, 0);
      fetcher_result=global_fetcher->fetch(neuurl, gensym, global_profile,
					   global_profile_handler,rpc,filters);

      if (xpause=(global_profile_handler->walkpause(global_profile)))
	sleep(xpause); // Walk delay!
      
      if (fetcher_result[_SIZEOFDOC]>0)
	status_info+=fetcher_result[_SIZEOFDOC];
  
      scheduler->ive_been_at(fetcher_result[_URLS_VISI]);
      scheduler->new_possible_urls(fetcher_result[_NEW_URLS_], fetcher_result[_URLS_VISI]);

      // Not indexing documents with error codes in size!
      if((fetcher_result[_TEXT_BODY]!="") && (fetcher_result[_SIZEOFDOC]>0))  
	indexer->process_text(global_profile, global_profile_handler,
			      fetcher_result[_TEXT_BODY],
			      fetcher_result[_TEXT_TITL],
			      fetcher_result[_METAINFO_],
			      fetcher_result[_URLS_VISI][0],
			      fetcher_result[_SIZEOFDOC],
			      fetcher_result[_SUMMARY__],
			      fetcher_result[_CHARSET__],
			      fetcher_result[_DATE_____],
			      );

      if (fetcher_result[_SIZEOFDOC]<-1) // If less than -1, it is an ERROR code! (-2..-7)
	if (global_profile_handler->struclogactive(global_profile)!="2")
	  scheduler->handle_error_code(fetcher_result[_URLS_VISI], fetcher_result[_SIZEOFDOC]);
      
      if ((safety++)==((global_profile_handler->safety_save(global_profile))))
      {
	safety=0;
	safe_freeze();
      }

      if(gensym==(global_profile_handler->max_doc_dl(global_profile)))
      {
	shutdown();
	noshutdownflag=1;
      }
    }
    else {
      plog("Finished - no more URLs to visit.");
      shutdown();
    }
  }
  while (neuurl!="" || noshutdownflag);
}


int start_control(string restore, string profil, object profile_pointer, object rpc)
{
  int i,k, status;
  string temporar;

  start_time=time();
  
  global_profile=profil;
  global_profile_handler=profile_pointer;  // Already cloned by crawler.

  
  global_fetcher=Crawler.fetcher();
  global_fetcher->set_robot_home(robot_home);

  indexer=Crawler.indexer();
  indexer->set_robot_home(robot_home);
  
  plog("I'm running profile "+profil);

  if(restore=="restore")
    {
      plog("Instead of creating new database, we continue from freezing point");
      
      scheduler=Crawler.scheduler(profil, global_profile_handler, 1, robot_home);
      create_status_file(profil, REST, 0);
      indexer->wake_up_restore(profil, global_profile_handler);
      status=scheduler->restore_from_freeze();
      if(!status)
      {
	return -1;
      }
    } else
    {
      scheduler=Crawler.scheduler(profil, global_profile_handler, 0, robot_home);
      
      if (temporar=indexer->wake_up_create(profil, global_profile_handler))
      {
	 // If wake_up_create returns anything, it is a error message!
	plog("ERROR: "+temporar);
	create_status_file(profil, ERRR, 0);
	global_profile_handler->modify_robot_id(global_profile, time(), ERROR);
	_exit(1);
      }   
    } 
  status_info=0;
  status_info2=time();
  safety=0;
  plog("I'm up and running at "+ctime(time()));
//   signal(signum("SIGHUP"), real_alarmet); 
  crawl_around(rpc);
  return -1;
}

void set_robot_home(string foo)
{
  robot_home=foo;
}
