//
// This source-code is a part of the Roxen IntraSeek search engine package.
//  1996-1998 Idonex AB
// Email: info@idonex.se 
//

#include <simulate.h>
#include "crawler.h"

inherit "url";

#define QUERY_LOG_DETAIL global_profile_handler->crawler_log_detail
//  Leave untouched

string global_redirect_new_url;
string global_result_string;
function global_call_back_when_finished;
string *global_visited_urls;
string global_profil;
object global_profile_handler;
int global_max_download_bytes;
int fetcherid;
string uri;
string protocol;
string site;
int    port;
//int summary_text_length;

object global_pointer_to_file_object;
string global_remember_url;
string robot_home;


void crawlerlog(string txt)
{
  Stdio.write_file(robot_home+"logs/crawler.log", txt+"\n");
}

void plog(string data)
{
  if(data && global_profile_handler)
    global_profile_handler->log(global_profil, "Fetcher "+fetcherid+": "+data);
}


void debugdump(string dumptext)
{
  crawlerlog("Fetcher: #"+fetcherid+": "+dumptext+"\n");
}

string construct_url()
{
  return ""+protocol+"://"+site+":"+port+"/"+uri;
}

string pluck_out_response_message(string haystack)
{
  return haystack[0..(search(haystack, CRLF+CRLF)+2)];
}

string pluck_out_text_body(string haystack)
{
  return haystack[(search(haystack, CRLF+CRLF))..];
}

int my_smallest(int a, int b)
{
  // A Very special-cased smallest interger checker....
  if(a==-1) a=0;
  if(b==-1) b=0;
  if ((a==0) && (b==0)) return 0;
  if (a==0) return b;
  if (b==0) return a;
  if (a<b) return a;
  return b;
}

string check_charset_in_response_header(string content)
{
  //	    <META HTTP-EQUIV="Content-Type"  CONTENT="text/html; charset=ISO-8859-2">
  int start, end;
  string typen;
  start=search((content=lower_case(content)), lower_case(PREFIX_CONTENT_TYPE));
  if (((start=search(content,"charset="))!=-1)
      && (end=my_smallest(search(content, CR, (start+8)),
			  (my_smallest(search(content, LF, (start+8)),
				       search(content, ";", (start+8)))))))
    return lower_case(content[(start+8)..(end-2)]);
  return DEFAULT_CHARSET; // Not specified, assuming latin 1....
}

int check_size_in_response_header(string html_header)
{
  int start, end, sizen;
  html_header=lower_case(html_header);
  start=search(html_header, lower_case(PREFIX_CONTENT_LENGTH));
  if (start==-1)
    return -1;
  else
  {
    start=start+sizeof(PREFIX_CONTENT_LENGTH);
    end=search(html_header,CR, start);
    if (end==-1)
      return -1;
    if (sscanf(html_header[start..(end-2)], "%d", sizen))
      return sizen;
    return -1;
  }
}

		   
int handle_error_codes(string response)
{
  string statusline;
  string *exploded;
  string *exploded2;
  string httpversion;
  int    statuscode, i;
  string reasonphrase;
  
  exploded=explode(response, CRLF);
  statusline=exploded[0];
  exploded2=explode(statusline, " ");
  httpversion=exploded2[0];
  
  if(sizeof(exploded2)<2) return REQ_FORBIDDEN;
  
  i=sscanf(exploded2[1], "%d", statuscode);
  
  if(i==-1)
    if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_SHORT)
      plog("Warning: Cannot parse response data. Strange statuscode!");
  
  reasonphrase=implode(exploded2[2..sizeof(exploded2)], " ");
  
  if(statuscode==REQ_OK) return REQ_OK;   // OK
  
  if(statuscode==REQ_FORBIDDEN)
  {
    if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_SHORT)
      plog("Warning: I was denied entry. URL Forbidden by rule.");
    return REQ_FORBIDDEN;
  }
  
  if(statuscode==REQ_NOTFOUND)
  {
    if(fetcherid==0)
    {
      if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_SHORT)
	plog("Warning: This site didn't provide any /robots.txt");
    }
    else 
    {
      if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_NONE)
	plog("<font color=red>Error: Resource not found. Outdated?</font>");
    }
    return REQ_NOTFOUND;
  }
  
  if(statuscode==REQ_REDIRECT)            // Resource moved temporarly! Redirected to new URL!
  { 
    for(i==1; i<sizeof(exploded); i++)
    {
      if (strstr(exploded[i], PREFIX_LOCATION) != -1 )
      {
	// Found the new location!!!
	global_redirect_new_url=(exploded[i])[(sizeof(PREFIX_LOCATION))..(sizeof(exploded[i]))];
      }
    }  // possible loop optimization here later on!
    return REQ_REDIRECT;
  }
  if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_SHORT)
    plog("Warning: Not yet implemented handler for error code :"+statuscode+" ("+reasonphrase+")");
  return statuscode;
}


void fetcher_timed_out()
{
  if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_NONE)
  {
    plog("<font color=red>"
	 "Error: Fetch abort! Timed out after "+
	 (global_profile_handler->timeout(global_profil))+" seconds when "
	 "connecting to "+global_remember_url+"<br></font>\n");
    if (global_result_string!="")
      plog("<font color=red>"
	   "Error: However, I was able to download parts of the document, "
	   "in total "+ sizeof(global_result_string) +" bytes, which will "
	   "be thrown away and not indexed. IF you have a slow connection, "
	   "you might want do increase the time for timeouts to give "
	   "Intraaseek fetcher more time to download the files?</font>");
  }
  
  destruct(global_pointer_to_file_object);
  global_result_string="";
}


constant MONTHS=(["Jan":0, "Feb":1, "Mar":2, "Apr":3, "May":4, "Jun":5,
	         "Jul":6, "Aug":7, "Sep":8, "Oct":9, "Nov":10, "Dec":11,
		 "jan":0, "feb":1, "mar":2, "apr":3, "may":4, "jun":5,
		  "jul":6, "aug":7, "sep":8, "oct":9, "nov":10, "dec":11,]);

int httpdate_to_time(string date)
{     
   if (intp(date)) return -1;
   // Tue, 28 Apr 1998 13:31:29 GMT
   // 0    1  2    3    4  5  6
   int mday,hour,min,sec,year;
   string month;
   if(sscanf(date,"%*s, %d %s %d %d:%d:%d GMT",mday,month,year,hour,min,sec)==6)
     return mktime((["year":year-1900,
		     "mon":MONTHS[lower_case(month)],
		     "mday":mday,
		     "hour":hour,
		     "min":min,
		     "sec":sec,
		     "timezone":0]));
   
   
   return -1; 
}


array fetch(string url, int id, string profil,
	    object profile_handler, object rpc, object filters)
{
  // The fetcher should return a mixed array of result consist of:
  // int fetcher_id, string textbody, string texttitle, string *urls_visited,
  // string *new_possible_urls, string metainfo, int sizeofdoc, string summary, string charset)
  // There are defines for these positions in crawler.h

  int return_code, ok_conn, h_error_code, queryerr;
  mixed *expand_result;
  int sizen;
  string host, temp_header_strr;
  mixed *tempo;
  string fetcher_bodytext, fetcher_titletext, fetcher_metainfo, fetcher_summary, fetcher_charset, charset;
  string *fetcher_new_possible_urls;
  int lastmodified;
  string cookie_bunch;
  string cookie_req="";
  string *all_cookies=(profile_handler->cookie_bunch(profil))/",";
  foreach(all_cookies, string pair)
    cookie_req+="Cookie: "+pair+CRLF;
  if(cookie_req=="") cookie_req=CRLF;

  
  fetcherid = id;
  
  global_redirect_new_url="";
  global_result_string="";
  global_visited_urls=({});
  global_profil=profil; 
  global_profile_handler=profile_handler;
  global_max_download_bytes=profile_handler->max_download_bytes(global_profil);
  global_visited_urls+=({url});
  global_result_string="";
  
  expand_result = universal_expand_url(url);
  
  return_code =  expand_result[0];
  uri =  expand_result[1];
  site =  expand_result[2];
  port =  expand_result[3];
  protocol =  expand_result[4];

    
  object fi=Stdio.File();
  
  if (return_code)
    {
      if (fetcherid==0)
	uri="/"+uri;  // Fetcher zero is always called from scheduler esp. to check /robots.txt
      
      if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_NONE)
	plog("Report: Connecting to "+uri+" from "+site+":"+port);
      
      if (fetcherid==0)  // Is this a robot.txt fetching fetcher?!
	if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_MEDIUM)
	  plog("Notice: a '/robots.txt'-fetcher started!");
      
      catch(ok_conn=(fi->connect(site, port)));  // To catch or not to catch. Thats the question :-)
      
      if(!ok_conn)
      {
	fi->close();
	if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_NONE)
	  plog("<font color=red>Error: Couldn't connect to get "+uri+" from "+site+" on port "+port+"!</font><br>");
	return ({ id, "","", global_visited_urls, ({}), "", ERR_CON_FAIL, "", "",0 });
      } else
      {
	// OK, now we have a connection to the right port, ask for the document.
	// Also, start a Call-Out for Timeouts If we don't get anything.
	// To the timeout we send the file object as argument so
	// it can destruct it!
	
	global_result_string="";
	global_remember_url=uri;
	global_pointer_to_file_object=fi;
	host = site+":"+port;
		
#ifdef __NT__
	call_out(fetcher_timed_out,global_profile_handler->timeout(global_profil));
#else
  	signal(signum("SIGALRM"), fetcher_timed_out);
  	alarm((global_profile_handler->timeout(global_profil)));
#endif	
		
	fi->write("GET /"+uri+" "+"HTTP/1.0"+CRLF);

	// open sitebuilder backdoor
	string backdoor_userpass;
  	if(rpc)
	{
	  backdoor_userpass=rpc->get_backdoor_userpass();
	  string auth=
	    "Authorization: Basic "+
	    MIME.encode(backdoor_userpass,"base64")
	    +CRLF;
   	  fi->write(auth);
	}
	fi->write("pragma: no-cache"+CRLF);
	//	werror("URL: "+url+"\n");
	
	fi->write(AGENT_IDENTIFICATION+CRLF+ACCEPT_REQ_FIELD+CRLF+
		  AGENT_FROM+(profile_handler->all_robots_from())+
		  CRLF+AGENT_HOST+host+CRLF);
	fi->write(cookie_req+CRLF);  // Cookiereq always ends with an crlf.
	fi->write(CRLF);
	
	
	global_result_string=fi->read(); //MAX_DATA_READ);
	
	if (fi)
	  fi->close();  // If it wasn't destructed by Timeout, close it...
	else
	  global_result_string="";

	if(!global_result_string)
	  global_result_string="";
	//	werror("result: %O\n",global_result_string);

	// close sitebuilder backdoor
	if(rpc)
	  rpc->remove_backdoor_userpass(backdoor_userpass);
	
	//  	alarm(99999999);   // Send the alarm away!
		
	if((id!=0) && (global_result_string==""))
	{
	  if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_NONE)
	    plog("<font color=red>Error: Couldn't connect to "
		 +site+" on port "+port+", connection failed!</font><br>");
	  return ({ id, "","", global_visited_urls, ({}), "", ERR_CON_FAIL, "", "",0 });
	}
	
	h_error_code=handle_error_codes(temp_header_strr=
					pluck_out_response_message(global_result_string));

//	werror(global_result_string);
	switch(h_error_code)
	{
	 case REQ_REDIRECT:
	  // Looping, redirected to: global_redirect_new_url
          // I removed the logging of redirects! (ERR_REDIRECT). Not interesting.
	  return ({ id, "","", global_visited_urls, ({ global_redirect_new_url }),
		    "", -1, "", "", 0 });
	  break;

	 case REQ_OK:
	  // If id is zero this is a special fetcher destinated to fetch robots txt only. 
	  if (id==0) 
	    return ({ 0, global_result_string, "", global_visited_urls,
		      ({}), "", sizeof(global_result_string), "", "", 0 });
	  else
	  {
	    array req=MIME.parse_headers(global_result_string);
//	    werror(sprintf("%O",req));
	    string content_type=req[0]["content-type"];
	    if(search(content_type,";")+1)
	      content_type=(content_type/";")[0];
	    mixed lastmod=req[0]["last-modified"];
	    if(lastmod&&(int)lastmod)
	      lastmodified=(int)lastmod;
	    else if(lastmod)
	      lastmodified=httpdate_to_time(lastmod);
	    else
	      lastmodified=0;
//	    werror(sprintf("Headers: %O\nLastmodified: %d\n",req[0],(int)lastmodified));
	    if(filters->content_types[content_type]) 
	    {
	      sizen=check_size_in_response_header(temp_header_strr);
	      charset=check_charset_in_response_header(temp_header_strr);
	      tempo=filters->content_types[content_type]->
		filter(req[1],site, uri, port, protocol,
		       global_profile_handler->summary_text_length(global_profil));
//	      werror(sprintf("FOO: %O",tempo));
              //     ({ array links, string body, string title, string metadata,
	      //        string summary,string charset, int index, int follow });

//	      werror(sprintf("%O",tempo));
	      
	      fetcher_new_possible_urls=Array.map(tempo[0],lambda(string url)
							   {
//							     sscanf(url,"%s?",url);
							     return url;
							   });
	      fetcher_bodytext=tempo[1];
	      fetcher_titletext=req[0]["title"]||tempo[2];
	      fetcher_metainfo=tempo[3];
	      fetcher_summary=tempo[4];

	      // Note; If user has set charset with META tag, it will override the
              // charset that the server told about!
	      if (tempo[5]!="")     
		fetcher_charset=tempo[5];
	      else 
		fetcher_charset=charset;
		
	      if ((!tempo[6]) && (!tempo[7]))   // no index no follow!
	      {
		if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_MEDIUM)
		  plog("Notice: META robot tag for noindex and nofollow found. "
		       "Not indexing, and not following links!");
		return ({ id, "","", global_visited_urls, ({}), "", -1, "", "",lastmodified });
	      }
	      else
		if (!tempo[6]) 
		{
		  if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_MEDIUM)
		    plog("Notice: META robot tag for noindex found. Not indexing this information.");
		  return ({ id, "","", global_visited_urls, fetcher_new_possible_urls, "", -1, "", "",lastmodified });
		}
		else
		  if (!tempo[7])   // Tempo 4 is index. So if not INDEX....
		  {
		    if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_MEDIUM)
		      plog("Notice: META robot tag for nofollow found. Not following links found.");
		    return ({ id, fetcher_bodytext, fetcher_titletext, global_visited_urls,
			      ({}), fetcher_metainfo, sizen, fetcher_summary, fetcher_charset,
			      lastmodified }); 
		  }
		  else
		    return ({ id, fetcher_bodytext, fetcher_titletext, global_visited_urls,
			      fetcher_new_possible_urls, fetcher_metainfo, sizen,
			      fetcher_summary, fetcher_charset,
			      lastmodified}); 
	    }
	    else   // Illegal type, not text...
	      return ({ id, "","", global_visited_urls, ({}), "", ERR__ILLEGAL, "", "", 0 });
	  }
	  break;

	 case REQ_FORBIDDEN :
	  return ({ id, "", "", global_visited_urls, ({}), "", ERR_FORBIDDN, "", "", 0 });
	  break;

	 case REQ_NOTFOUND :
	  return ({ id, "", "", global_visited_urls, ({}), "", ERR_NOTFOUND, "", "", 0 });
	  break;

	 default :
	  return ({ id, "", "", global_visited_urls, ({}), "", ERR__UNKNOWN, "", "", 0 });
	  break;
	}
      }
    }
  else
  {
    if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_SHORT)
      plog("Warning: Couldn't expand the incoming URL ("+url+").");
    return ({ id, "", "", global_visited_urls, ({}), "", ERR__ERR_URL, "", "", 0 });
  }
}

void set_robot_home(string foo)
{
  robot_home=foo;
}
