// This source-code is a part of the Roxen IntraSeek search engine package.
//  1996-1998 Idonex AB
// Email: info@idonex.se 
//
// $Id: indexer.pike,v 1.37 1999/06/07 22:51:18 js Exp $


#include <simulate.h>
#include "../modules/languages.h"
#include "crawler.h"

import spider;
import Stdio;
import Simulate;

string robot_home;

// Public methods:
//
// void process_text(string database_id, object profile_handler, string textbody,
//                   string texttitle, string metainfo, string url, int sizeofdoc, string summary,
//                     int|void date)
// void dump_to_disc(string database_id, object profile_handler) //full_database_filename)
//


void crawlerlog(string txt)
{
  write_file(robot_home+"logs/crawler.log", txt);
}


object profile_handler;


int error=0;
int global_cnt_bytes;

mapping cut_huge_index=([]);
mapping huge_index=([]);
mapping pages=([]);
mapping full_text_storage=([]);

mapping symb=([]);

array earray;

void remove_file(string from)
{
  rm(from);
}

void remove_dir(string from)
{
  catch {
    array entries=get_dir(from)-({"..","."});
    foreach(entries, string fn)
      if(file_stat(from+"/"+fn)[1]==-2)
	remove_dir(from+"/"+fn);
      else
	rm(from+"/"+fn);
    rm(from);
  };
}

array empty(array from)
{
  if(earray) return earray;
  return earray=Array.map(allocate(sizeof(from)), lambda(){return"";});
}

void debugdump(string dumptext)
{
  werror("Indexer: "+dumptext+"\n");
}

int gensym(string database_id)
{
  return (symb[database_id]++);     // could be optimized :-)
}

      

string normalize_title(string texttitle)
{
  texttitle=texttitle[0..(MAX_TITLE_LENGTH-1)];
  if (texttitle=="" || texttitle==0)
    texttitle=" - MISSING TITLE - ";
  // debugdump("-----> "+texttitle+" <-----");
  return texttitle;
}

array(string) normalize_text(string textbody, object profile_handler,
			     string database_id, string charset)
{
  // Either returns an array of words, or an empty array if something went wrong!!!
  //
  int i;
  int text_ok=1;
  string temp;
  array(string) text;
  
  // First replace all entities 
  textbody=replace(textbody, entity_from, both_goes_to);
  textbody=replace(textbody, decimal_from, both_goes_to);

  // ... then replace all the special characters
  textbody=replace(textbody, div_from, div___to);

  // Check all? Maybe 50% is sufficent? Maybe less...
  for (i=0; (text_ok && (i<sizeof(textbody)) && (i<1024)); i++) 
    if(textbody[i]<32)  // Problem: Found ASCII value "+textbody[i]+" in textbody.
      text_ok=0;

  if(text_ok)
  {
    // First, we want to get all standard characters correctly lowercased.
    textbody=lower_case(textbody);
    // Make everything LOWERCASED!!! (take care of ISO!)
    if (low[charset])
      textbody=replace(textbody, low[charset][0], low[charset][1]);  

    text=textbody/" ";
    for (i=0; i<sizeof(text); i++)
    {
      temp=text[i];
      // Remove long words
      if(sizeof(temp)>AVOID_WORDS_LONGER_THAN)  
	text[i]="";
      //  Remove bogus word
      if(strlen(replace(temp, FILTER_OUT, empty(FILTER_OUT)))!=strlen(temp))  
	text[i]="";
      // Remove single characters as well!
      if((sizeof(temp)==1) || (int)temp) 
	text[i]="";
    }
    text=text-(profile_handler->stoplist(database_id));  // Remove the stopwords!
    text=text-({""})-({});      // Remove empty leftover junk.
  }
  return (text_ok?text:({}));  // return empty list if something went wrong.
}


array(string) remove_stopwords(array(string) text, object profile_handler, string database_id)
{
  return text-(profile_handler->stoplist(database_id));  // Remove the stopwords!
}


int add_page(string database_id, array(string) text, string title_text,
	     string summary_text, string the_url, int sizeofdoc, int date)
{
  int new;
//  crawlerlog("Adding page "+the_url+" with text "+summary_text+"\n");
  pages[database_id][(new=gensym(database_id))]=
    ""+the_url+DELI+(sizeofdoc/1000.0)+DELI+summary_text+DELI+title_text+DELI+date;
  return new;
}

void add_words_to_index_cut(string database_id, array(string) text,
			    int pagenum, int weight)
{
  string word, cutword;
  array(string) tempo;
  
  foreach(text, word)
    if(sizeof(word)>2)
    {
      cutword=word[0..2];
      if(!(cut_huge_index[database_id][cutword]))
  	cut_huge_index[database_id][cutword]=({word });
      else
	if(-1==member_array(word, cut_huge_index[database_id][cutword]))
	  cut_huge_index[database_id][cutword]+=({ word });
    }
}


void add_words_to_index(string database_id, array(string) text,
			int pagenum, int weight)
{
  string word;
  mapping tempo;
  foreach(text, word)
  {
    if(!(tempo=huge_index[database_id][word]))
      huge_index[database_id][word]=tempo=([]);
    tempo[pagenum]=tempo[pagenum]+weight;
    huge_index[database_id][word]=tempo;  
  }
}


void add_words_to_full_text(string database_id, string textbody, int pagenum)
{
  object gz = Gz;
  if(sizeof(indices(gz)))
    full_text_storage[database_id][pagenum]=gz->deflate()->deflate( textbody );
  else
    full_text_storage[database_id][pagenum]= textbody;
}


int process_text(string database_id, object profile_handler, string textbody,
		 string texttitle, string metainfo, string url, int sizeofdoc,
		 string summary, string charset, int|void date)
{
  //
  // returns pagenumber that has just been added!
  //
  int pagenum;
  array(string) text;
  string title_text;
  
  if (error)
  {
    debugdump("Warning. I haven't been able to initialize correctly. Aborting!");
    return 0;
  }
  
  text=normalize_text(textbody, profile_handler, database_id, charset);
  textbody=text*" "; // Re-use this variable...
  text=remove_stopwords(text,profile_handler, database_id);
  global_cnt_bytes+=sizeof(textbody);


  // Makes a word list, and also check if the text is "OK".
  //    
  if (text!=({})) // if the text indeed was OK, carry on with:
  {
    title_text=normalize_title(texttitle);
    pagenum = add_page(database_id, text, title_text, summary, url, sizeofdoc, date);
    
    if(pagenum)
    {

      if ((profile_handler->quoted_active(database_id))=="1")
	add_words_to_full_text(database_id, textbody, pagenum);
      
      if ((profile_handler->glob_active(database_id))=="1")
	add_words_to_index_cut(database_id, text, pagenum, 1); 
      
      add_words_to_index(database_id, text, pagenum, 1);  // 1=standard weight!
      add_words_to_index(database_id,
			 remove_stopwords(normalize_text(texttitle, profile_handler,
							 database_id, charset),
					  profile_handler, database_id),
			 pagenum, 3); // Extra weight for title stuff.
      add_words_to_index(database_id,
			 remove_stopwords(normalize_text(metainfo, profile_handler, database_id, charset),
					  profile_handler, database_id),
			 pagenum, 6); //REALLY Extra weight for meta stuff.
    }
  }
  return pagenum;
}

						   
void dump_to_disc(string database_id, object profile_handler)
{
  string rawmap;
  string word;

  array(string) page_index=0;
  array(string) page_value=0;
  array(string) word_index=0;
  
  array(string) typisk;
  array(string) typisk2;
  
  mapping mem_word_map=0;
  mapping disc_word_map=0;
  
  float percent;
  object db;
  
  int i,j,k,timus,loopsteps;
  string current_read;
  string current_write;
  string raw_data;
  string already_exist;
  string fname,fname2;
  
  mapping obsoletes=([]);

  string tempmap;
  
  timus=time();

  debugdump("Im asked for a save of "+database_id+"...");
  debugdump("IM CALLED I know of "+sizeof(indices(huge_index[database_id]))+" words");
  
  current_write="n";
  
  fname=((profile_handler->working_at(database_id))+
	 current_write+"."+database_id)+".yabu"; //.pages.gdbm";
  debugdump("\n["+database_id+"] Writing data to: "+fname+"\n");
    
  db=Yabu.lookup(fname, "wS");

  //////////////////////////   pages   ///////////////////////////////////////////

  
  // If the loop runs twice, this only has to be done once....
  if(!page_index)  
  {
    page_index=indices(pages[database_id]); 
    page_value=values(pages[database_id]);
  }

  for(i=0; i<sizeof(page_index); i++)
  {
    tempmap=page_value[i];
    object gz = Gz;
    if(sizeof(indices(gz)))
      tempmap=gz->deflate()->deflate( tempmap );
    // Map a pagenumber to the COMPACT form.
    db["pages"][(string)page_index[i]]=page_value[i];
    // Map the URL to the page number.
    db["pages"][ (string)explode(page_value[i],DELI)[0] ] = (string)page_index[i];  
  }
  
  debugdump("["+database_id+"] Storing Pagecounter 'symb': "+symb[database_id]);
  db["pages"]["$symb$"]= ""+symb[database_id];

  //////////////////////////   index   ///////////////////////////////////////////
  
  k=0;
  // if(!word_index) // If the loop runs twice, this only has to be done once....
  word_index=indices(huge_index[database_id]);
  debugdump("Pushing the current index into Yabu database indexfile ("+
	    sizeof(word_index)+" words)");
    
  foreach(word_index, word)
  {
    // This fetches out a mapping from the memory index.
    mem_word_map=huge_index[database_id][word];
    // Fetch string (an encoded mapping). 
    if (rawmap=(db["index"][word]))
    {
      k++;
      // Decode the mapping.
      disc_word_map=decode_value(rawmap);  
    }
    else
      disc_word_map=([]);
    db["index"][word]=(string)(encode_value(mem_word_map+disc_word_map));
  }
  debugdump("("+(time()-timus)+" sec)");
  debugdump("Done appending INDEX to database. Of the stored "+sizeof(word_index)+
	    " terms, "+k+" already were in the index.");
  
  //////////////////////////   cut   ///////////////////////////////////////////
  
  k=0;
  debugdump("Appending: "+fname2);
  // not create, just read write!! Appending things!
  word_index=indices(cut_huge_index[database_id]);
  debugdump("Pushing the current cut index into Yabu database indexfile ("+
	    sizeof(word_index)+" words)");
  foreach(word_index, word)
  {
    // This fetches out a mapping from the memory index.
    typisk=cut_huge_index[database_id][word];
    // Fetch string (an encoded mapping). 
    if (rawmap=(db["cut"][word]))
    {
      k++;
      // Decode the array
      typisk2=decode_value(rawmap);  
    }
    else
      typisk2=({});
    db["cut"][word]=(string)(encode_value(typisk|typisk2));
  }
  debugdump("("+(time()-timus)+" sec)");
  
  if ((profile_handler->glob_active(database_id))=="1")
    debugdump("Done appending Cut-INDEX to database. Of the stored "+
	      sizeof(word_index)+" terms, "+k+" already were in the index.");
  else
    debugdump("Database left empty, as the wildcard option for "
	      "this profile is set to no.");
  
  //////////////////////////   full   ///////////////////////////////////////////
  
  k=0;
  
  debugdump("Appending: "+fname2);
  // not create, just read write!! Appending things!
  debugdump("Pushing the current full text into Yabu database indexfile ("+
	    sizeof(full_text_storage[database_id])+" pages)");
  
  foreach(indices(full_text_storage[database_id]), int pagenum)
  {
    k+=sizeof(full_text_storage[database_id][pagenum]);
    db["full"][(string)pagenum]=full_text_storage[database_id][pagenum];
  }
  
  debugdump("("+(time()-timus)+" sec)");
  db->sync();

  if ((profile_handler->quoted_active(database_id))=="1")
  {
    object gz = Gz;
    if(sizeof(indices(gz)))
      debugdump("The FullText data was packed from "+
		global_cnt_bytes+" down to "+k+" bytes using GZ.");
    else
      debugdump("Warning! No GZ found in Pike. Used uncompressed data!.");
    debugdump("Done appending FullText to database. "+k+" bytes stored.");
  }
  else
    debugdump("Database left empty, as the quoted search "
	      "option for this profile is set to no.");

  /////////////////////////////////////////////////////////////////////////////

  global_cnt_bytes=0;
  pages[database_id]=([]); // save mem
  huge_index[database_id]=([]); // save mem
  cut_huge_index[database_id]=([]);
  full_text_storage[database_id]=([]);
  
}

string destroy_fname;

string wake_up_create(string database_id, object profile_handler)
{
  object databas;
  string fname;
  
  debugdump("Creating "+database_id+" at "+ctime(time()));
  symb[database_id]=1;
  
  pages[database_id] = ([]);
  huge_index[database_id] = ([]); 
  cut_huge_index[database_id]  = ([]);
  full_text_storage[database_id]=([]);
  global_cnt_bytes=0;

  if (get_dir((profile_handler->working_at(database_id)))==0)
    return "No such working directory: "+profile_handler->working_at(database_id);

  if (get_dir((profile_handler->store_at(database_id)))==0)
    return "No such storage directory: "+profile_handler->store_at(database_id);
  
  fname=((profile_handler->working_at(database_id))+"n."+database_id)+".yabu";

//   if(file_stat(fname+"/lock"))
//     return "Crawling already in progress.";
  
  remove_dir(fname);
  object db=Yabu.lookup(fname,"wS");
//   Stdio.File(fname+"/lock","cw")->write("Kilroy was here");
//   destroy_fname=fname+"/lock";
  db["index"]->purge;
  db["pages"]->purge;
  db->sync();
  debugdump("Start at "+fname);
  
  return 0;
}

void destroy()
{
//   if(destroy_fname)
//     rm(destroy_fname);
}

void wake_up_restore(string database_id, object profile_handler)
{
  string temporar;
  int tempsymb;
  object db;
  
  debugdump("Assuming:  "+((profile_handler->working_at(database_id))+
			   "n."+database_id)+".yabu");

  debugdump("Restoring symbol from "+((profile_handler->working_at(database_id))+
				      "n."+database_id)+".yabu  (pages)");
  db=Yabu.lookup(((profile_handler->working_at(database_id))+"n."+database_id)+
	       ".yabu", "r");
  temporar=db["pages"]["$symb$"];
    
  if(-1==sscanf(temporar, "%d", tempsymb))
    debugdump("Serious error! Couldn't initialize the unique page counter.... ");
  
  symb[database_id]=tempsymb;
  debugdump("Last added page was "+tempsymb);
  
  pages[database_id] = ([]);
  huge_index[database_id] = ([]);
  full_text_storage[database_id]=([]);
  cut_huge_index[database_id]  = ([]);
  global_cnt_bytes=0;
}

void set_robot_home(string foo)
{
  robot_home=foo;
}
