//
// This source-code is a part of the Roxen IntraSeek search engine package.
//  1996-1998 Idonex AB
// Email: info@idonex.se 
//


/*
  This object handles the profiles.

  The crawler (scheduler.pike) uses it to parse the profile configurations
  with, and the auto-updated uses it to classify an URL to the corresponding
  database.
  
  It reads the profile configs from disc, and parses them.
  Must be initialized using load_and_parse_profiles(string fullfilename)
  where fullfilename is the path to the profile.

  It has these functions that can be externally called:
      string classify(url);
      control_url(string profil, string url);
      load_and_parse_profiles(string fullfilename)
      active_profiles();
      
*/

#include "crawler.h"

import Stdio;

#define QUERY_LOG_DETAIL crawler_log_detail

mapping profiles=([]);

mapping accept_regexps=([]);;
mapping avoid_regexps=([]);;

mapping original_accept_regexps=([]);;
mapping original_avoid_regexps=([]);;

string *stoplists;
string robot_home;

int inited_stop=0;

int day;      // 0=any, 1=monday, etc
int per_week; // 0=don't care, 1=every 2=every second etc.
int hour;     // 0=midnight
int xxtime;


string working_at(string id)
{
  if ((profiles[id])  && (profiles[id]["working_path"]))
    return profiles[id]["working_path"];
  return DEFAULT_WORKING_AT;
}


void log(string profile, string data)
{
  write_file(working_at(profile)+"crawler."+profile+".log", data+"\n");
}

string view_log(string profile)
{
  if (!file_stat(working_at(profile)+"crawler."+profile+".log"))
    return "Error: Couldnt find any crawler log at "+working_at(profile)+"crawler."+profile+".log";
  return read_bytes(working_at(profile)+"crawler."+profile+".log");
}


int str2int(string s)
{
  int k;
  if (s && s!="")
    if(-1==sscanf(s, "%d", k))
    {
      return 0;
    }
  return k;
}


string delete_log(string profile)
{
  rm(working_at(profile)+"crawler."+profile+".log");
}

void set_fall_back_accept(string profil, string defaccept)
{
  // Acceptpattern if * defined. like by IPP!
  profiles[profil]["defaccept"]=defaccept;
}

string *fall_back_accept(string profil)
{

  return ({ profiles[profil]["defaccept"] });
}

void set_fall_back_world(string profil, string defworld)  // Startpage if "/" defined! Like by Ipp!
{
  profiles[profil]["defworld"]=defworld;
}

string *fall_back_world(string profil)
{
  return ({ profiles[profil]["defworld"] });
}

string *active_profiles()
{
  return (indices(profiles)-({})-({""}));
}

object *query_accept_regexps(string id)
{
}

object *query_avoid_regexps(string id)
{
}

mapping get_profile(string id)
{
  return profiles[id];
}

string store_at(string id)
{
  if ((profiles[id]) && (profiles[id]["store_path"]))
    return profiles[id]["store_path"];
  return robot_home+"databases/";
}

int safety_save(string id)
{
  if ((profiles[id]) && (profiles[id]["safety_save"]))
    return profiles[id]["safety_save"];
  return DEFAULT_SAFETY_SAVE;

}

int max_doc_dl(string id)
{
  if ((profiles[id]) && (profiles[id]["max_doc_dl"]))
    return profiles[id]["max_doc_dl"];
  return DEFAULT_MAX_DOC_DL;
    
}

string *stoplist(string id)
{
  if ((profiles[id]) && (profiles[id]["stoplist"]))
    return profiles[id]["stoplist"];
  return ({ });
}

int max_download_bytes(string id)
{
  if ((profiles[id]) && (profiles[id]["max_page_fetch"]))
    return profiles[id]["max_page_fetch"];
  return DEFAULT_MAX_BYTES_FETCH;
 
}

int stopword_suggest_percent(string id)
{ // obsolete!
  return profiles[id]["stopword_suggest_percent"];
}

string *ind_stoplist(string id)
{
  if(profiles[id] && arrayp(profiles[id]["ind_stoplist"]))
  {
    return profiles[id]["ind_stoplist"];
  }
  else
    return ({ "" });
}

string *allstoplists(string id)
{
  if(profiles[id] && ((profiles[id]["stoplists"]||"")!=""))
    return profiles[id]["stoplists"];
  return ({ DEFAULT_STOPLIST });
}

string *start_pages(string id)
{
  if(profiles[id] && (profiles[id]["start_pages"]) && (profiles[id]["start_pages"]!="") && (profiles[id]["start_pages"]!="0"))
    return profiles[id]["start_pages"];
  else
    return ({});
}

string *avoid_pattern(string id)
{
   if(profiles[id] &&  (profiles[id]["avoid_pattern"]) && (profiles[id]["avoid_pattern"]!="") && (profiles[id]["avoid_pattern"]!="0"))
    return profiles[id]["avoid_pattern"];
  else
    return ({ });
	  
}

string *accept_pattern(string id)
{
  if(profiles[id]  && (profiles[id]["accept_pattern"]) && (profiles[id]["accept_pattern"]!="") && (profiles[id]["accept_pattern"]!="0"))
    return profiles[id]["accept_pattern"];
  else
    return ({});
}

string realname(string id)
{
  if(profiles[id] && (profiles[id]["realname"]!=""))
    return ""+profiles[id]["realname"];
  return "";
}

string query_log_active(string id)
{
  // 0=undef, 1=on, 2=off.
  if(profiles[id] && (profiles[id]["query_log_active"]))
    return (profiles[id]["query_log_active"]);
  return DEFAULT_QUERY_LOG_ACTIVATED;
}

string active(string id)
{
  // 0=undef, 1=on, 2=off.
  if(profiles[id] && (profiles[id]["active"]))
    return (profiles[id]["active"]);
  return DEFAULT_PROFILE_ACTIVATED;
}

int query_log_max_size(string id)
{
  if(profiles[id] && (profiles[id]["query_log_max_size"]))
    return (profiles[id]["query_log_max_size"]);
  return DEFAULT_QUERY_LOG_MAX_SIZE;
}

int quicklinks(string id)
{
  if(profiles[id] && (profiles[id]["quicklinks"]))
    return (int)(profiles[id]["quicklinks"]);
  return (int)DEFAULT_QUICK_LINKS;
}

int entries(string id)
{
  if(profiles[id] && (profiles[id]["entries"]))
    return (int)(profiles[id]["entries"]);
  return (int)DEFAULT_ENTIRES;
}

int cutoff(string id)
{
  if(profiles[id] && (profiles[id]["cutoff"]))
    return (int)(profiles[id]["cutoff"]);
  return (int)DEFAULT_CUTOFF;
}

int timeout(string id)
{
  if(profiles[id] && (profiles[id]["timeout"]))
    return (int)(profiles[id]["timeout"]);
  return (int)DEFAULT_TIMEOUT;
}

int walkpause(string id)
{
  if(profiles[id] && (profiles[id]["walkpause"]))
    return (int)(profiles[id]["walkpause"]);
  return (int)DEFAULT_WALK_DELAY;
}


int crawler_schedule_1(string id)
{
  if ((profiles[id]) && (1!=zero_type(profiles[id]->schedule_1)))
    return (int)(profiles[id]["schedule_1"]);
  return (int)DEFAULT_SCHEDULE_1;
}

int crawler_schedule_2(string id)
{
  if ((profiles[id]) && (1!=zero_type(profiles[id]->schedule_2)))
    return (int)(profiles[id]["schedule_2"]);
  return (int)DEFAULT_SCHEDULE_2;
}

int crawler_schedule_3(string id)
{
  if ((profiles[id]) && (1!=zero_type(profiles[id]->schedule_3)))
    return (int)(profiles[id]["schedule_3"]);
  return (int)DEFAULT_SCHEDULE_3;
}


int crawler_log_detail(string id)
{
  // 0=full (default), 1=medium, 2=short, 3=none (except fatal errors)
  if(profiles[id] && (profiles[id]["detail"])<4)
    return (profiles[id]["detail"]);
  return DEFAULT_CRAWLER_LOG_LEVEL;
}

string struclogactive(string id)
{
  // "0"=undef, "1"=on, "2"=off.
  if(profiles[id] && (profiles[id]["struclogactive"]))
    return (profiles[id]["struclogactive"]);
  return DEFAULT_STRUC_LOG_LOG_ACTIVATED;
}


int nice_increment(string id)
{
  // 0 = no nice.
  
  if ((profiles[id]) && (profiles[id]["nice_increment"]))
  {
    if ((int)(profiles[id]["nice_increment"]) > NICE_MAX)
      return (int)NICE_MAX;
    if ((int)(profiles[id]["nice_increment"]) < NICE_MIN)
      return (int)NICE_MIN;
    return (int)(profiles[id]["nice_increment"]);
  }
  return (int)DEFAULT_NICE_INCREMENT;
}


string glob_active(string id)
{
  // "0"=undef, "1"=on, "2"=off.
  if(profiles[id] && (profiles[id]["glob_active"]))
    return (profiles[id]["glob_active"]);
  return DEFAULT_GLOB_ACTIVATED;
}

string quoted_active(string id)
{
  // "0"=undef, "1"=on, "2"=off.

  if(profiles[id] && (profiles[id]["quoted_active"]))
    return (profiles[id]["quoted_active"]);
  return DEFAULT_QUOTED_ACTIVATED;
}

int summary_text_length(string id)
{
  // This is the length of the sample text
  // (summary) displayed at search results.
  if(profiles[id] && (profiles[id]["summary_text_length"]))
    return (int)(profiles[id]["summary_text_length"]);
  return DEFAULT_SUMMARY_TEXT_LENGTH;
}

string cookie_bunch(string id)
{
//  werror("id="+id+"\n");
//  werror(sprintf("Cookies: %O", profiles[id]["cookie_bunch"]));

  if(profiles[id] && (profiles[id]["cookie_bunch"]))
    return (profiles[id]["cookie_bunch"]);
  return "";
}


/////////////////////////////////////////////////////////////////////////////


string *explode_file(string filename) //*handle_includes(string *raw_data, string path)
{
  string *result=({});

  if(!file_stat(filename))
  {
    // log("Couldn't include file:"+filename+"\n");
    return ({});
  } 
  result+=replace((read_bytes(filename)||""), " ", "\n") / "\n";
  return result;
}

string *default_avoids(string path)
{
  string s;
  if (!file_stat(path+DEFAULT_AVOID))
    return ({ "error" });
  if (s=read_bytes(path+DEFAULT_AVOID))
    return ((s/"\n")*" ")/" "-({""});
}

string *available_stoplists(string path)
{
  string test;
  object st=Regexp(".*stop$");
  string *raw=get_dir(path);
  stoplists=({});
  foreach(raw, test)
    if (st->match(test))
      stoplists+=({ test });
  return stoplists;
}

void create_profile(string id, array data)
{
  if (!profiles[id])
    profiles[id]=([]);
  if ((!data[0]) || (data[0]==""))
    profiles[id]["realname"]="noname";
  else
    profiles[id]["realname"]=replace(data[0], ({"\n"}), ({""}) ); // No spaces!
  profiles[id]["store_path"]=data[1];
  profiles[id]["working_path"]=data[2];
  if(profiles[id]["working_path"]=="")
   profiles[id]["working_path"]=DEFAULT_WORKING_AT;
  profiles[id]["accept_pattern"]=data[3];
  if(!profiles[id]["accept_pattern"])
    profiles[id]["accept_pattern"]=({});
  profiles[id]["avoid_pattern"]=data[4];
  profiles[id]["start_pages"]=data[5];
  if(!profiles[id]["start_pages"])
    profiles[id]["start_pages"]=({});
  profiles[id]["ind_stoplist"]=data[6];
  if((!profiles[id]["ind_stoplist"]) || (profiles[id]["ind_stoplist"]==""))
    profiles[id]["ind_stoplist"]=({});
  profiles[id]["stoplists"]=data[7];
  profiles[id]["safety_save"]=data[8];
  profiles[id]["max_doc_dl"]=(int)data[9];
  profiles[id]["max_page_fetch"]=(int)data[10];
  profiles[id]["query_log_active"]=data[11];
  profiles[id]["query_log_max_size"]=(int)data[12];
  
  profiles[id]["active"]=data[13];
  profiles[id]["quicklinks"]=(int)data[14];
  profiles[id]["entries"]=(int)data[15];
  profiles[id]["detail"]=(int)data[16];
  profiles[id]["cutoff"]=(int)data[17];
  profiles[id]["timeout"]=(int)data[18];
  profiles[id]["walkpause"]=(int)data[19];
  profiles[id]["struclogactive"]=data[20];
  profiles[id]["schedule_1"]=(int)data[21];
  profiles[id]["schedule_2"]=(int)data[22];
  profiles[id]["schedule_3"]=(int)data[23];
  profiles[id]["nice_increment"]=(int)data[24];
  profiles[id]["glob_active"]=data[25];
  profiles[id]["quoted_active"]=data[26];
  profiles[id]["summary_text_length"]=(int)data[27];
  profiles[id]["cookie_bunch"]=data[28];
}
  

void remove_profile(string which)
{
  m_delete(profiles, which);
}

void save_profiles(string path)
{
  string this;
  string bar;

  object filen=File();
  filen->open(path+"profiles.old", "wtc");
  filen->write(""+read_bytes(path+"profiles.bak"));
  filen->close();
  filen->open(path+"profiles.bak", "wtc");
  filen->write(""+read_bytes(path+"profiles.txt"));
  filen->close();
  filen->open(path+"profiles.txt", "wtc");
  filen->write("//\n"
               "// This is the IntraSeek crawler configuration file.\n"
               "// Dont modify it by manually, use the web based config interface.\n"
               "//\n");
  foreach(active_profiles(), this)
  {
    if ((this) && (this!="0") && (this!="") && (this!="") && (profiles[this]))
    {
      filen->write("$start "+this+"\n"
                   "$realname "+replace(""+realname(this), " ", PROFILES_DELI)+"\n"
                   "$store_path "+store_at(this)+"\n"
                   "$working_path "+working_at(this)+"\n"
                   "$accept_pattern "+(accept_pattern(this)*"\n")+"\n"
                   "$avoid_pattern "+((avoid_pattern(this)*"\n")-"\r")+"\n"
                   "$start_pages "+(start_pages(this)*"\n")+"\n"
                   "$stoplist "+(ind_stoplist(this)*"\n")+"\n");
      if (allstoplists(this))
	foreach(allstoplists(this), bar)
	  filen->write("#include="+bar+"\n");
      filen->write("$safety_save "+safety_save(this)+"\n"
                   "$max_doc_dl "+max_doc_dl(this)+"\n"
                   "$max_page_fetch "+max_download_bytes(this)+"\n"
                   "$query_log_active "+query_log_active(this)+"\n"
                   "$query_log_max_size "+query_log_max_size(this)+"\n"
                   "$active "+active(this)+"\n"
                   "$quicklinks "+quicklinks(this)+"\n"
                   "$entries "+entries(this)+"\n"
                   "$detail "+crawler_log_detail(this)+"\n"
                   "$cutoff "+cutoff(this)+"\n"
                   "$timeout "+timeout(this)+"\n"
                   "$walkpause "+walkpause(this)+"\n"
		   "$struclogactive "+struclogactive(this)+"\n"
                   "$schedule_1 "+crawler_schedule_1(this)+"\n"
                   "$schedule_2 "+crawler_schedule_2(this)+"\n"
                   "$schedule_3 "+crawler_schedule_3(this)+"\n"
                   "$nice_increment "+nice_increment(this)+"\n"
                   "$glob_active "+glob_active(this)+"\n"
                   "$quoted_active "+quoted_active(this)+"\n"
                   "$summary_text_length "+summary_text_length(this)+"\n"
                   "$cookie_bunch "+cookie_bunch(this)+"\n"
                   "$end "+this+"\n"
                   "_______________________________________________________________________________\n"
                   "-------------------------------------------------------------------------------\n");
    }
  }
  filen->close();
}


void load_and_parse_profiles(string path, string filename)
{
  string tem, temp_str, current_profile;
  string *tem2, *temarr;
  string parsing;
  string current_profile;
  string fullfilename;
  mapping current;
  int i;
  
  profiles=([]); // empty - full reset

  fullfilename=path+filename;
  if(!file_stat(fullfilename))
     {
       return 0;
     }
  tem=read_bytes(fullfilename);
  tem=replace(tem," ","\n"); 
  tem=replace(tem, PROFILES_DELI ," ");
  tem=((tem-"\r")-"\t");
  
  tem2=tem/"\n";

  current_profile="";
  parsing="";
  foreach(tem2,tem)
  {
    if((tem[0..1]!="//") && (tem!="") && (tem!="0") && (tem))
    {
    if(tem[0..0]=="$")
    {
      parsing=tem[1..99];
    } else
    {
      switch(parsing)
      {
       case "start":
	if ((tem=="") || (tem==" "))
	  current_profile="unknown";
	else
	  current_profile=tem;
	current=([]);
	break;
      case "end":
	profiles[current_profile]=current;
	current_profile="";
	break;
      case "realname":
       if (current["realname"])
	 current["realname"]=current["realname"]+" "+tem;
       else
	 current["realname"]=tem;
       break;
       case "accept_pattern":
	if (current["accept_pattern"])
	  current["accept_pattern"]=({tem})+current["accept_pattern"];
	else
	  current["accept_pattern"]=({tem});
	break;
       case "avoid_pattern":
	if (current["avoid_pattern"])
	{
	  //werror("kultig");
	  if(tem[0..8]=="#include=")
	    current["avoid_pattern"]=explode_file(path+"resource/"+tem[9..99])+current["avoid_pattern"];
	  else
	    if(tem!="" && tem!="\n" && tem!=" ")
	      current["avoid_pattern"]=({tem})+current["avoid_pattern"];
	}
	else
	  if(tem[0..8]=="#include=")
	    current["avoid_pattern"]=explode_file(path+"resource/"+tem[9..99]);
	  else
	  {
	    if(tem!="" && tem!="\n" && tem!=" ")
	      current["avoid_pattern"]=({tem});
	  }
	break;
       case "start_pages":
	if (current["start_pages"])
	  current["start_pages"]=({tem})+current["start_pages"];
	else
	  current["start_pages"]=({tem});
	break;
       case "stoplist":
	if(sizeof(tem)>9&&tem[0..8]=="#include=")
	{
	    if(current["stoplists"])
	      current["stoplists"]=({tem[9..99]})+current["stoplists"];
	    else
	      current["stoplists"]=({tem[9..99]});

	    temarr=explode_file(path+"resource/"+tem[9..99]);
	     
	    if (current["stoplist"])
	      current["stoplist"]=temarr+current["stoplist"];
	    else
	      current["stoplist"]=temarr;
	     
	}  else 
	  if(current["ind_stoplist"])
	    current["ind_stoplist"]=({tem})+current["ind_stoplist"];
	  else
	    current["ind_stoplist"]=({tem});
	break;
       case "store_path":
	current["store_path"]=tem;
	break;
       case "working_path":
	current["working_path"]=tem;
	break;
       case "safety_save":
	if(-1==sscanf(tem, "%d", i))
	  i=1000;
	else
	  current["safety_save"]=i;
	break;
       case "max_doc_dl":
	if(-1==sscanf(tem, "%d", i))
	  i=10000;
	else
	  current["max_doc_dl"]=i;
	break;
       case "stopword_suggest_percent":
	if(-1==sscanf(tem, "%d", i))
	  i=10000;
	else
	  current["stopword_suggest_percent"]=i;
	break;
       case "max_page_fetch":
	if(-1==sscanf(tem, "%d", i))
	  i=150000;
	else
	  current["max_page_fetch"]=i;
	break;

	// New since alfa 7:
	
       case "active":
	current["active"]=tem;
	break;
       case "query_log_active":
	current["query_log_active"]=tem;
	break;
       case "query_log_max_size":
	current["query_log_max_size"]=str2int(tem);
	if ((current["query_log_max_size"])==0)
	  current["query_log_max_size"]=DEFAULT_QUERY_LOG_MAX_SIZE;
	break;
       case "quicklinks":
	current["quicklinks"]=str2int(tem);
	if ((current["quicklinks"])<1)
	  current["quicklinks"]=(int)DEFAULT_QUICK_LINKS;
	break;
       case "entries":
	current["entries"]=str2int(tem);
	if ((current["entries"])<1)
	  current["entries"]=(int)DEFAULT_ENTIRES;
	break;
       case "detail":
	current["detail"]=str2int(tem);
	if (current["detail"]>3)
	  current["detail"]=(int)DEFAULT_CRAWLER_LOG_LEVEL;
	break;
       case "walkpause":
	current["walkpause"]=str2int(tem);
	break;
       case "cutoff":
	current["cutoff"]=str2int(tem);
	if ((current["cutoff"])<1)
	  current["cutoff"]=(int)DEFAULT_CUTOFF;
	break;
       case "timeout":
	current["timeout"]=str2int(tem);
	if ((current["timeout"])<1)
	  current["timeout"]=(int)DEFAULT_TIMEOUT;
	break;
       case "schedule_1" :
	current["schedule_1"]=str2int(tem);
	break;
       case "schedule_2" :
	current["schedule_2"]=str2int(tem);
	break;
       case "schedule_3" :
	current["schedule_3"]=str2int(tem);
	break;
       case "struclogactive":
	current["struclogactive"]=tem;
	break;
       case "nice_increment" :
	current["nice_increment"]=str2int(tem);
	break;
       case "glob_active" :
	current["glob_active"]=tem;
	break;
       case "quoted_active" :
	current["quoted_active"]=tem;
	break;
       case "summary_text_length" :
	current["summary_text_length"]=str2int(tem);
	break;
       case "cookie_bunch":
	current["cookie_bunch"]=tem;
	break;
      }
    }
    }
  }
  
}

void compile_all_regexps()
{
  string  temp_str, current_profile;

  foreach(active_profiles(), current_profile)
  {
    accept_regexps[current_profile]=({});
    avoid_regexps[current_profile]=({});
    if (avoid_pattern(current_profile))
      foreach(avoid_pattern(current_profile),temp_str)  // Compile all avoids to regexps.
      {
	if(temp_str!="")
	{
	  temp_str=replace(temp_str,".","\\.");   // ".+" is any number of any char.
	  temp_str=replace(temp_str,"*",".*");   // ".+" is any number of any char.
	  temp_str=temp_str+"$";
	  avoid_regexps[current_profile]+=({Regexp(temp_str)});
	}
      }

    string *accepts;

    if(accept_pattern(current_profile))
    {
      if ((sizeof(accept_pattern(current_profile))) && (accept_pattern(current_profile)[0]=="/*" ))  // Check for /* which is an IPP special.
	accepts=fall_back_accept(current_profile);  // Ipp uses this.
      else
	accepts=accept_pattern(current_profile);
      
      foreach(accepts,temp_str)
      {
	if ((temp_str||"")!="")
	{
	  temp_str=replace(temp_str,".","\\.");   // ".+" is any number of any char.
	  temp_str=replace(temp_str,"*",".*");   // ".+" is any number of any char.
	  temp_str=temp_str+"$";
	  accept_regexps[current_profile]+=({Regexp(temp_str)});
	}
      }
    }
  }
}


int control_url(string id, string in_url)
{
  //
  // Given an URL and an ID, it says if the URL is OK to include in the
  // database or not.
  // Simply returns true or false.
  //

  object cx;
  int i=0;
//  log(id, "Checking for id "+id+" '"+in_url+"' with regexps:");

  for(i=0; i<sizeof(avoid_regexps[id]); i++)
  {
    cx=avoid_regexps[id][i];
    if (cx->match(in_url)) {
      if (QUERY_LOG_DETAIL(id)<LOG_LEVEL_MEDIUM)
	log(id," Rejected "+in_url+" Matched ["+(avoid_pattern(id))[i]+"]");
      return 0;
    }
  }
  for(i=0; i<sizeof(accept_regexps[id]); i++)
  {
    cx=accept_regexps[id][i];
    if (cx->match(in_url)) {
      if (QUERY_LOG_DETAIL(id)<LOG_LEVEL_MEDIUM)
	log(id," Accepted "+in_url+" Matched ["+(accept_pattern(id))[i]+"]");
      return 1;
    }
  }
  
  if (QUERY_LOG_DETAIL(id)<LOG_LEVEL_MEDIUM)
    log (id," Rejected (no accept rule for "+in_url+")");
  return 0;
  
}

string *classify(string url)  // Returns all profiles that the URL fits to!
{
  //
  // Give this function an URL and it will return all the ID classes that
  // the URL fits to!
  //
  string current_profile;
  string *temp=({});
  foreach(active_profiles(), current_profile)
    if(control_url(current_profile, url))
      temp+=({current_profile});
  return ((temp==({}))?0:temp);
}

mapping load_robot_id(string profile)
{
  string snoddas;
  mapping tempmap=([]);
  string *tempogra;
  if(file_stat(robot_home+"crawlerid/"+profile))
  {
    snoddas=read_bytes(robot_home+"crawlerid/"+profile);
    if (snoddas && snoddas!="")
    {
      tempogra=snoddas/"\n";
      tempmap["pid"]=tempogra[0];
      tempmap["start"]=tempogra[1];
      tempmap["stop"]=tempogra[2];
      tempmap["status"]=tempogra[3];
      if (sizeof(tempogra)>4)
	tempmap["modified"]=tempogra[4];
      return tempmap;
    }
  }
  return 0;
}

void delete_robot_id(string profile)
{
  log(profile, "Removing  "+profile+"s id file");
  rm(robot_home+"crawlerid/"+profile);
}

void create_robot_id(string profile, string pid, string start, string stop, string status)
{
  object filen;
  filen=File();
  rm(robot_home+"crawlerid/"+profile);
  filen->open(robot_home+"crawlerid/"+profile, "wtc");
  filen->write(pid+"\n"+start+"\n"+stop+"\n"+status+"\n"+((string)time()));
  filen->close();
}

void modify_robot_id(string profile, string stop, string status)
{
  object filen;
  mapping temporar;
  temporar=load_robot_id(profile);
  if (temporar)
    create_robot_id(profile, temporar["pid"], temporar["start"], ""+stop, ""+status);
}
		    
int pid_exist(string pid)
{
/*   If sig is 0  (the  null  sig-*/
/*     nal),  error checking is performed but no signal is actually*/
/*     sent.  This can be used to check the validity of pid.*/
  int pp;
  if (pp=str2int(pid))
    return (kill(pp, 0));
  else
    return 0;
}

int has_crawlerid(string profile)
{
  if(file_stat(robot_home+"crawlerid/"+profile))
    return 1;
  else
    return 0;
}

int is_running(string profile)
{
  mapping tillf;
  if (tillf=load_robot_id(profile))
  {
    if (pid_exist(tillf["pid"]))
      return 1;
    return 0;
  } else
    werror("No id file?");
}

void consist_log(string profile, string data)
{
  write_file(store_at(profile)+profile+".cons.log.html", data+"\n");
}

void delete_consist_log(string profile, string data)
{
  rm(store_at(profile)+profile+".cons.log.html");
}
string view_consist_log(string profile)
{
  if (!file_stat(store_at(profile)+profile+".cons.log.html"))
    return "Error: Couldn't find any site structure log at "+store_at(profile)+profile+".cons.log.html";
  return read_bytes(store_at(profile)+profile+".cons.log.html");
}

string describe_time(int seconds)
{
  if (seconds>31104000)
    return "more than a year";
  else
    if (seconds>15552000)
      return "more than half a year";
    else
      if (seconds>5184000)
        return "more than two months";
      else
        if (seconds>2592000)
          return "more than a month";
        else
        {
          int d,h,m,s;
          string r="";
          if (seconds>=86400)  r+=""+(d=(seconds/86400)%30)+" day"+(d==1?"":"s")+", ";
          if (seconds>=3600)  r+=""+(h=(seconds/3600)%24)+" hour"+(h==1?"":"s")+", ";
          if (seconds>=60)  r+=""+(m=(seconds/60)%60)+" minute"+(m==1?"":"s")+" and ";
          r+=""+(s=(seconds%60))+" second"+(s==1?"":"s");
          return r;
        }
  return "?";
}

int find_next(int o)
{
  mapping m;
  int t = o || xxtime;

  m = localtime(t);
  m->hour=hour; m->min=0; m->sec=0;
  switch(per_week)
  {
   case 3:
    if(m->mday > 8) return find_next(t+86400*7);
    break;
   case 2:
    if((m->yday/7)%2) return find_next(t+86400*7);
    break;
   case 0:
    return -1;
  }

  if(day)
  {
    m->mday -= m->wday-1;
    m->mday += day-1;
    if(mktime(m) <= xxtime) return find_next(t+86400*7);
  } else
    if(mktime(m) <= xxtime) return find_next(t+86400);
//  write("Time to run is "+ctime(mktime(m)));
  return mktime(m);
}


int next_schedule_start(string id, int xtime)
{

  per_week=crawler_schedule_1(id);
  day=crawler_schedule_2(id);
  hour=2*crawler_schedule_3(id);
  xxtime=xtime;
  return find_next(xtime);
}

void set_robot_home(string foo)
{
  robot_home=foo;
}

string robot_from;

void set_all_robots_from(string foo)
{
  robot_from=foo;
}

string all_robots_from()
{
  return robot_from||"";
}
