//
// This source-code is a part of the Roxen IntraSeek search engine package.
//  1996-1998 Idonex AB
// Email: info@idonex.se 
//

/*
  This scheduler complies with the standard /robots.txt rules
  as defined in http://info.webcrawler.com/mak/projects/robots/norobots.html
*/

#if constant(Process.create_process)
import ".";
#endif

#include "crawler.h"

inherit "url.pike"; 
import Simulate;

#define QUERY_LOG_DETAIL profile_handler->crawler_log_detail
//  Leave untouched

string robot_home;

string *tovisit;
string *beenat;
string toret;
string uri, site, protocol;
int port;
program p;     // the single fetcher!
mapping robot_txt;

object special_fetcher, profile_handler;
string global_profil;

object global_struc_databas;

mapping (string:int)url_to_id=([ ]);
mapping (int:string)id_to_url=([ ]);
mapping used_error_list=([ ]);


int global_gensym=1;

void crawlerlog(string txt)
{
  Stdio.write_file(robot_home+"logs/crawler.log", txt+"\n");
}

void plog(string data)
{
  if(data && profile_handler)
    profile_handler->log(global_profil, "Schedul: "+data);
}

string *parse_robot_html(string robottxt, string internal_site,
			 string internal_protocol, int internal_port)
{
  int start, end, temp, rejected, i, j;
  string temptxt;
  string *robotlines;
  string *etc;
  string *collect_rejected;

  // tar en robot.txt, protocol, site, port och
  // returnerar en array med frbjudna starter p URLer,
  // tex. "http://www.lysator.liu.se/vfsh/"
  // plog("I WILL PARSE: "+robottxt);
  // Remove comments...
  
  while (-1!=(start=search(robottxt, "#")))   // ((s/"\n")-({""}))*"\n"
  {
    end=search(robottxt, "\n", start);
    temp=sizeof(robottxt);
    if (end==-1)
      end=temp;
    // Simply remove the comment! Nifty!
    robottxt=robottxt[0..((start>1)?(start-1):0)]+
      robottxt[((end<temp)?(end+1):end)..temp];  
  }
  
  // Format the inputted text nicely, removing double "\n\n"'s and double spaces.
  // robottxt=((robottxt/"\n")-({""}))*"\n";
  
  temp=1;
  while(temp)
  {
    temptxt=replace(robottxt, "\n\n", "\n");
    temptxt=replace(temptxt, "  ", " ");
    temp= (sizeof(robottxt)!=sizeof(temptxt));
    robottxt=temptxt;
  }
  // you can't convert all of robots.txt to lowercase, since it can
  // contain mixed case pathnames, which are perfectly legal
  // robottxt=lower_case(robottxt);
 
  // plog("I HAVE PARSED TO: "+robottxt);
  // now we can handle the text more easily!
  // so far, we only listen to * as we regard ourself as VFSH

  robotlines=robottxt/"\n";
  for(i=0; i<sizeof(robotlines); i++)
  {
    //plog("YY:"+robotlines[i]);
    temptxt = lower_case(robotlines[i]);
    if (temptxt[0..10]=="user-agent:")
    {
      // Several agents can appear on one line!
      etc=robotlines[i][10..sizeof(robotlines[i])]/" ";  
      rejected=0;
      
      // Check if we are one of these agents!
      for(j=0; j<sizeof(etc); j++)
	if(etc[j]=="*" || etc[j]==CRAWLER_ID)  
	{
	  collect_rejected=({});
	  rejected=1;
	  // plog("XX: I was rejected...  ...from:");
	}
    }
    if (temptxt[0..8]=="disallow:") 
    {
      if (rejected)
      {
	// Several agents can appear on one line!
	etc=robotlines[i][8..sizeof(robotlines[i])]/" ";  
	for(j=0; j<sizeof(etc); j++)
	{
	  if (sizeof(etc[j])>1)
	  {
	    temptxt=internal_protocol+"://"+internal_site+
	      ((internal_port==80)?"":":"+port)+etc[j];
	    collect_rejected+=({temptxt});
	    // plog("XX: among others: "+temptxt);
	  }
	}
      }
    }
  }
  return collect_rejected;
}


int robot_txt_check(string url)
{
  int return_code, i;
  string *xavoid;
  mixed *expand_result;
  mixed *fetcher_results;
  
  int return_code, local_port;
  string avoid, local_site, local_protocol;
  mixed *expand_result;
  string *forbidden_urls;
  
  expand_result = universal_expand_url(url);

  return_code =  expand_result[0];
  uri =  expand_result[1];
  site =  expand_result[2];
  port =  expand_result[3];
  protocol =  expand_result[4];

  if (!(avoid=robot_txt[""+site+((port==80)?"":":"+port)]))
  {
    object filters=Crawler.filters();
    fetcher_results=special_fetcher->fetch(
      "http://"+site+((port==80)?"":":"+port)+"/robots.txt",
      0, global_profil, profile_handler,0,filters);
    if ((fetcher_results[_SIZEOFDOC]>0) && (fetcher_results[_TEXT_BODY]!=""))
    {
      // INdeed, Only one site HAS been visited to fetch robots.txt!
      expand_result = universal_expand_url(fetcher_results[_URLS_VISI][0]);  
      local_site =  expand_result[2];
      local_port =  expand_result[3];
      local_protocol =  expand_result[4];
      forbidden_urls=parse_robot_html(fetcher_results[_TEXT_BODY],
				      local_site, local_protocol, local_port);
    }
    else
      forbidden_urls=({});  // If the server didnt HAVE any /robots.txt
    
    if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_MEDIUM)
      plog("Notice: Forbidding things for: "+""+site+((port==80)?"":":"+port));
    robot_txt[""+site+((port==80)?"":":"+port)]=forbidden_urls;
    if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_MEDIUM)
      plog("the fetcher 0 is finished fetching /robot.txt from: ["+
	   local_site+((port==80)?"":":"+port)+"]");
  }
  
  // When we get to this point, the fetcher have been invoked in a blocking mode (ie
  // no parallell running and processed the /robots.txt. Now the robot_txt should
  // be set as well, and we can easily do this:
    
  xavoid=robot_txt[""+site+((port==80)?"":":"+port)];      
  
  if(!xavoid)
  {
    if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_SHORT)
      plog("Warning: Empty, or Very bogus exclusion scheme in the mapping for"+
	 ""+site+((port==80)?"":":"+port));
    xavoid=({});
  }

// Now, check if the current URL in some way matches any of the URL's on the
// forbidden list for that specific server:
// the avoid array contains STARTS of arrays to avoid, ie:
// a entry like "http://www.lysator.liu.se/thomas/photo" forbids every URL
// under that photo path, as well as the photo path itself!
  
  for (i=0; i<sizeof(xavoid); i++)
  {
    // plog("checking: "+avoid[i]+" <-> "+url);
    if(-1!=search(url, xavoid[i]))
    {
      // plog("info: Excluded a denied page!");
      return 0;
    }
  }
  return 1;
}


string give_me_url()
{
  if(sizeof(tovisit||({}))>0)
  {
    toret=tovisit[0];
    tovisit=tovisit[1..sizeof(tovisit)];
    return toret;
  }
  else
    return "";
}

void ive_been_at(string *urls)
{
  int i;
  beenat+=urls;
}


mixed *query_status()
{
  return ({ sizeof(beenat||({})), sizeof(tovisit||({})) });
}


int id_for_url(string url)
{
  if (!(url_to_id[url]))
  {
    global_gensym++;
    url_to_id[url]=global_gensym;
    id_to_url[global_gensym]=url;
    return global_gensym;
  }
  else
    return url_to_id[url];
}

string url_for_id(int id)
{
  return id_to_url[id];
}


void add_reference_to(int from, string to)
{
  string rawmap;
  
  if (global_struc_databas)
  {
    if (rawmap=global_struc_databas["struct"][to])
    {
      string *urls=decode_value(rawmap);
      global_struc_databas["struct"][to]=(string)(encode_value(urls|({from})));
      // Union. Values present in both arrays will only be present once in the result.
    }
    else
      global_struc_databas["struct"][to]=(string)(encode_value(({from})));
  }
}

int *query_references(string to)
{
  string rawmap;
  if (rawmap=global_struc_databas["struct"][to])
    return decode_value(rawmap);
  else
    return ({});
}

void handle_references(string *urls, string *from)
{
  string url1,url2;
  int id;
  foreach(from, url1)
    foreach(urls, url2)
      add_reference_to(id_for_url(url1), (string)id_for_url(url2));
}


void new_possible_urls(string *urls, string *from)
{
  int i,j,k,flag;
  int numm;
  numm=sizeof(tovisit);
   
  // Never LOOP visiting the URL's. If the URL is on the BEEN at or TOVISIT list
  // don't schedule it twice!

  for(i=0; i<sizeof(urls); i++)
    if ((-1==member_array(urls[i], beenat)) && (-1==member_array(urls[i], tovisit)))
      if (profile_handler->control_url(global_profil, urls[i]))
	if  (sizeof(urls[i])>5)
	  if (robot_txt_check(urls[i]))
	    tovisit+=({urls[i]});
  
  handle_references(urls,from);
  
  if (QUERY_LOG_DETAIL(global_profil)<LOG_LEVEL_SHORT)
    plog(" Visited:"+sizeof(beenat)+". Ready: "+sizeof(tovisit)+
	 ". New: "+((sizeof(tovisit)-numm)?(sizeof(tovisit)-numm):"none")+".");
  
//  plog(" To visit: "+replace(sprintf("%O", tovisit),
//  ({ "<", ">" }), ({ "&lt;", "&gt;" }) )+" ");

}

int *query_error_list(int errorcode)
{
  // returns a list of all documents that got the errorcode 'errorcode'
  string rawmap;
  
  if (rawmap=global_struc_databas["struct"][(string)errorcode])
    return decode_value(rawmap);
  else
    return ({});
}

void handle_error_code(string *urls_visited, int sizeofdoc)
{
  // Note that sizeofdoc here is *not* the ACTUAL size of the document,
  // But rather a negative value. Each and every of these indicates
  // a different error code.

  
  string rawmap, url;
  
  string *newerrors=({});
  
  foreach(urls_visited, url)
    newerrors+=({ id_for_url(url) });

  // The negative error values won't interfer with
  // the positive page numbers stores in the same database.
  if (rawmap=global_struc_databas["struct"][(string)sizeofdoc])  
  {
    string *errors=decode_value(rawmap);
    global_struc_databas["struct"][(string)sizeofdoc]=(string)(encode_value(errors|newerrors));
    // Union. Values present in both arrays will only be present once in the result.
  }
  else
    global_struc_databas["struct"][(string)sizeofdoc]=(string)(encode_value(newerrors));

  used_error_list[sizeofdoc]=1;
  
}

void write_references_log(string profile)
{
  string url;
  int id, error, doc, fleff;
  int *docs;
  fleff=0;

  // Skapa en linklog ocks!?
  
  profile_handler->delete_consist_log(profile);
  
  if (profile_handler->struclogactive(profile)=="2")
    profile_handler->consist_log(profile,
				 "<h2>The logging of errors and warnings found "
				 "for the profile '"+profile+"' has been disabled.<br>"
				 "If you want to be able to view this log, you have "
				 "to go to the configuration of the profile and "
				 "re-enable it.</h2>");
  else
  {
    profile_handler->consist_log(profile,
				 "<h2>This is the Intraseek log of "
				 "errors and warnings found for the profile '"+
				 profile+"'.</h2>");
    if(sizeof(indices(used_error_list)))
    {
      // "<html><head><title>Intraseek Error log for profile: "+profile+".</title></head><body>"
      foreach(indices(used_error_list), int error)
      {
	docs=query_error_list(error);
	if (sizeof(docs))
	{
	  fleff=1;
	  switch (error) {
	   case  ERR_REDIRECT :
	    profile_handler->consist_log(profile,
					 "<p><p><hr noshade><h3><font color=orange>"
					 "Warning: The site had redirects on these "
					 "URLs:</h3></font><hr noshade>");
	    break;
	   case ERR_CON_FAIL : 
	    profile_handler->consist_log(profile,
					 "<p><p><hr noshade><h3><font color=red>"
					 "Error: Connection failed to these URLs:</h3>"
					 "</font><hr noshade>");
	    break;
	   case ERR_FORBIDDN :
	    profile_handler->consist_log(profile,
					 "<p><p><hr noshade><h3><font color=red>"
					 "Error: Forbidden to visit these URLs:</h3>"
					 "</font><hr noshade>");
	    break;
	   case ERR_NOTFOUND :
	    profile_handler->consist_log(profile,
					 "<p><p><hr noshade><h3><font color=red>"
					 "Error: These URL's couldn't be found:</h3>"
					 "</font><hr noshade>");
	    break;
	   case ERR__ERR_URL :
	    profile_handler->consist_log(profile,
					 "<p><p><hr noshade><h3><font color=orange>"
					 "Warning: These URL's were erronous:</h3></font>"
					 "<hr noshade>");
	    break;
	   case ERR__UNKNOWN :
	    profile_handler->consist_log(profile,
					 "<p><p><hr noshade><h3><font color=orange>"
					 "Warning: These URL's had unknown errors:</h3>"
					 "</font><hr noshade>");
	    break;
	   case ERR__ILLEGAL :
	    profile_handler->consist_log(profile,
					 "<p><h3><font color=orange>Warning: These "
					 "documents were of illegal type to index "
					 "(but not excluded in any way from visits). "
					 "(binary data, etc):</h3></font><hr noshade>");
	    break;
	  }
	  profile_handler->consist_log(profile, "<ol>");
	  foreach(docs, doc)
	  {
	    profile_handler->consist_log(profile,
					 "<li><a href=\""+url_for_id(doc)+"\">"+
					 url_for_id(doc)+"</a>:<ul>");
	    foreach(query_references((string)doc), id)
	      profile_handler->consist_log(profile,
					   "<li> Linked from: <a href=\""+
					   url_for_id(id)+"\">"+
					   url_for_id(id)+"</a>\n"); 
	    profile_handler->consist_log(profile, "</ul>");
	  }
	  profile_handler->consist_log(profile, "</ol>");
	}
      }
    }
    if (!fleff)
      profile_handler->consist_log(profile,
				   "<p>Congratulations, Intraseek found no errors!<p>");
  }
  profile_handler->consist_log(profile,
			       "<p><hr noshade><h4>This file was created by Intraseek at "+
			       ctime(time())+".</h4>");
  
  string fname=((profile_handler->working_at(global_profil))+"n."+global_profil+".structure.yabu");
  rm(fname+"/struct.chk");
  rm(fname+"/struct.inx");
  rm(fname);
  
}

void freeze()
{
  int tid;
  plog("Freeze! Storing all scheduler data structures ("+
       sizeof(beenat)+" pages visited, "+sizeof(tovisit)+" more known).");
  plog("Freeze! Storing to disc: "+(profile_handler->store_at(global_profil))
       +global_profil+".scheduler_freeze.is");
  rm((profile_handler->store_at(global_profil))+global_profil+".scheduler_freeze.is");
  write_file((profile_handler->store_at(global_profil))+global_profil+".scheduler_freeze.is",
	     sprintf("mapping site2=%O;\nstring *a=%O;\nstring *b=%O;\n",
		     robot_txt,tovisit,beenat));

  if (profile_handler->struclogactive(global_profil)=="1")
  {
    tid=time();
    plog("Syncing databases...\n");
    global_struc_databas->sync();
    plog("Done. Took "+profile_handler->describe_time(tid-time())+"...\n");
  }
    
}
	     
int restore_from_freeze()
{
  object o; 
  plog("UnFreezing! Restoring all scheduler data structures from disc!");
  if(file_stat((profile_handler->store_at(global_profil))+
	       global_profil+".scheduler_freeze.is"))
  {
    o=clone(compile_file((profile_handler->store_at(global_profil))+
			 global_profil+".scheduler_freeze.is"));
    robot_txt=o->site2;
    tovisit=o->a;
    beenat=o->b;
    destruct(o);
    plog("UnFreezing done OK.");
    return 1;
  }
  else
  {
    plog("No scheduler freeze file exist at "+
	 (profile_handler->store_at(global_profil))+
	 global_profil+".scheduler_freeze.is");
    return 0;
  }
  
  if (profile_handler->struclogactive(global_profil)=="1")
  {
    plog("Reopening database for storage of Site structure information.\n");
    string fname=((profile_handler->working_at(global_profil))+
		  "n."+global_profil+".structure.yabu");
    plog("Using the file: "+fname+"\n");
    global_struc_databas=Yabu.lookup(fname, "w");
  }
  else
    global_struc_databas=0;  // No site structure loggin will be used for this session!
}


void create(string profil, object profile_pointer,
	    int restore, string new_robot_home)
{
  string *temp;
  string tem;
  int i;
  profile_handler=profile_pointer;
  global_profil=profil;
  robot_home=new_robot_home;
  
  robot_txt=([]);  // Later on should this one be restored from saved file!

  // We need an unique fetcher to fetch /robot.txt
  special_fetcher=Crawler.fetcher();
  special_fetcher->set_robot_home(robot_home);
  random_seed(17);
  tovisit=profile_handler->start_pages(global_profil);
  if (sizeof(tovisit) && tovisit[0]=="/")
    tovisit=profile_handler->fall_back_world(global_profil);  // Ipp uses this!
  
  beenat=({});

  if (!restore)
  {
    if (profile_handler->struclogactive(global_profil)=="1")
    {
      crawlerlog("Creating database for storage of Site structure information.\n");
      string fname=((profile_handler->working_at(global_profil))+
		    "n."+global_profil+".structure.yabu");
      rm(fname+"/struct.chk");
      rm(fname+"/struct.inx");
      rm(fname); 
      crawlerlog("Will be stored in the file: "+fname+"\n");
      global_struc_databas=Yabu.lookup(fname, "w");

      if (!global_struc_databas)
	crawlerlog("Failed to open database.");
    }
    else
      global_struc_databas=0;  // No site structure loggin will be used for this session!
  }
}

int stop_func()
{
   // Zap the temporary GDBM database!?
}

