//
// This source-code is a part of the Roxen IntraSeek search engine package.
//  1996-1997 Informationsvvarna AB
// Email: info@infovav.se 
//
//
// This is the module that handles the Roxen IntraSeek engine
// It should be added as a normal roxen module!
//

#include <module.h>
#include <roxen.h>
//#include <simulate.h>
#include "languages.h"

inherit "module";
inherit "config_interface";

import spider;
import Stdio;

#define MAXDISP 15

#define DEF_LANG "en"
#define CR "\n"
#define FROM ({ "<", ">", ")and not", "and not(", " and not ", ")xor", ")not",")or", ")and", "xor(", "not(","or(", "and("," xor ", " not "," or ", " and " })
#define TO   ({ "&lt;", "&gt;", " - ", " - ", " - ", " ^ ", " - ", " | ", " + "," ^ ", " - ", " | ", " + "," ^ ", " - ", " | ", " + " })
#define SYMBOLS ({ "+", "-", ")", "(", "^", "|" })

#define FEL(x) "Error mounting Yabu database stored at <tt>"+x+"</tt>. The database is either corrupt or you don't have permission to open it.<br>Corrupt databases can appear if you run out of discspace when the database was created."

#define MAX_STACK 10
#define MAX_CACHE 75

//#define SIGNOFF "&nbsp;&nbsp;&nbsp;<font size=-2>&lt;<a href=http://www.roxen.com/>Roxen Intraseek</a>&gt;</font>"
#define SIGNOFF ""
#define DEFAULT_CHARSET "iso-8859-1"
#define INTERVAL_CHECK1 2
#define INTERVAL_CHECK2 2
#define THIS this_object()

#define STAND_ERR(x) "The profile handler (a vital part of IntraSeek) hasn't loaded correctly, and therefore it is not possible to use this service right now. The profile loader failed to load from: <tt>"+x+"</tt>. Most likely, the path is wrong. If not, Roxen debug log has to be checked for errors. To get rid of this message, the IntraSeek module must be reloaded."

#define ERR_NO_DB 1
#define ERR_CORRUPT 2

string cvs_version="$Id: intraseek_engine.pike,v 1.92 1999/08/22 01:31:13 js Exp $";

mapping (string:int)_interval_counter1=([]);
mapping (string:int)_interval_counter2=([]);

int global_not_display_flag;

int active=0;

string layout, admin, engine_home, profiles_home, file_prefix, default_database_id;
string maintainer, global_current_base, raw_data, stop_list_file_name, global_problem;

string  *active_databases=({});
string *global_languages=({});
mixed *global_remember_callouts=({});

int engine_log_on, loaded, global_dir_exist;

object profile_handler;

mapping database_pagebase=([]);
mapping database_index=([]);
mapping database_cut=([]);
mapping database_full=([]);

mapping (string:int) errors=([]);

mapping stoplists=([]);
mapping realnames=([]);
mapping obsolete_pages=([]);
//mapping global_new_database=([]);

mapping wordmap_cache=([]);
mapping cut_globmatch_cache=([]);
mapping cut_disk_cache=([]);

int hit_wordmap_cache=0;
int hit_cut_globmatch_cache=0;
int hit_cut_disk_cache=0;

int bom_wordmap_cache=0;
int bom_cut_globmatch_cache=0;
int bom_cut_disk_cache=0;


int rm(string fn)
{
  int code=predef::rm(fn);
//  werror("-- rm(\""+fn+"\") =>"+code+"\n");
  return code;
}


void engine_log_dump(string dumptext)
{
  if(engine_log_on)
    write_file(engine_home+"logs/engine.log", ""+ctime(time())[0..19]+" --- "+dumptext+"\n");
}


int move_file(string from, string to)
{
  //
  // Moves 'from' to 'to'
  // Fall-back to copy if mv fails.
  // Returns result of operation.
  //
  
  if (mv(from,to))
    return 1;
  engine_log_dump("Couldn't use mv, using copy from "+from+" to "+to+".");
  object in = File();
  object out = File();
  string tmp;
  if (!in->open(from, "r"))
    return 0;
  if (!out->open(to, "wtc"))
    return 0;
  do {
    tmp = in->read(32768);
    if (!tmp)
      return 1;
    if (-1==(out->write(tmp)))
    {
      engine_log_dump("Failed to copy. Write protected dir or disk full?");
      return 0;
    }
  } while (tmp != "");
  in->close();
  out->close();
  engine_log_dump("Finished copying. Removing "+from);
  return rm(from);
}


int move_dir(string from, string to)
{
  mkdirhier(to+"/");
  array entries=get_dir(from);
  if(!entries)
    return 0;
  else
    entries-=({"..","."});
  foreach(entries, string fn)
  {
    if(file_stat(from+"/"+fn)[1]==-2)
    {
      if(!move_dir(from+"/"+fn,to+"/"+fn))
	return 0;
    }
    else
    {
      if(!move_file(from+"/"+fn,to+"/"+fn))
	return 0;
    }
  }
  rm(from);
  return 1;
}


array register_module()
{
  string s="";
  
  if(loaded) {
    string q = query_location();
    if(sizeof(q) && q[0] == '/') q = q[1..];
    s = "<br>Go to the <a href='"+
      my_configuration()->query("MyWorldLocation")+"/intraseek/'>admininstration page</a>.<br>";
  }

  return ({
    (MODULE_PARSER|MODULE_LOCATION|MODULE_PROVIDER),
      "IntraSeek 1.0.18 Search Engine",
      "Roxen IntraSeek 1.0.18 Search Engine.<br>"+s+"<p>\n\n\n(C)1996-1998 Idonex AB.<br>"
      "Intraseek is a search engine designed and optimized for operating in "
      "the Roxen environment. It is written in Pike - the same fast, "
      "flexible and powerful programming language that Roxen Challenger is "
      "based upon.",
      0,
      1,
      });
}

function valid_url;

void create(object conf)
{
  defvar("provider_backdoor", 0, "Provider backdoor", TYPE_INT|VAR_EXPERT);

  defvar("cusername", roxen->query("ConfigurationUser"),
	 "Admin user name", TYPE_STRING,
	 "This user name grants full access to the configuration "
	 "part of Roxen IntraSeek.");
  
  defvar("cpassword", roxen->query("ConfigurationPassword"),
	 "Admin password", TYPE_PASSWORD,
	 "This password grants full access to the configuration "
	 "part of Roxen IntraSeek.");

  defvar("mail", "yourname@yourcompany.com", "Administrator email", TYPE_STRING,
	 "Will be used by the crawler when presenting itself to the webserver. Could be left empty.");

  string mwd=combine_path(combine_path(getcwd(), backtrace()[-1][0]), "..");
  catch { mwd=combine_path(combine_path(mwd, ".."),
#ifdef __NT__
			   mwd
#else
			   readlink(mwd)
#endif			   
			   ); };
  mwd=combine_path(mwd, "../");
  
  defvar("enginepath", mwd, "Engine home location", TYPE_DIR,
	 "This is the path to where intraseek is stored in your filesystem.");
  defvar("profilespath", mwd, "Profile config location", TYPE_DIR,
	 "This is where the profile configurations should be saved. If you want to share the "
	 "same intraseek code between several virtual servers, each one has to have a different"
	 " profilespath, but they can still have the same enginepath.");
  
  defvar("elog_on", 1, "Engine log activated", TYPE_FLAG, "");
  defvar("mountpoint", "/intraseek/", "Mountpoint for config", TYPE_LOCATION,
	 "This is were you will find the Intraseek web-based configuration "
	 "interface on your webserver.");
  
}

string fixpath(string in)
{
  while(in!="" && (in[strlen(in)-1..strlen(in)-1]==" "))
    in=in[0..(strlen(in)-2)];
  if (in[strlen(in)-1..strlen(in)-1]!="/")
    in=in+"/";
  return in;
}

void load_profile_handler()
{
  program profile;
    
  if(!file_stat(engine_home+"Crawler.pmod/profile.pike"))
    global_problem=STAND_ERR(engine_home+"Crawler.pmod/profile.pike");
  else
  {
    global_problem=0;
    
    if (profile_handler)
      destruct(profile_handler);
    
    catch(profile=compile_file(engine_home+"Crawler.pmod/profile.pike"));
    if(!profile) {
      global_problem=STAND_ERR(engine_home+"Crawler.pmod/profile.pike");
      return;
    }
    profile_handler=clone(profile);
    profile_handler->set_robot_home(engine_home);
    profile_handler->set_all_robots_from(admin);
    profile_handler->load_and_parse_profiles(profiles_home, "profiles.txt");
  }
}

string query_location()
{
  return query("mountpoint");
}


void remove_file(string from)
{
  rm(from);
}

void remove_dir(string from)
{
  array entries=get_dir(from)-({"..","."});
  foreach(entries, string fn)
    if(file_stat(from+"/"+fn)[1]==-2)
      remove_dir(from+"/"+fn);
    else
      rm(from+"/"+fn);
  rm(from);
}

int new_database_exist(string database_id)
{
//  werror("new_database_exist('%s')\n",database_id);
  string database_stored_at;
  string database_working_at;
  string from, to;
  
  database_working_at=profile_handler->working_at(database_id);
  database_stored_at=profile_handler->store_at(database_id);
  //  werror(database_stored_at+database_id+".new.flag\n");
  if(file_stat(database_stored_at+database_id+".new.flag"))
  {
    engine_log_dump("Found new database ("+database_id+")!");
    from=profile_handler->working_at(database_id)+"n."+database_id+".yabu";
    to=profile_handler->store_at(database_id)+"a."+database_id+".yabu";
    if(from!=to)
    {
      engine_log_dump("Moving '"+database_id+"' databases from temporary storage at "+from+" to "+to+".");
      if(!move_dir(from, to))
	engine_log_dump("FAILED moving "+from+" to "+to+".");
    } else
      engine_log_dump("Database "+from+" now exist at "+from+".\n");
        
    engine_log_dump("Removing flag from "+database_stored_at+database_id+".new.flag");
    remove_file(database_stored_at+database_id+".new.flag");

    return 1;
  }
  return 0;
}



int open_database(string database_id)
{
  // returns either an error code or a 0 if everything went OK.

  string database_stored_at=profile_handler->store_at(database_id);
  string fname;
  
  realnames[database_id]=profile_handler->realname(database_id); // Reset realnames!
  
  if(!file_stat(""+database_stored_at+"a."+database_id+".yabu"))
  {
    engine_log_dump("<p>   Trying to use \""+database_id+"\" databases, which could not be found at "+database_stored_at+"a."+database_id+".yabu\n");
    return ERR_NO_DB;  // No database found. Maybe a new one is being built?
  }
  
  engine_log_dump("Mounting: '"+database_id+"' from:");
  
//   if(database_pagebase[database_id])
//     database_pagebase[database_id]->close(); ;  // If flipping, it was open and has to be closed before re-mounted!
//   if(database_index[database_id])
//     database_index[database_id]->close();  // If flipping, it was open and has to be closed before re-mounted!
//   if(database_cut[database_id])
//     database_cut[database_id]->close();  // If flipping, it was open and has to be closed before re-mounted!
//   if(database_full[database_id])
//     database_full[database_id]->close();  // If flipping, it was open and has to be closed before re-mounted!
  
  fname=(string)database_stored_at+"a."+database_id+".yabu";
  object db=Yabu.lookup(fname, "r");
  catch(database_pagebase[database_id]=db["pages"]);
  if (!(database_pagebase[database_id])) { engine_log_dump(FEL(fname)); return ERR_CORRUPT; }  // This is fatal!
  catch(database_index[database_id]=db["index"]);
  if (!(database_index[database_id])) { engine_log_dump(FEL(fname)); return ERR_CORRUPT; } // This is also fatal.
  catch(database_cut[database_id]=db["cut"]);
  if (!(database_cut[database_id])) engine_log_dump("          Cutoff search disabled.");
  catch(database_full[database_id]=db["full"]);
  if (!(database_full[database_id])) engine_log_dump("          Full search disabled.");
    
  stoplists[database_id]=profile_handler->stoplist(database_id);
    
  engine_log_dump("           Read stoplist, which had "+sizeof(stoplists[database_id])+" entries.");
  engine_log_dump("           Mounted ["+database_id+"] "+realnames[database_id]);
    
  wordmap_cache[database_id]=([]);  // Empty Cache!
  cut_globmatch_cache[database_id]=([]);  // Empty Cache!
  cut_disk_cache[database_id]=([]);  // Empty Cache!

  return 0;  // Everything OK  (0)
}

int str2int(string s)
{
  int k;
  if (s && s!="")
    sscanf(s, "%d", k);
  return k;
}

string format_page(string *parts,  mapping arguments)
{
  string lang, textcolor, result;
  int noneflag=str2int((string)parts[8]);

  if ((!noneflag) && (!parts[1])) return "";
  
  if(!(textcolor=(arguments->textcolor))) textcolor="#800040";
 
  if (!(lang=(arguments->lang)))
    lang=DEF_LANG;
  else
    if (search(global_languages, lang)==-1)
      return "Intraseek, Language '"+lang+"' not supported. Use the 'lang=' tag and one of these: "+(global_languages*",")+"<br>\n";
  
  if (noneflag)
    result=replace(parts[6], "$$foundnone", languages[lang][6]);
  else
  {
    result=parts[0];
    result+=replace(languages[lang][3], ({"%1", "%2"  }),
		    ({ (replace(parts[1], "$$delim",
				languages[lang][12])), parts[2] }));
    result+=": "+parts[3]+"<br>";
    if (parts[4]!="") result+=languages[lang][4]+":   "+parts[4]+"<br>          ";   // Ifnot too many
    if (parts[5]!="") result+=languages[lang][5]+":   "+parts[5]; // Ignored , if any...
    result+=replace(parts[6],
		    ({ "$$score", "$$matches", "$$size", "$$textcolor", }),
		    ({ languages[lang][7], languages[lang][8], languages[lang][9], textcolor }));
    result+=replace(parts[7], ({"$$prev", "$$next" }), ({ languages[lang][10], languages[lang][11] }));
  }
#ifdef HANDELBANKEN  
  result+="</form>";
#endif
  return result;
}


string *alternative_words(string term)
{
  //
  // Takes a glob, for example "gnu*"
  // and returns a list of words that matches it, for example
  // ({ "gnus", "gnuare", "gnupp" })
  //
  // Alternatives are stored in a database.
  //
  
  string rawmap;
  string *all;
  string cutterm;
  if (sizeof(term)>2)
  {
    if(!cut_globmatch_cache[global_current_base][term])
    {
      bom_cut_globmatch_cache+=1;
      cutterm=term[0..2];
      if (!(cut_disk_cache[global_current_base][cutterm]))
      {
	bom_cut_disk_cache+=1;
	rawmap=database_cut[global_current_base][cutterm];
	if (rawmap)
	  all=decode_value(rawmap);
	else
	  all=({});
	cut_disk_cache[global_current_base][cutterm]=all;
      }
      else
      {
	hit_cut_disk_cache+=1;
	all=(cut_disk_cache[global_current_base][cutterm]);
      }
      cut_globmatch_cache[global_current_base][term]=glob(term, all);
    }
    else
      hit_cut_globmatch_cache+=1;
    return cut_globmatch_cache[global_current_base][term];
  }
  return ({});
}


mapping wordmap_for_term(string term)
{
  string rawmap;
  mapping wordmap;
  
  if(!wordmap_cache[global_current_base][term])  // The cacheinfo may also be set by other functions!
  {
    bom_wordmap_cache+=1;
    rawmap=database_index[global_current_base][term];
    if (rawmap)
      wordmap_cache[global_current_base][term]=(wordmap=decode_value(rawmap));
    else
    {
      hit_wordmap_cache+=1;
      wordmap_cache[global_current_base][term]=0;
    }
  }
  return (wordmap_cache[global_current_base][term]);
}



mixed parse_tree(string *xq)
{
  //
  // Build parse tree
  //
  
  array(mixed) stack_ar=allocate(MAX_STACK);
  int level=0;
   
  level=0;
  stack_ar[level]=({ });
  foreach(xq, string term)
  {
    switch (term) {
     case "(" :
      level++;
      if(level>MAX_STACK)
	level=MAX_STACK;
      stack_ar[level]=({ });
      break;
     case ")" :
      level--;
      if (level>-1)
      {
	stack_ar[level]+=({ stack_ar[level+1] });
	stack_ar[level+1]=0;
      }
      break;
     default :
      if(!stack_ar[level])
	stack_ar[level]=({});
      stack_ar[level]+=({ term });
      break;
    }
  }
  return stack_ar[0];
}


string *evaluate_tree(mixed indata)
{
  int i;
  mixed first;
  mixed result=({});
  string last_operator="|";
  mixed current_pages;
  string *all_words;
  string megaraw, ppaaggee, original;
  int position;
  mapping buildmap;
  
  
  for(i=0; i<sizeof(indata); i++)
  {
    first=indata[i];
    if(arrayp(first))
      current_pages=evaluate_tree(first);
    else
    {
      if((first=="|") || (first=="+") || (first=="^") || (first=="-"))
      {
	last_operator=first;
      } else  {
	
	if (search(first, "_$_")!=-1 && database_full[global_current_base])	{
	  
	  //
	  // Handle quoted expressions.
	  // First, we do a search for all included terms with the AND inbetween!
	  //  (Ignoring stopwords here)
	  // Note that quoted searches are identified by _$_ and that spaces are
	  // replaced by this character!
	  //

	  if (!wordmap_cache[global_current_base][first])
	  {
	    bom_wordmap_cache+=1;
	    original=(first/"_$_")*" ";
	    //
	    // We insert AND's inbetween all words and do a new query....
	    //
	    string *valid_pages=evaluate_tree(parse_tree((((first/"_$_")*" + ")/" ")-({})));
	    //
	    // Then we use those pages, one by one
	    //
	    
	    wordmap_cache[global_current_base][first]=([]);
	    buildmap=([]);
	    
	    int cutoff=profile_handler->cutoff(global_current_base);

	    foreach(valid_pages, ppaaggee)
	    {
	      string s = sprintf("%d", ppaaggee);  // ugly casting, I know.
	      megaraw=database_full[global_current_base][s];
	      object gz = Gz;
	      if(sizeof(indices(gz)))
		catch (megaraw=gz->inflate()->inflate( megaraw ));
	      position=0;
	      //
	      // And now we use pike string search to search through the whole text.
	      // Position always increasing. naive!
	      //
	      while (-1!=(position=(search(megaraw, original, position))))
	      {
		if (!buildmap[s])
		  buildmap[(int)ppaaggee]=1;    // Initial
		else
		  buildmap[(int)ppaaggee]+=1;   // Count occurances
		position+=sizeof(original)-1;   // Increase position pointer
	      }

	      if (sizeof(buildmap) > cutoff)
		break;
		  
	      //
	      // Finally we build a  simulated wordmap_for_term for this very term!
	      // Not that the term stored in wordmap cache here, is the one with
	      // spaces replaced by the magic delimeter _$_
	      //

	    }
	    wordmap_cache[global_current_base][first]=buildmap;
	  }
	  else
	    hit_wordmap_cache+=1;
	  if (wordmap_cache[global_current_base][first])
	    current_pages= indices(wordmap_cache[global_current_base][first]);
	  else
	    current_pages=({});
	}
	else  // Yes. Its not possible to combine quoted strings and globs.
	{
	  if (search(first, "?")!=-1 || (search(first, "*")!=-1))
	  {
	    //
	    // Found a glob.
	    //
	    all_words=alternative_words(first)||({});  // Finds words that matches the glob
	  }
	  else all_words=({first});
	  current_pages=({ });
	  foreach(all_words, string xfirst)
	  {
	    //
	    // Checking each of the alternative terms
	    //
	    if(wordmap_for_term(xfirst))
	      current_pages+=indices(wordmap_for_term(xfirst));  // Slr upp i Yabu hr!
	  }
	}
      }
    }
    
    if(!((first=="|") || (first=="^") || (first=="+") || (first=="-")))
    {
      switch (last_operator) {
       case "+" :  // AND (INTERSECTION)
	result=result&current_pages;
	break;
       case "-" :  // NOT
	result=result-current_pages;
	break;
       case "^" :  // NOT
	result=result^current_pages;
	break;
       default :  // OR
	result=result|current_pages;
	break;
      }
    }
    
  }
  return result;
}


string *query_database(string inbase, string query, string str_start,
		       string query_base, string charset,object|void req_id,
		       string target)
{
  string *xq2, *t1, *t2, *ignored;
  int i,j,k,l, startt, tidus, tidus2, tidus3, tidus4, tidus5, tidus6;
  float max, maxtest;
  string term, temp, ootemp, base, rawmap;
  int *file_sta;
  int kvotat;
  string origquery;
  mapping cell, wordmap,page_points, page_count;
  array(string) xq=({});
  array(string) valid_pages=({});
  string part0,part1,part2,part3,part4,part5,part6,part7;
  int noneflag=0;
  object o=File();

  startt=0;
  
  if(!str_start)
    str_start="";
    
  base=(string)inbase;
#ifdef HANDELSBANKEN  
  if(req_id->variables->lastquery&&req_id->variables->andsearch=="yes")
    query="("+req_id->variables->lastquery+") AND "+query;
#endif  

  if(target)
    target = "target=\""+target+"\" ";
  else
    target = "";

  if(-1==search(active_databases, base))
    base=default_database_id;  // just incase some luser modifies the URL to something dumb. Fallback!
  
  if(-1==search(active_databases, base)) // Did the fallback work? If not, this is either an unactive database or a nonexistant one.
  {
    part1=part2=part3=part4=part5="";
    part6="<br><font size=+1><b>$$foundnone</b></font><p>No such active profile: "+base+".<p>";
    noneflag=1;
  }
  else
  {
  
    global_current_base=base;  // Set the the global variable.
  
    i=sscanf(str_start,"%d",startt);

    if ((new_database_exist(base)) || (errors[base]))
    {
      // If the database had an error, or a new database exist.
      // Try to remount the database!
      engine_log_dump("New database for "+base+". Remounting");
      errors[base]=open_database(base);  // Remount this database!
    }

    if (errors[base]) // If errors are still reported, return a disclaimer or something
    {
      part1=part2=part3=part4=part5="";
      part6="<br><font size=+1><b>$$foundnone</b></font><p>(No database found. Either a new one is being built right now, or no valid database has been built yet.)<p>";
      noneflag=1;
    }
    else
    {
    
      if ((query) && (query!="0")) 
      {
	if(profile_handler->query_log_active(base) != "2")
	{
	  o->open(engine_home+"logs/"+base+".query.log", "wac"); // Write, append, create if not exist.
	  o->write(((string)http_date(time()))[0..25]+" - "+query+CR);  // Append loginformation to log file.
	  if ((_interval_counter1[base]++)>INTERVAL_CHECK1) // Dont check size on EVERY question, that is a waste of cpu....
	  {
	    _interval_counter1[base]=0;  // Reset the interval counter to zero.
	    if (file_sta=o->stat())    // Ask file for its file-stat!
	      if (file_sta[1]>((int)(profile_handler->query_log_max_size(base))))  // Check size of log file.
	      { 	    // We have exceeded maximum size of query log.
		o->close();
		engine_log_dump("Query log size exceeded. Clearing log. Making backup to .old file.");
		string from=engine_home+"logs/"+base+".query.log";
		string to=engine_home+"logs/"+base+".query.log.old";
		if(!move_file(from,to))  // Make a backup!
		  engine_log_dump("Failed making backup of query log from "+from+" to "+to+".");
	      } 
	  }
	  o->close();
	}

	if(wordmap_cache[inbase] && (sizeof(wordmap_cache[inbase]))>MAX_CACHE)
	  wordmap_cache[inbase]=([]);  // Empty Cache!

	if(cut_globmatch_cache[inbase] && (sizeof(cut_globmatch_cache[inbase]))>MAX_CACHE)
	  cut_globmatch_cache[inbase]=([]);  // Empty Cache!
    
	if(cut_disk_cache[inbase] && (sizeof(cut_disk_cache[inbase]))>MAX_CACHE)
	  cut_disk_cache[inbase]=([]);  // Empty Cache!

	cell=([]);
	page_points=([]);
	page_count=([]);
	ignored=({});
	max=0.0;
	j=k=0;

    
	query=lower_case(query); // First we standard lowercase all normal chars...
    
	for (int po=0; po<sizeof(FROM); po++)
	  query=replace(query, FROM[po], TO[po]);
    
	// query=replace(query, FROM, TO);  // Much smarter.
    
	if ((charset) && (charset!="") && (low[charset]))
	  query=replace(query, low[charset][0], low[charset][1]);  // Make everything LOWERCASED!!! (take care of ISO!)
	else
	  query=replace(query, low[DEFAULT_CHARSET][0], low[DEFAULT_CHARSET][1]);  // If no tag, DEFAULT_CHARSET is assumed

	query=replace(query, ({ "(", ")" }), ({" ( ", " ) "}) );
	origquery=query;

	int quotstart, quotend;

	if (quotstart=search(query, "\"")!=-1)
	{
	  quotend=search(query, "\"", quotstart);
	  if (quotend==-1)
	  {
	    query=replace(query, "\"", " ");
	  }
	  else
	  {
	    string slutgilt="";
	    string *quotexplod=query/"\"";
	    for (int zx=0; zx<sizeof(quotexplod); zx++)
	    {
	      if (zx%2)
		slutgilt+=replace(quotexplod[zx], " ", "_$_");
	      else
		slutgilt+=quotexplod[zx];
	    }
	    query=slutgilt;
	  }
	}

	xq=query/" ";
	xq=xq-({""});
	ignored=stoplists[base]&xq;
	xq=xq-stoplists[base];
	valid_pages=evaluate_tree(parse_tree(xq));
	xq = xq - SYMBOLS;

    
	o->open(engine_home+"logs/"+base+".query.words.log", "wac"); // Write, append, create if not exist.
	o->write((string)(xq*"\n")+"\n");  // Append loginformation to log file.
	if ((_interval_counter2[base]++)>INTERVAL_CHECK2) // Dont check size on EVERY question, that is a waste of cpu....
	{
	  _interval_counter2[base]=0;  // Reset the interval counter to zero.
	  if (file_sta=o->stat())    // Ask file for its file-stat!
	  {
	    if (file_sta[1]>((int)(profile_handler->query_log_max_size(base))))  // Check size of log file.
	    { 	    // We have exceeded maximum size of query log.
	      o->close();
	      engine_log_dump("Query log size exceeded. Clearing log. Making backup to .old file.");
	      string from=engine_home+"logs/"+base+".query.words.log";
	      string to=engine_home+"logs/"+base+".query.words.log.old";
	      if(!move_file(from,to))  // Make a backup!
		engine_log_dump("Failed making backup of query log from "+from+" to "+to+".");
	    }
	  }
	}
	o->close();
    
   
	int cutoff=profile_handler->cutoff(base);
    
	foreach(xq, string orgterm)
	{
	  foreach(  (((search(orgterm, "?")!=-1 || (search(orgterm, "*")!=-1))?(alternative_words(orgterm)):({orgterm}))||({})), term)
	  {
	    j++;
	    wordmap=wordmap_for_term(term);
	    if (wordmap)
	    {
	      k++;
	      t1=indices(wordmap);
	      t2=values(wordmap);
	      cell[term]=sizeof(t1);
	      for (i=0; ((i<cutoff) && (i<sizeof(t1))); i++)
	      {
		page_count[t1[i]]+=1;
		maxtest=page_points[t1[i]]-=(1+log(1+(float)t2[i]));
		if (maxtest<max)
		  max=maxtest;
	      }
	    }
	  }
	}

	mixed gz = Gz;
#ifdef HANDELSBANKEN
	if(req_id->variables->sort=="by_date")
     	{
	  foreach(indices(page_points), int i)
	  {
	    temp = database_pagebase[base][(string)i];
	    if(sizeof(indices(gz)))catch(temp=gz->inflate()->inflate(temp));
	    array xq=(""+temp)/DELI;
	    int date=(int)xq[-1];
	    page_points[i]=(float)date;
	  }
	}
#endif	
	page_points=mkmapping(valid_pages, valid_pages)&page_points;
          // INTERSECTION (values taken from right argument)


	
	if(req_id->misc->filter_documents)
	{
	  mapping urls=([]);
	  foreach(indices(page_points), int i)
	  {
	    temp = database_pagebase[base][(string)i];
	    if(sizeof(indices(gz)))catch(temp=gz->inflate()->inflate(temp));
	    array xq=(""+temp)/DELI;
	    string url=xq[0];
	    if(!sscanf(url,"http://%*s:%*d/%s",url))
	      sscanf(url,"http://%*s/%s",url);
	    url="/"+url;
	    urls[i]=url;
	  }
	  page_points=req_id->misc->filter_documents(urls,req_id)&page_points;
	}
// 	werror(sprintf("Before: %O\n",page_points));
	if(req_id->misc->sb)
	  foreach(indices(page_points), int i)
	  {
	    temp = database_pagebase[base][(string)i];
	    if(sizeof(indices(gz)))catch(temp=gz->inflate()->inflate(temp));
	    //	    werror(sprintf("temp: %O\n",temp));
	    array xq=(""+temp)/DELI;
	    string url=xq[0];
	    //werror(sprintf("url: %O\n",url));
	    if(!sscanf(url,"http://%*s:%*d/%s",url))
	      sscanf(url,"http://%*s/%s",url);
	    //werror(sprintf("url: %O\n",url));
	    if(sizeof(url) && url[-1]=='/')
	      url=url[..sizeof(url)-2];
	    // Only show documents the user in req_id is allowed to see...
// 	    werror(url+"\n");
	    // FIXME:
	    int pp=req_id->misc->sb->ac_find_file_pp(url);
// 	    werror(sprintf("pp: %O\n",pp));
	    int perm=req_id->misc->sb->mac->access_try(pp,req_id);
// 	    werror(sprintf("perm: %O\n",perm));
	    if(perm==0)
	      m_delete(page_points,i);
	  }
// 	werror(sprintf("After: %O\n",page_points));
	t1=indices(page_points);
	t2=values(page_points);
	sort(t2,t1);  // sort. the second ... (ena lista, andra som bieffekt (=indexen))
#ifdef HANDELSBANKEN
	if(req_id->variables->sort=="by_date")
     	{
	  t1=reverse(t1);
	  t2=reverse(t2);
	}
#endif
      }

      int oneach=profile_handler->entries(base);

      if ((query) && (query!="0")  && (query!=""))
      {
	if(k)
	{
	  mapping hidden = ([]);
	  int hidden_before_this;
 	  int hidden_on_this;
	  mixed gz = Gz;
	  part0="<hr noshade>\n<font size=+2><input type=hidden name=lastquery value='"+html_encode_string(query)+"'>";
	  part1=(string)(1+startt-hidden_before_this)+"$$delim"+
	    (((startt+oneach)>sizeof(t1))?(sizeof(t1)-sizeof(hidden)):
	     ((startt+oneach-hidden_on_this)-hidden_before_this));
	  part2=(string)(sizeof(t1)-sizeof(hidden));
	  part3="<b>"+replace(origquery, " ", "")+"</b></font><tt>";
	  part4="";
	  l=0;
	
	  global_not_display_flag=0;
	
	  foreach(Array.uniq(xq), string orgterm)
	    foreach(((search(orgterm, "?")!=-1 || (search(orgterm, "*")!=-1)))?(alternative_words(orgterm)):({orgterm}), term)
	    {
	      if(cell[term]>0)
	      {
		l++;
		global_not_display_flag+=1;
		ootemp=term;
		kvotat=0;
		if(search(term, "_$_")!=-1)
		{
		  term=replace(term, "_$_", " ");
		  kvotat=1;
		}
#define MIN(X,Y) ((X)<(Y)?(X):(Y))
		part4+="<a "+target+"href=\""+http_encode_string(query_base)+
		  "?p="+(kvotat?"%22":"")+http_encode_string(term)+
		  (kvotat?"%22":"")+"&base="+http_encode_string(inbase)+"\">"+
		  term+"</a>:"+cell[ootemp]+ //MIN(,sizeof(t1)-sizeof(hidden))
		  "  ";
		if(l==5)      
		{
		  part4+="\n                     ";
		  l=0;
		}
	      }
	    }
    
	  if (global_not_display_flag>MAXDISP)
	    part4="";
	  part5="";
	  if(sizeof(ignored))
	    foreach(ignored, term)
	      part5+=""+term+"  ";
	
	  part6="</tt><font size=-2><hr noshade>";
	
	  int hidden_so_far = 0;
	  for (i=startt; ((i<startt+oneach) && (i<sizeof(t1))); i++)
	  {
	    if(hidden[i]) {
	      hidden_so_far++;
	      continue;
	    }
	    temp =  database_pagebase[base][""+t1[i]];
	  
	    if(sizeof(indices(gz)))
	      catch (temp=gz->inflate()->inflate( temp ));

	    xq=(""+temp)/DELI;
	    term=xq[0];
	    
	    part6+="<TABLE width=95% BORDER=0 CELLPADDING=1 CELLSPACING=0>\n"
	      "<TR>"
	      "<TD ROWSPAN=3 VALIGN=TOP ALIGN=LEFT>"
	      "<B>   "+(i+1-hidden_so_far)+")</B></TD>"
	      "<TD VALIGN=TOP ALIGN=LEFT><FONT FACE=\"Arial,Helvetica\"><A HREF=\""+term+"\">"+
	      html_encode_string(xq[3])+"</a>.</TD>"+
	      ((req_id->variables->sort=="by_date")?
	       (
		 "<TD VALIGN=TOP ALIGN=RIGHT><FONT SIZE=-1>&nbsp;Date: "+http_date((int)xq[-1])+", $$matches: "+page_count[(t1[i])]+"&nbsp;<BR>[$$size: "+sprintf("%5.1f", (((float)(xq[1]))))+" Kb]</FONT></TD>"):
	       (
		 "<TD VALIGN=TOP ALIGN=RIGHT><FONT SIZE=-1>&nbsp;$$score: "+(sprintf("%2.1f",((float)(100*(((float)(t2[i]))/(max))))))+"%, $$matches: "+page_count[(t1[i])]+"&nbsp;<BR>[$$size: "+sprintf("%5.1f", (((float)(xq[1]))))+" Kb]</FONT></TD>"))+
	      "</TR>"
	      "<TR>"
	      "<TD COLSPAN=2><FONT COLOR=\"$$textcolor\">"+html_encode_string(xq[2])+"</FONT></TD>"
	      "</TR>"
	      "<TR>"
#ifndef HANDELSBANKEN	      
	      "<TD VALIGN=TOP ALIGN=LEFT COLSPAN=2><FONT size=-1 FACE=\"Arial,Helvetica\" COLOR=\"#ff5050\"><A HREF=\""+term+"\">"+html_encode_string(term)+"</A></FONT></TD>"
#else
	      ""
#endif	      
	      "</tr>"
	      "<tr>"
	      "<TD COLSPAN=3><HR NOSHADE></TD>"
	      "</TR>"
	      "</TABLE>\n";
	  }
	} else {
	  if ((query) && (query!="0") && (query!=""))
	  {
	    part6="<br><font size=+1><b>$$foundnone</b></font><p>";
	    noneflag=1;
	  }
	}
	part6+="</font>";
	part7="";
	query = (query/" " - ({""})) * " ";
	if(startt>=oneach)
	  part7+="&nbsp;<a "+target+"href=\""+http_encode_string(query_base)+
	    "?p="+replace(http_encode_string(query), "+", "%2B")+
	    "&start="+(startt-oneach)+"&base="+http_encode_string(inbase)+"\">[$$prev]</a>&nbsp;";

	int pageline=profile_handler->quicklinks(base);

	if(sizeof(t1)>oneach)
	{
	  for(i=0; ((i<pageline) && ((i*oneach)<sizeof(t1)));  i++)
	  {
	    if((i*oneach)==startt)
	      part7+="&nbsp;"+(i+1)+"&nbsp;";
	    else
	      part7+="&nbsp;<a "+target+"href=\""+http_encode_string(query_base)+
		"?p="+replace(http_encode_string(query), "+", "%2B")+
		"&start="+(i*oneach)+"&base="+http_encode_string(inbase)+"\">"+(i+1)+"</a>&nbsp;";
	  }
	}
      
	if ((((startt+oneach)<sizeof(t1))) && ((startt+oneach)<(pageline*oneach)))
	  part7+="&nbsp;<a "+target+"href=\""+http_encode_string(query_base)+
	    "?p="+replace(http_encode_string(query), "+", "%2B")+
	    "&start="+(startt+oneach)+"&base="+http_encode_string(inbase)+"\">[$$next]</a>&nbsp;";
    
      }
      if (!t1)
      {
	part0=part1=part2=part3=part4=part5=part6=0;
	noneflag=0;
      }
      else
	if (sizeof(t1)==0)
	{
	  part6="<br><font size=+1><b>$$foundnone</b></font><p>";
	  noneflag=1;
	}
    }
  }  
  part7+=SIGNOFF;
  
  return ({ part0,part1,part2,part3,part4,part5,part6,part7,noneflag });
}
  
int build_active_databases_list()
{
  string *old_active_databases=active_databases;
  active_databases=({});
//  profile_handler->load_and_parse_profiles(profiles_home,"profiles.txt");
  foreach(profile_handler->active_profiles(), string id)
    if (id && (id!=""))
      active_databases+=({ id });
  return equal(old_active_databases, active_databases);
}

void db_start() 
{
  // Try to mount all active database!
  
  build_active_databases_list();
  if(sizeof(active_databases)==0)
    return; // result+CR+"Warning! No active databases. Nothing mounted.<br>\n";
  default_database_id=active_databases[0];  // Use this as fallback if naughty id is entered in the URL...
  foreach (active_databases, string database_id)
  {
    _interval_counter1[database_id]=0;
    _interval_counter2[database_id]=0;
    if(new_database_exist(database_id))
      engine_log_dump("New database for "+database_id+" found and swapped in. Now this is beeing used!");
    errors[database_id]=open_database(database_id);
  }
}

string status()
{
  string temp;
  array(string) log;
  string res="";
  res+="Everything is OK!\n<p>"
    "Cache hit ratio:<p><ol>"
    "<li>Cache for wordmap: "+hit_wordmap_cache+" hits of "+(hit_wordmap_cache+bom_wordmap_cache)+" queries.<br>"
    "<li>Cache for cutglob: "+hit_cut_globmatch_cache+" hits of "+( hit_cut_globmatch_cache+bom_cut_globmatch_cache)+" queries.<br>"
    "<li>Cache for cutdisk: "+hit_cut_disk_cache+" hits of "+(hit_cut_disk_cache+bom_cut_disk_cache)+" queries.<br>"
    "</ol>";
  return res; 
}



string expand_searchfield(string tag_name, mapping arguments, object req_id, object file, mapping defines)
{
  string *id;
  int i;
  string base, query, default_id, lang, part1, part2, part3, part4, ids, def;
  string query_base="";
  string result="";
  int defflag=0;

  // This tag sets the global variable 

  if (global_problem) return global_problem;
  
  if (arguments->action)
    query_base += " action=\""+arguments->action+"\"";
  else
    query_base += " action=\""+http_encode_string(req_id->not_query)+"\"";
  if (arguments->target)
    query_base += " target=\""+arguments->target+"\"";
  
  base=req_id->variables->base||"";
  query=req_id->variables->p||"";
  lang=(arguments->lang);
  ids=arguments->ids;
  def=arguments->default_id;
  part1="";
  part1+="<form"+query_base+">";
#ifdef HANDELSBANKEN
  part1+="<table bgcolor=#659ACF cellpadding=5><tr><td colspan=2>";
  part1+="<table><tr>";
  if(stringp(req_id->misc->filters))
    part1+="<td>Search within filter:</td>";
  part1+="<td>Search for:</td><td></td></tr><tr>";
  if(stringp(req_id->misc->filters))
    part1+="<td>"+req_id->misc->filters+"</td>";
  part1+="<td>";
#endif  
      
  if (!def && !ids)  // If no ids specified, fallback on active ids.
  {
    id=profile_handler->active_profiles();
    if (sizeof(id))
    {
      if (sizeof(id)>1)
	ids=id*",";   // If more than one, list them all
      else
	def=id[0];  // Or if only one, set the default_id to the only one!
    }
    else
      return "Intraseek error. No ID's specified, or configured.";   // No active profiles at all.
  }
  
  if(def)  // if a default one, make this hidden!
  {
    part2="<input type=hidden name=base value=\""+html_encode_string(def)+"\">";
    defflag=1;
  }
  else  // Otherwise, make a selection list out the available ones!
  {
    part2=" <select name=\"base\">";
    id=ids/",";
    for (i=0; (i<sizeof(id)); i++)
    {
      if (id[i]==base)
	part2+="<option value=\""+html_encode_string(id[i])+"\" selected>"+
	  html_encode_string(realnames[id[i]])+"</option>";
      else
	part2+="<option value=\""+html_encode_string(id[i])+"\" >"+
	  html_encode_string(realnames[id[i]])+"</option>";
    }
    part2+="</select>";
  }

#ifdef HANDELSBANKEN
  part3=" <input size=30 name=p value=\"";
  if ((query) && (query!="0") && (query!=""))
    part3+=html_encode_string(query);
  
  part3+="\"></td><td>&nbsp;<input type=submit value=\"";
  part4="\">&nbsp;<input type=hidden name=start value=0></td></tr></table>";
#else
  part3=" <input size=30 name=p value=\"";
  if ((query) && (query!="0") && (query!=""))
    part3+=html_encode_string(query);
  
  part3+="\">&nbsp;<input type=submit value=\"";
  part4="\">&nbsp;<input type=hidden name=start value=0>";
#endif  

#ifdef HANDELSBANKEN
  part4+="</td></tr>";
  if(req_id->variables->p)
    part4+=
      "<tr><td><input type=radio name=andsearch value=yes CHECKED> "
      "Search in result</td>"
      "<td><input type=radio name=andsearch value=no> "
      "New search</td></tr>";
  part4+=
    "<tr><td><input type=radio name=sort value=by_confidence CHECKED> "
    "Sort result by confidence</td>"
    "<td><input type=radio name=sort value=by_date> "
    "Sort by date</td></tr></table>";
#endif  


#ifdef HANDELSBANKEN
  if(!req_id->variables->p)
#endif    
    part4+="</form>";
  
  if ((!(lang=(arguments->lang))) || (search(global_languages, lang)==-1))
    lang=DEF_LANG;
  
  if (defflag) // If we have a default database, ie no select-button...  (Different sentences are formed)
#ifdef HANDELSBANKEN
    result=part1+part2+part3+languages[lang][0]+part4;
#else
    result=part1+part2+languages[lang][1]+part3+languages[lang][0]+part4;
#endif  
  else
    result=part1+(replace(languages[lang][2], "%0", part2))+part3+languages[lang][0]+part4;
  return result;
}

string expand_result(string tag_name, mapping arguments, object req_id,
		     object file, mapping defines)
{
  if (global_problem) return global_problem;
  return format_page(query_database((string)req_id->variables->base,
				    (string)req_id->variables->p,
				    (string)req_id->variables->start,
				    (arguments->action||req_id->not_query||""),
				    (arguments->charset),
				    req_id,
				    arguments->target),
		     arguments);
}

string expand_last_update(string tag_name, mapping arguments, object req_id, object file, mapping defines)
{
  // Not officially supported. Use at own risk
    if (global_problem) return global_problem;

  string base=(string)req_id->default_id;
  if(-1==search(active_databases, base))
    base=default_database_id;  // just incase some luser modifies the URL to something dumb. Fallback!
  if ((!errors[base]) && (-1!=search(active_databases, base))) // Does the base exist, and has been mounted OK?
    return ctime((int)(database_pagebase[base]["$last_update$"]));
  else
    return "<i>Unknown</i>";
}

string expand_page_counter(string tag_name, mapping arguments, object req_id, object file, mapping defines)
{
  // Not officially supported. Use at own risk
  if (global_problem) return global_problem;

  string base=(string)req_id->default_id;
  if(-1==search(active_databases, base))
    base=default_database_id;  // just incase some luser modifies the URL to something dumb. Fallback!
  if ((!errors[base]) && (-1!=search(active_databases, base))) // Does the base exist, and has been mounted OK?
    return database_pagebase[base]["$symb$"];
  else
    return "<i>Unknown</i>";
}
 

string expand_title(string tag_name, mapping arguments, object req_id, object file, mapping defines)
{
  if (global_problem) return global_problem;

  string lang=(arguments->lang);
  if ((!(lang=(arguments->lang))) || (search(global_languages, lang)==-1))
    lang=DEF_LANG;
  if(req_id->variables->p)
    return html_encode_string((string)languages[lang][13]+""+req_id->variables->p);
  else
    return html_encode_string(languages[lang][0]);
}

int is_this_the_engine_home_location(string where)
{
  if (file_stat(where+"/Crawler.pmod/profile.pike"))
    return 1;
  return 0;
}

string check_variable(string name, mixed value)
{
  // Check if `value' is O.K. to store in the variable `s'.  If so,
  // return 0, otherwise return a string, describing the error.
  string idt;

  switch (name) {
   case "enginepath" :
    if (!(is_this_the_engine_home_location(value)))
      return "This is not the IntraSeek home directory!";
    THIS->engine_reset(1);
    break;

   case "profilespath" :
    THIS->engine_reset(1);
    break;
    
   case "mail" :
    THIS->engine_reset(1);
    break;

   case "elog_on" :
    THIS->engine_reset(1);
    break;
    
   case "mountpoint" :
    my_world=0;
    break;
  }
  return 0;
}


mapping query_tag_callers()
{
  return (["intraseek_form":expand_searchfield,
	   "intraseek_title":expand_title,
	   "intraseek_results":expand_result,
	   "intraseek_updated":expand_last_update,
	   "intraseek_pages":expand_page_counter ]);
}


void autolaunch(string profilen)
{
  engine_log_dump("Launching  "+profilen+" Now! ... At "+ctime(time()));
  this_object()->do_launch(profilen, 0);
  call_out(this_object()->restart_callouts, 120);  // We can safely wait a while before restarting this callout
}

void restart_callouts()
{
  foreach(global_remember_callouts, mixed callid)
    remove_call_out(callid);
  global_remember_callouts=({});
  int xtime=time();
  int ntime;
  foreach((profile_handler->active_profiles()), string profilen)
  {
    if (((profile_handler->crawler_schedule_1(profilen)!=0)) && ((profile_handler->active(profilen))))
      if (-1!=(ntime=(profile_handler->next_schedule_start(profilen,xtime))))
      {
	global_remember_callouts+=({ call_out(autolaunch, (10+(ntime-xtime)), profilen) });
	engine_log_dump("Scheduling autolaunch for "+profilen+" at "+ctime(ntime));
      }
  }
}


void engine_reset(int level)
{
  // level=0=First time
  // level=1=Just refresh mounted databases and such!
  string temp;
  engine_home=combine_path((string)query("enginepath"), "");
  profiles_home=combine_path((string)query("profilespath"), "");
  admin=(string)query("mail");
  engine_log_on=(int)query("elog_on");
  
  if (level==0)  // First time! Always try to load profile handler
    load_profile_handler();

  if (level==0 || level==1)
  {
    if (profile_handler)
    {
      restart_callouts();
      db_start();  // Remount all databases.
      global_languages=indices(languages);
    } else
      load_profile_handler();       // Try to load the profile handler.... again.
  }
  
}

int module_dependencies(object configuration, array (string) modules)
{
  if(configuration) {
    foreach (modules, string module)
      configuration->enable_module(module);
    if(roxen->root)
      roxen->configuration_interface()->build_root(roxen->root);
  }
  return 1;
}

void stop()
{
  // Zap config interface when module is reloaded!
  string s=combine_path(engine_home||"", "modules/config_interface.pike");
  m_delete(master()->programs, s);
}

void start(int num, object configuration)  // number and config object sent as argument
{
  if(configuration)
    roxen->set_var("intraseek_set_valid_url/"+configuration->name, 
		   lambda(function q){ valid_url=q; });
  

  array err;
  module_dependencies(configuration, ({ "graphic_text#0",
					"obox#0",
					"configtablist#0"  }) );
  global_problem=0;
  loaded=1;

  roxen->set_var("intraseek_sitebuilder_hook", this_object());   // SiteBuilder hook
  
  if (err=catch {
    if (!active)
    {
      engine_reset(0);
      active=1;
    }		
  })
    engine_log_dump("Error in Intraseek module start: %s\n"+describe_backtrace(err));
}

class provider_globalwizard {

  //
  // THIS IS FOR IPP!
  //
  
  inherit "wizard";

  constant sched1=({ "Never!", "Weekly", "Biweekly", "First week every month" });
  constant sched2=({ "Daily", "Mondays", "Tuesdays", "Wednesdays", "Thursdays",
		     "Fridays", "Saturdays", "Sundays" });
  constant sched3=({ "00:00", "02:00", "04:00", "06:00", "08:00", "10:00",
		     "12:00", "14:00", "16:00", "18:00", "20:00", "22:00" });

  string name = "Intraseek global settings";

  string page_0(object id, mapping conf)
  {
    return
      "<h3>Automatic crawler launch</h3>\n"
      "<table cellpadding=5 cellspacing=0 border=0>"
      "<tr><td valign=top align=right>"
      "<br></td><td valign=top>"
      "<var name=schedule_1 type=select choices='"+sched1*","+
      "' default='"+sched1[conf->schedule_1]+"'>"
      ", <br>"
      "<var name=schedule_2 type=select choices='"+sched2*","+
      "' default='"+sched2[conf->schedule_2]+"'>"
      " at "
      "<var name=schedule_3 type=select choices='"+sched3*","+
      "' default='"+sched3[conf->schedule_3]+"'>"
      "</td></tr></table>\n";
  }

  void wizard_done(object id, mapping conf)
  {
    conf->schedule_1 = search(sched1, id->variables->schedule_1);
    conf->schedule_2 = search(sched2, id->variables->schedule_2);
    conf->schedule_3 = search(sched3, id->variables->schedule_3);
  }

}


void provider_init(object customer, mapping conf)
{

  //
  //
  // This is for IPP!
  //
  //
  
  array data = allocate(28);
  string workdir;

  data[0]=customer->get_field("name");
  mkdir(workdir = combine_path(customer->get_field("logdir"),
			       "intraseek.data"));
  data[1]=workdir+"/";
  data[3]=({"/*"});
  data[5]=({"/"});
  data[20]="2";
  data[21]=conf->schedule_1;
  data[22]=conf->schedule_2;
  data[23]=conf->schedule_3;

  set("provider_backdoor", 1);
  set("profilespath", data[1]);
  my_configuration()->save(1);

  engine_reset(1);

  profile_handler->create_profile(replace(customer->cid(), " ", "_"), data);
  profile_handler->save_profiles(data[1]);

  catch {
    do_launch(customer->cid(), 0, 0);
  };
}

// SiteBuilder plugin stuff

constant sb_plugin_name = "Intraseek";
constant sb_minimal_access = 2; // There's no read access only mode (yet)
mapping|string sb_plugin_handle(string sub, string pre, object id, int access)
{
  if(access<2)
    return 0;
  id->misc->content_editor_base=pre;
  mixed answer=this_object()->find_file(sub,id);

  if(answer->error)
    return answer;
  else
    return answer->data;
}

multiset query_provides()
{
  return (< "intraseek", "sitebuilder_plugin" >);
}
