// Roxen LogView Calendar Class
//
// Copyright (C) 1996, 1997, 1998 Idonex
//

#define GREEN "#338844"

import Array;

constant month_names = ({ "January", "February", "March",
			 "April", "May", "June",
			 "July", "August", "September",
			 "October", "November", "December" });

constant smonth_names = ({ "Jan", "Feb", "Mar", "Apr",
			  "May", "Jun", "Jul", "Aug",
			  "Sep", "Oct", "Nov", "Dec" });

constant day_names = ({ "Monday", "Tuesday", "Wednesday",
			"Thursday", "Friday", "Saturday", "Sunday" });

constant sday_names = ({ "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun" });


/*** Julian date functions ***/

int ymd2jd(int y, int m, int d)
{
  int a;

  if (m<3)
    {
      y--;
      m+=12;
    }
  a = y/100;
  return 1720995 + d + 2 - a + (a/4) +
    (36525*y)/100 + (306001*(m+1))/10000;
}

int yd2jd(int y, int d)
{
  int a;
  y--;
  a = y/100;
  return 1721426 + d - a + (a/4) + (36525*y)/100;
}

array(int) jd2ymd(int jd)
{
  array(int) dt = allocate(3);

  int alpha, a, b, c, d, e;

  alpha = (100*jd - 186721625)/3652425;
  a     = jd + 1 + alpha - alpha/4;
  b     = a + 1524;
  c     = (100*b - 12210)/36525;
  d     = (36525*c)/100;
  e     = (10000*b-10000*d)/306001;
  dt[2] = b - d - (306001*e)/10000;
  dt[1] = (e<14) ? e-1 : e-13;
  dt[0] = (dt[1]<3) ? c-4715 : c-4716;
  return dt;
}

int week2jd(int year, int week)
{
  if(week == -1) {
    year++;
    week = -week;
  }
  if(week < 0) {
    year--;
    week = -week;
  }
  int first_time = yd2jd(year, 0);
  int first_wday = first_time % 7;
  first_time -= first_wday;
  if(first_wday > 3)
    first_time += 7;
  return (week<0? first_time - 7 : first_time + 7 * (week - 1));
}

int leapyear(int year) {
  if ((year%400) == 0)
    return 1;
  if ((year%100) == 0)
    return 0;
  if ((year%4) == 0)
    return 1;
  return 0;
}

int weekday(int year, int month, int day)
{
  return ymd2jd(year, month, day)%7;
}

int week(int year, int month, int day) {
  int delta_day;
  int first_time = yd2jd(year, 0);
  int yday = ymd2jd(year, month, day)-first_time;
  int first_day = first_time%7;
  int first_next_day = yd2jd(year+1, 0)%7;
  
  if ((yday > (364+leapyear(year)-first_next_day)) &&
      (first_next_day <= 3))
    return 1;
  
  if (first_day <= 3) {
    // Week begins on Thursday or erlier, means week one here
    delta_day = first_day;
  } else {
    // Week begins on Friday or later, means same week as last year.
    if (yday + first_day <= 6) {
      // Return the last week previous year.
      return week(year-1, 12, 28);
    }
    delta_day = first_day-7;
  }
  // Calculate week. Begin with week 1.
  return (yday + delta_day)/7 + 1;
}

int lastdayofmonth(int year, int month)
{
  return (month==12? 31 : ymd2jd(year, month+1, 1)-ymd2jd(year, month, 1));
}

int lastweekofyear(int year)
{
  return week(year, 12, 28);
}

array(int) weeksofyear(int year)
{
  int w0 = week(year, 1, 1);
  array(int) arr = indices(allocate(lastweekofyear(year)+1))[1..];
  return (w0 == 1? arr : ({ w0 }) + arr);
}

array(int) daysofmonth(int year, int month)
{
  return indices(allocate(lastdayofmonth(year, month)+1))[1..];
}

array(int) weeksofmonth(int year, int month)
{
  int w;
  mapping(int:int) wks = ([]);
  foreach(daysofmonth(year, month), int day) {
    w = week(year, month, day);
    if(month == 1 && w > 50)
      w = -w;
    wks[w] = 1;
  }
  array(int) a = sort(indices(wks));
  if(month == 12 && a[0] == 1)
    a = a[1..] + ({ -1 });
  return a;
}

array(int) monthsofyear(int year)
{
  return indices(allocate(12+1))[1..];
}

/*** HTML Calendar ***/

array(int) has_data_week_list(int year, array(array(int)) has_data_dates)
{
  multiset(int) weeks = (< >);
  foreach(has_data_dates, array(int) ymd) {
    if(ymd[0] == year) {
      int w = week(@ymd);
      if(w>50 && ymd[1]==1)
	w = -w;
      else if (w==1 && ymd[1]==12)
	w = 4711;
      weeks[w] = 1;
    }
  }
  array(int) ww = sort(indices(weeks));
  if(sizeof(ww) && ww[-1]==4711)
    ww[-1]=-1;
  return ww;
}

array(array(int)) has_data_weekdays_list(int year, int wk,
					 array(array(int)) has_data_dates)
{
  array(array(int)) days = ({ });
  if(wk==-1) {
    wk = 1;
    year++;
  }
  if(wk<0) {
    wk = -wk;
    --year;
  }
  foreach(has_data_dates, array(int) ymd)
    if(ymd[0]>=year-1 && ymd[0]<=year+1) {
      int w = week(@ymd);
      int y = ymd[0];
      if(w == 1 && ymd[1] == 12)
	y++;
      if(w > 50 && ymd[1] == 1)
	y--;
      if(y == year && w == wk)
	days += ({ ymd });
    }
  return days;
}

array(int) has_data_monthdays_list(int year, int month,
				   array(array(int)) has_data_dates)
{
  multiset(int) days = (<>);
  foreach(has_data_dates, array(int) ymd)
    if((ymd[0] == year) && (ymd[1] == month))
      days[ymd[2]] = 1;
  return sort(indices(days));
}

array(int) has_data_month_list(int year, array(array(int)) has_data_dates)
{
  multiset(int) months = (<>);
  foreach(has_data_dates, array(int) ymd)
    if(ymd[0] == year)
      months[ymd[1]] = 1;
  return sort(indices(months));
}

array(int) has_data_year_list(array(array(int)) has_data_dates)
{
  multiset(int) years = (<>);
  foreach(has_data_dates, array(int) ymd)
    years[ymd[0]] = 1;
  return sort(indices(years));
}

/*** ***/

string html_monthsofyear(array(array(int)) has_data_dates,
			 int year, int this_month, string base)
{
  string s = "<table border=0><tr>";
  array(int) months = monthsofyear(year);
  array(int) datamonths = has_data_month_list(year, has_data_dates);
  for (int i = 0, j = 0; i < sizeof(months); i++) {
    int has_data = j<sizeof(datamonths) && datamonths[j] == months[i];
    s += "<td valign=top align=right><b>";
    if(has_data) {
      s += "<a href=\""+base+"month/"+year+"/"+datamonths[j]+"\">";
      if(this_month==months[i])
	s += "<font color="GREEN">"+smonth_names[i]+"</font>";
      else
	s += smonth_names[i];
      s += "</a>";
    } else {
      s += "<font color=#bbbbbb>"+smonth_names[i]+"</font>";
    }
    s += "</b></td>\n";
    if(i == 5)
      s += "</tr><tr>";
    if(has_data)
      j++;
  }
  return s+"</tr></table>";
}

string html_calendar(array(string) args, string base,
		     array(array(int)) has_data_dates)
{
  array(int) years = has_data_year_list(has_data_dates);

  if(!sizeof(years))
    return 0;

  mapping(string:mixed) ymwd = ([]);
  ymwd->type = (sizeof(args)?args[0]:0);
  array(int) t = map(args[1..], lambda(string s) {
    int d; sscanf(s, "%d", d); return d; } );

  string h;
  switch (ymwd->type) {
  case "year":
    ymwd->year = t[0];
    h = (string) ymwd->year;
    break;
  case "month":
    ymwd->year = t[0];
    ymwd->month = t[1];
    h = month_names[ymwd->month-1]+", "+ymwd->year;
    break;
  case "week":
    ymwd->year = t[0];
    ymwd->month = t[1];
    ymwd->week = t[2];
    h = "Week "+ymwd->week+", "+ymwd->year;
    break;
  case "day":
    ymwd->year = t[0];
    ymwd->month = t[1];
    ymwd->day = t[2];
    ymwd->week = week(ymwd->year, ymwd->month, ymwd->day);
    h = day_names[weekday(ymwd->year, ymwd->month, ymwd->day)]+
        ", "+month_names[ymwd->month-1]+" "+ymwd->day+", "+ymwd->year;
    break;
  default:
    return "<p><b>nada</b>";
  }

  string c = "<table border=0><tr align=left><th>Year</th><th>Month</th>";
  if(ymwd->month)
    c += "<th> Week &nbsp;</th>"
         "<th>Mo</th><th>Tu</th><th>We</th><th>Th</th><th>Fr</th>"
         "<th>Sa</th><th>Su</th>";
  c += "</tr><tr></tr><tr valign=top><td><table border=0><tr><td>"+
       map(years, lambda(int year, int dy, string base) {
        return "<b><a href='"+base+"year/"+year+"'>"+
	   (year==dy?"<font color="GREEN">"+year+"</font>":(string) year)+
	   "</a><br>\n";
       }, ymwd->year, base)*""+
       "</td></tr></table></td>"
       "<td>"+html_monthsofyear(has_data_dates,
				ymwd->year, ymwd->month, base)+"</td>";

  if(ymwd->month) {
    c += "<td align=center><table border=0>";
    array(int) data_wks = has_data_week_list(ymwd->year, has_data_dates);
    array(int) wks = weeksofmonth(ymwd->year, ymwd->month);
    mapping data_weekdays = ([]);
    foreach(wks, int wk) {
      int dwk = wk<0?-wk:wk;
      data_weekdays[wk] = has_data_weekdays_list(ymwd->year, wk,
						 has_data_dates);
      c += "<tr><td><b>";
      if(sizeof(data_wks & ({ wk }))) {
	c += "<a href='"+base+"week/"+ymwd->year+"/"+ymwd->month+"/"+wk+"/'>";
	if(wk == ymwd->week)
	  c += "<font color="GREEN">"+dwk+"</font>";
	else
	  c += ""+dwk;
	c += "</a>";
      } else
	c += "<font color=#bbbbbb>"+dwk+"</font>";
      c += "</b></td></tr>\n";
    }
    c += "</table></td>";

    for(int day = 0; day < 7; day++) {
      c += "<td><table border=0>";
      foreach(wks, int wk) {
	array(int) ymd = jd2ymd(week2jd(ymwd->year, wk) + day);
	int has_data = 0;
	foreach(data_weekdays[wk], array(int) tmp_ymd)
	  if(equal(ymd, tmp_ymd))
	    has_data = 1;
	c += "<tr align=right><td><b>";
	if(has_data) {
	  c += "<a href='"+base+"day/"+ymd[0]+"/"+ymd[1]+"/"+ymd[2]+"/'>";
	  if(ymd[0] == ymwd->year &&
	     ymd[1] == ymwd->month &&
	     ymd[2] == ymwd->day)
	    c += "<font color="GREEN">"+ymd[2]+"</font>";
	  else
	    c += ""+ymd[2];
	  c += "</a>";
	} else
	  c += "<font color=#bbbbbb>"+ymd[2]+"</font>";
	c += "</b></td></tr>\n";
      }
      c += "</table></td>";
    }
  }    
  c += "</tr></table>";
  
  c = "<obox right=310 title='"+
      replace("Period of time ("+h+")", " ", "&nbsp;")+"'>"+c+"</obox>";
  
  return c;
}
