
#define MAX_EVENTLOG_SIZE 42
#define GARB_TIME (2*24*60*60)

class event {
  static private object parent;
  static private string mode = "wtc";
  static private int event_number = 0;
  static private mapping events = ([]), lookup = ([]);

  static private mixed max(mixed ... args)
  {
    if(sizeof(args) == 1)
      return args[0];
    return predef::max(args[0], max(@args[1..]));
  }

  static private void garb()
  {
    int garbtime = time() - GARB_TIME;
    foreach(indices(events), int n) {
      mapping m = events[n];
      if(max(@m->time) < garbtime) {
	string key = m->type+"\0"+m->msg;
	m_delete(lookup, key);
	m_delete(events, n);
      }
    }    
  }

  static private string save_msg(string msg, string type)
  {
    msg = (upper_case(type)+": "+
	   capitalize(parent->language("eng", "date")(time(1)))+"\n"+msg);

    object event_file = Stdio.File();
    string filename = combine_path(parent->query("logpath"), "events");
    if(event_file->open(filename, mode)) {
      event_file->write((mode=="wtc"?"":"\n")+msg+"\n");
      event_file->close();
      mode = "wac";
    }
  }

  static private string describe(mixed m)
  {
    if(arrayp(m)) {
      array a = master()->describe_backtrace(m)/"\n";
      a[0] = "<b>"+a[0]+"</b>";
      return a*"\n";
    }
    return m;
  }
  
  static private string add_msg(array args, string type)
  {
    garb();

    string msg = "#unknown";
    if(sizeof(args) == 1)
      msg = describe(args[0]);
    else if(sizeof(args) > 1)
      msg = sprintf(@args);
    save_msg(msg, type);

    string key = type+"\0"+msg;
    if(zero_type(lookup[key])) {
      events[event_number] = ([ "type":type, "time":({time(1)}), "msg":msg ]);
    } else {
      int old_event_number = lookup[key];
      mapping old_event = events[old_event_number];
      m_delete(events, old_event_number);
      old_event->time = ({ time(1) }) + old_event->time;
      events[event_number] = old_event;
    }
    lookup[key] = event_number;
    event_number--;
    return msg;
  }  
  
  void notice(mixed ... args)
  {
    add_msg(args, "notice");
  }
  
  void warning(mixed ... args)
  {
    add_msg(args, "warning");
  }
  
  void error(mixed ... args)
  {
    werror("(LogView) "+add_msg(args, "error")+"\n");
  }
  
  void clear(string msg)
  {
    lookup = ([]);
    events = ([]);
    notice(msg||"Event Log cleared.");
  }

  string implode_nicely(array(string) x)
  {
    if (sizeof(x)<2)
      return x*"";
    else
      return x[0..sizeof(x)-2]*", "+" and "+x[-1];
  }

  string nice_timeline(array times)
  {
    array a = ({});
    string r = "", old_tp, tp, old_hr, hr;
    foreach(times, int t) {
      int h, m;
      string s;
      if(sscanf(s=parent->language("eng", "date")(t), "%s %d:%d",tp,h,m)==3)
	hr = sprintf("%02d:%02d", h, m);
      else
	sscanf(parent->language("eng", "date")(t), "%s %s", tp,hr);

      if(tp == old_tp) {
	if(old_hr != hr)
	  a += ({ hr });
      } else {
	if(sizeof(a))
	  r += capitalize(implode_nicely(a))+". ";
	a = ({ tp+" "+hr });
	old_tp = tp;
      }
      old_hr = hr;
    }
    r += capitalize(implode_nicely(a))+".";
    return r;
  }

  string html()
  {
    mapping imgs = ([ "notice":"err_1",
		      "warning":"err_2",
		      "error":"err_3" ]);
    mapping alts = ([ "notice":"Notice:",
		      "warning":"Warning:",
		      "error":"Error:" ]);

    string r = "<br><table border=0>\n";  // <dd>
    foreach(sort(indices(events)), int n) {
      mapping entry = events[n];
      int x = time();
      r += ("<tr><td valign=top><img border=0 alt="+alts[entry->type]+
	    " src=/internal-roxen-"+imgs[entry->type]+"></td><td>\n"+
#if constant(roxen)
	    roxen->configuration_interface()->describe_times(entry->time)+
#else
	    nice_timeline(entry->time)+
#endif
	    "<br>\n"+replace(entry->msg, "\n", "<br>")+"<p></td></tr>\n");
    }
    return r + "</table>\n";
  }
  
  void create(object parent_in)
  {
    parent = parent_in;
  }
}
