/*
** The table class
*/

#define TABLE_ERR(msg) throw(({ "(Table) "+msg+"\n", backtrace() }))

class table {
  static private mixed metadata;
  static private mapping fieldmap;
  static private array fields, table;

  static private void initialize_fields()
  {
    array a = indices(allocate(sizeof(fields)));
    fieldmap = mkmapping(fields + a, a+a);
  }
  
  static private array|int remap(array|int cs, int|void forgive)
  {
    array v = ({});
    int ap = arrayp(cs);
    if(!ap) cs = ({ cs });
    foreach(cs, string f)
      if(zero_type(fieldmap[f])) {
	if(!forgive)
	  TABLE_ERR("Unknown field '"+f+"'");
      } else
	v += ({ fieldmap[f] });
    return ap?v:v[0];
  }

  /*
  ** Non desctructive operations
  */
  
  mixed cast(string type)
  {
    switch(type) {
    case "array":
      return copy_value(table);
    }
  }

  array _indices()
  {
    return copy_value(fields);
  }  

  array _values()
  {
    return copy_value(table);
  }  
  
  int _sizeof()
  {
    return sizeof(table);
  }

  object reverse()
  {
    table = predef::reverse(table);
    return this_object();
  }
  
  array col(int|string c)
  {
    return copy_value(column(table, remap(c)));
  }

  array row(int r)
  {
    return copy_value(table[r]);
  }

  array `[](int|string c)
  {
    return col(c);
  }

  int `==(object t)
  {
    return equal(fields, indices(t)) && equal(table, values(t));
  }

  object append_bottom(object t)
  {
    if(!equal(indices(t), fields))
      TABLE_ERR("Table fields are not equal.");
    return object_program(this_object())(table+values(t), fields, metadata);
  }

  object append_right(object t)
  {
    if(sizeof(t) != sizeof(table))
      TABLE_ERR("Table sizes are not equal.");
    fields += indices(t);
    array v = values(t);
    for(int r = 0; r < sizeof(table); r++)
      table[r] = table[r] + v[r];
    return this_object();
  }

  mixed op_col(function f, int|string c, mixed ... args)
  {
    c = remap(c);
    mixed x = table[0][c];
    for(int r = 1; r < sizeof(table); r++)
      f(x, table[r][c], @args);
    return x;
  }

  mixed sum_col(int|string c)
  {
    return `+(@column(table, remap(c)));
  }

  mixed average_col(int|string c)
  {
    return sum_col(c)/sizeof(table);
  }

  mixed min_col(int|string c)
  {
    return op_col(min, c);
  }

  mixed max_col(int|string c)
  {
    return op_col(max, c);
  }

  object cols(int|string ... cs)
  {
    array t = ({});
    cs = remap(cs);
    for(int r = 0; r < sizeof(table); r++)
      t += ({ rows(table[r], cs) });
    return object_program(this_object())(t, rows(fields, cs), metadata);
  }

  object not_cols(int|string ... cs)
  {
    return cols(@remap(fields) - remap(cs, 1));
  }

  object copy()
  {
    return object_program(this_object())(table, fields, metadata);
  }

  object filter(function f, array(int|string)|int|string cs, mixed ... args)
  {
    array t = ({});
    cs = remap(arrayp(cs)?cs:({ cs }));
    foreach(table, mixed row)
      if(f(@rows(row, cs), @args))
	t += ({ row });
    return object_program(this_object())(t, fields, metadata);
  }

  /*
  ** Destructive operations
  */

  mixed meta(mixed|void m)
  {
    if(query_num_arg() == 1)
      metadata = m;
    return metadata;
  }

  object map_col(function f, array(int|string)|int|string cs, mixed ... args)
  {
    int ap = arrayp(cs);
    if(!catch(cs = remap(ap?cs:({ cs })))) {
      for(int r = 0; r < sizeof(table); r++) {
	mixed v = f(@rows(table[r], cs), @args);
	if(arrayp(v))
	  for(int i = 0; i < sizeof(v); i++)
	    table[r][cs[i]] = v[i];
	else
	  table[r][cs[0]] = v;
      }
    }
    return this_object();
  }

  // object map(mapping(int|string:function)|array|string|int f, mixed ... args)
  object map(mixed f, mixed ... args)
  {
    if(!sizeof(table))
      return this_object();

    if(stringp(f) || intp(f))
      f = ({ f });
    if(arrayp(f)) {
      f = mkmapping(f, Array.map(allocate(sizeof(f)),
				 lambda(int x, function q) { return q; },
				 args[0]));
      args = args[1..];
    }

    mapping m = ([]);
    array cs = remap(indices(f));
    f = mkmapping(cs, values(f));
    array(int) keys = indices(fields) - cs;
    foreach(table, array row) {
      string key = encode_value(rows(row, keys));
      if(array a = m[key])
	foreach(cs, int c)
	  a[c] = f[c](a[c], row[c], @args);
      else
	m[key] = copy_value(row);
    }
    table = values(m);
    return this_object();
  }

  object sum(int|string ... cs)
  {
    array v = ({});
    for(int i = 0; i < sizeof(cs); i++)
      v += ({ `+ });
    return map(mkmapping(cs, v));
  }

  object map_max(int|string ... cs)
  {
    array v = ({});
    for(int i = 0; i < sizeof(cs); i++)
      v += ({ max });
    return map(mkmapping(cs, v));
  }

  object sort(int|string ... cs)
  {
    if(!sizeof(cs))
      return this_object();
    int c;
    if(!catch(c = remap(cs[-1]))) {
      mapping m = ([]);
      for(int r = 0; r < sizeof(table); r++) {
	mixed d;
	if(!m[d = table[r][c]])
	  m[d] = ({ table[r] });
	else
	  m[d] += ({ table[r] });
      }
      array i = indices(m), v = values(m);
      predef::sort(i, v);
      table = v*({});
    }
    return sort(@cs[0..(sizeof(cs)-2)]);
  }

  object rsort(int|string ... cs)
  {
    return sort(@cs)->reverse();
  }
  
  object truncate(int n)
  {
    table = table[0..(n-1)];
    return this_object();
  }

  void decode(string s)
  {
    mapping m = decode_value(s);
    table = m->table;
    fields = m->fields;
    metadata = m->meta;
    initialize_fields();
  }

  string encode()
  {
    return encode_value(([ "table":table,"fields":fields,"meta":metadata ]));
  }

  void create(array(array)|void _table, array(string)|void _fields,
	      mixed|void _meta)
  {
    if(!arrayp(_table))
      TABLE_ERR("Table not array ("+typeof(_table)+")");
    if(!arrayp(_fields))
      TABLE_ERR("Fields not array ("+typeof(_fields)+")");

    table = copy_value(_table);
    fields = copy_value(_fields);
    metadata = copy_value(_meta);
    initialize_fields();
  }
}
