// LogView function

inherit "fu.pik";

#define MAX_SESSIONS 10000
#define SESSION_TIMEOUT 20*60

constant title = "3:Visitor sessions//2:Sessions";
constant fetch = ({ "Host", "Year", "Month", "Date",
		    "Hour", "Minute", "Second", "Path",
		    "Hits", "Referrer" });
constant as_is = 1;
constant invisible = 1;

int offset;
function refmatch, pagematch, ymd2jd;

mapping display(string display_type, array table)
{
  array fields = ({ "Host", "Hits", "Pages", "Start page", "End page",
		    "Start timestamp", "End timestamp",
		    "Status", "Year", "Month", "Date", "Hour" });
  array types = ({ "text", "num", "num", "text", "text",
		   "num", "num", "text", "text", "text", "text", "text" });
  return ([ "fields":fields, "types":types, "table":table ]);
}

array revise(array table)
{
  // werror("### WARNING! REVISE! ###\n");
  array v = ({});
  foreach(table, array a)
    if(a[7] != "OK")
      v += ({ a });
  return v;
}

array store(array table)
{
  return table;
}

array join(array old_table, array new_table)
{
  array v = ({});
  mapping m = ([]);
  foreach(old_table, array a) 
    if(a[7] == "OK")
      v += ({ a });
    else
      m[a[0]] = a;
  int n_sessions = sizeof(v) + sizeof(m);

  int lasttime;
  if(sizeof(new_table)) {
    sort(column(new_table, 5), new_table);
    lasttime = sort(column(new_table, 6))[-1];
  }

  foreach(new_table, array a) {
    string host = a[0];
    if(!m[host]) {
      if(a[0] != "unknown" && a[7] == "Start" && n_sessions < MAX_SESSIONS) {
	m[host] = a;
	n_sessions++;
      }
    } else {
      array b = m[host];
      if(b[6] + SESSION_TIMEOUT >= a[5]) {
	b[1] += a[1];
	b[2] += a[2];
	b[3] = pagematch(b[3]) || !pagematch(a[3])?b[3]:a[3];
	b[4] = pagematch(a[4])?a[4]:b[4];
	b[6] = a[6];
      } else {
	m_delete(m, host);
	if(n_sessions < MAX_SESSIONS) {
	  m[host] = a;
	  n_sessions++;
	}
	b[7] = "OK";
	v = ({ b }) + v;
      }
    }
  }

  foreach(values(m), array a) {
    if(a[6] + SESSION_TIMEOUT < lasttime)
      a[7] = "OK";
    v = ({ a }) + v;
  }
  return v;
}

void start()
{
  ymd2jd = LogView.Calendar.ymd2jd;
  pagematch = Regexp(log_str_Page_Regexp||"")->match;

  refmatch = match_multiple(log_str_Server_Domain_Name, log_str_Server_Alias);

  offset = ymd2jd(1990, 1, 1);
}

array fu(array data)
{
  array v = ({});
  foreach(data, array row) {
    if(row[0]) {
      string host = complete_host(row[0], log_str_Server_Domain_Name);
      if(host == "unknown") host = row[0];
      int timestamp = ((ymd2jd(@row[1..3])-offset)*24*3600 +
		       row[4]*3600 + row[5]*60 + row[6]);
      string page = row[7]||"";
      string hits = row[8];
      string referrer = row[9]||"";
      // werror(sprintf("%O %O %O\n", host, referrer, page));
      v += ({ ({ host, 1, pagematch(page)?1:0,
		 page, page, timestamp, timestamp,
		 "Start" }) + row[1..4] });
      // refmatch(referrer)?"Nada":"Start" }) + row[1..4] });
    }
  }
  return join(({}), v);
}
