
inherit "wizard";
inherit "wiz.pik";

void ws(mixed ... args) { werror(sprintf(@args)); }

constant name = "40:<p>Statistics groups...";
constant doc = "View group status and change group configuration.";
// constant doc = "Add new groups, select what logs belong to what groups, "
//                "enable/disable group functions, set group variables.";

mapping parse_config(object id, object o)
{
  mapping m = ([]);

  m->group = figure_out_group(id);

  m->fugroups = ({});
  foreach(o->list_fugroups(), string fugroup)
    if(id->variables["fugroup_"+fugroup] != "0")
      m->fugroups += ({ fugroup });

  m->logs = ([]);
  foreach(glob("logs:*", indices(id->variables)), string arg)
    if(id->variables[arg] != "0") {
      string logname, virtual;
      sscanf(arg, "logs:%s\n%s", logname, virtual);
      if(m->logs[virtual])
	m->logs[virtual] += ({ logname });
      else
	m->logs[virtual] = ({ logname });
    }

  m->group_args = ([]);
  foreach(glob("group_args_*", indices(id->variables)), string arg)
    m->group_args[arg[sizeof("group_args_")..]] = id->variables[arg];

  return m;
}

string verify_0(object id, object o)
{
  if(!id->variables->group)
    return 0;
  if(!figure_out_group(id))
    return "<font color=red><b>Please fill in the <em>Add new group</em> field."
           "</b></font>\n";
}

string page_0(object id, object o)
{
  string r = "";

  r += "Please select the group to configure. To add a new group,\n"
       "enter the new name and select <em>Add new group</em>.<p>";
  
  int i;
  array a = ({});
  foreach(o->list_groups(), string name) {
    a += ({ ({ "<font size=-1><var name=group type=radio"+
	       (i++?"":" default=checked")+
	       " value='group_"+name+"'></font> "+name }) });
  }
  a += ({ ({ "<font size=-1><var name=group type=radio "+
	     (i++?"":" default=checked")+" value=add_new_group></font> "
	     "<var name=new_group type=string size=30> "
	     "<font size=-1>Add new group.</font>" }) });
  r += html_table(({ "Group" }), a);

  if(string err = verify_0(id, o))
    r += err;
  
  // Clear wizard
  id->variables = ([ "action":id->variables->action,
		     "group":id->variables->group,
		     "new_group":id->variables->new_group ]);

  return r;
}

string log_table(array v, string title, mapping m, object o)
{
  array a = ({});
  mapping c = o->read_group_config(m->group);
  foreach(v, array row) {
    string d = "";
    if(c->logs && c->logs[row[1]] && mkmultiset(c->logs[row[1]])[row[0]])
      d = " default=on";
    a += ({ ({ "<font size=-1><var name='logs:"+row[0]+"\n"+row[1]+"' "
	       "type=checkbox"+d+"></font>&nbsp;"+row[0] /* , row[1] */ }) });
  }
  return html_table(({ "Log" /* , title */ }), a);
}

string page_1(object id, object o)
{
  string r = "";
  mapping m = parse_config(id, o);

  array l = ({});
  foreach(o->list_logs(), string name)
    l += ({ ({ name, o->virtual_name() }) });
  
  array v = ({});
  mapping semiremotes = o->list_virtual_logs();
  foreach(sort(indices(semiremotes)), string virtual)
    if(virtual != o->virtual_name())
      foreach(sort(semiremotes[virtual] || ({})), string name)
	v += ({ ({ name, virtual }) });
  
  mapping remotes = logs_of_the_galaxy(o);
  foreach(sort(indices(remotes)), string virtual)
    foreach(sort(remotes[virtual] || ({})), string name)
      v += ({ ({ name, virtual }) });

  r += "Please select none, one or several logs you want to import to\n"+
       "the statistics group ``"+m->group+"''.<p>\n";
  if(sizeof(l) /* || sizeof(v) */) {
    // r += "Please select the logs you want to belong to "+m->group+".<p>\n";
    if(sizeof(l))
      r += log_table(l, "Virtual Server", m, o);
#if 0
    if(sizeof(v))
      r += log_table(v, "Remote Virtual Server", m, o);
#endif
  } else
    r += "<em>No logs are available.</em>\n";

  return r;
}

string page_2(object id, object o)
{
  return enable_functions(parse_config(id, o)->group, id, o);
}

string page_3(object id, object o)
{
  return group_args(parse_config(id, o)->group, id, o);
}

string page_4(object id, object o)
{
  string r = "";
  mapping m = parse_config(id, o);

  r += "Please confirm the following settings:<p>";

  r += "<table>";
  r += "<tr><th align=right>Group: </th><td>"+m->group+"</td></tr>\n";
  r += "<tr><th align=right valign=top>Logs: </th><td>";
  foreach(indices(m->logs), string virtual)
    r += String.implode_nicely(m->logs[virtual])/*+" ("+virtual+")<br>"*/;
  r += "<tr><th align=right valign=top>Functions: </th><td>";
  foreach(m->fugroups, string name)
    r += name+"<br>";
  r += "</td></tr>\n";
  foreach(indices(m->group_args), string name)
    r += "<tr><th align=right>"+arg_title(name)+": </th><td>"+
         m->group_args[name]+"</td></tr>\n";
  r += "</table>";
  
  return r;
}

void wizard_done(object id, object o)
{
  mapping m = parse_config(id, o);
  o->save_group_config(m->group, m);
}
