
inherit "wizard";
inherit "wiz.pik";

constant name = "30:<p>Logs...";
constant doc = "View log status and change log configuration.";

mapping parse_config(object id, object o)
{
  mapping m = ([]);

  m->status = id->variables->status!="Disabled"?id->variables->status:0;
  
  m->mode = "log";
  string log_file = id->variables->log_file;
  string clf_file = o->clf_imports()[log_file];
  if(log_file == "jordgubbar!") {
    m->mode = "this";
    log_file = id->variables->new_this_log_file ||
               o->read_main_config()->this_server_log ||
               o->virtual_short_name();
  } else if(clf_file && (clf_file = id->variables->clf_file || clf_file)) {
    m->mode = "clf";
    m->clf_file = clf_file;
  }
  m->log_file = log_file - "/";

  if(m->clf_file)
    m->clf_file = combine_path(getcwd()+"/", m->clf_file);

  m->log_args = ([]);
  foreach(indices(o->list_log_args(log_file)), string s)
    if(string v = id->variables["log_args_"+s])
      m->log_args[s] = v;

  return m;
}

string page_0(object id, object o)
{
  string r = "";

  r += "Please select a log to configure.<p>\n";

  array v = ({});
  mapping clf = o->clf_imports();
  string t = o->read_main_config()->this_server_log;
  v += ({ ({ "<font size=-1><var name=log_file type=radio default=checked "
	     "value='jordgubbar!'></font> "+(t?t:"")+" "
	     "(log for this virtual server)",
	     t && o->read_log_config(t)->status?"Active":"Disabled",
	     t?mb_size(o->log_size(t)):"-" }) });
  foreach(o->list_logs() - ({ t }), string name) {
    string clf_file = clf[name];
    v += ({ ({ "<font size=-1><var name=log_file type=radio "
	       "value='"+name+"'></font> "+name,
	       clf_file && o->read_log_config(name)->status?"Import":"-",
	       mb_size(o->log_size(name)) }) });
  }
#if 0
  v += ({ ({ "<font size=-1><var name=log_file type=radio "
	     "value=add_new_group></font> "
	     "<var name=new_group type=string size=30> "
	     "<font size=-1></font>", "" }) });
#endif
  r += html_table(({ "Log", "Status", "Size (Mb)" }), v);

  // Clear wizard
  id->variables = ([ "action":id->variables->action,
		     "log_file":id->variables->log_file ]);
  
  return r;
}

string page_1(object id, object o)
{
  string r = "";

  mapping m = parse_config(id, o);
  mapping c = o->read_log_config(m->log_file);

  switch(m->mode) {
  case "this":
    r += "Change status by using the select box. <em>Active</em> means that\n"
         "LogView will log this virtual server ("+o->virtual_name()+").\n"
         "<em>Disabled</em> means that LogView will not log new data.<p>\n";
    r += "<b>Status</b> <var name=status type=select"
         " default="+(c->status?"Active":"Disabled")+
         " choices=Active,Disabled><p>";
    r += "Change the name by editing the text box below.<p>\n";
    r += "<b>Log Name</b><br>"
         "<var name=new_this_log_file type=string size=50 default='"+
         m->log_file+"'><p>\n";
    break;
  case "clf":
    r += "Change status by using the select box. <em>Import</em> means that\n"
         "the log will be read by LogView once every update. <em>Disabled</em>\n"
         "means that LogView will not read any new data from the log.<p>\n";
    r += "<b>Status </b> <var name=status type=select"
         " default="+(c->status?"Import":"Disabled")+
         " choices=Import,Disabled><p>\n";
    r += "Change the log file name by editing the text box below.<p>";
    r += "<b>Common Log Format file name</b> "
         "<font size=-1>(full path)</font><br>"
         "<var name=clf_file type=string size=50 default='"+
         m->clf_file+"'><p>\n";
    break;
  default:
    return 0;
  }
  
  return r;
}

string page_2(object id, object o)
{
  mapping m = parse_config(id, o);
  return log_args(m->log_file, id, o);
}

string page_3(object id, object o)
{
  string r = "";

  r += "Confirm the following settings:<p>";
  r += "<table>";

  mapping m = parse_config(id, o);
  switch(m->mode) {
  case "clf":
    r += "<tr><th nowrap align=right>Common Log:</th><td>"+
         m->clf_file+"</td></tr>\n";
  case "log":
    r += "<tr><th nowrap align=right>Binary Log:</th><td>"+
         m->log_file+"</td></tr>\n";
    break;
  case "this":
    r += "<tr><th nowrap align=right>This server log:</th><td>"+
         m->log_file+"</td></tr>\n";
    break;
  default:
    return "<b>que?</b>";
  }

  if(m->mode == "log") {
    r += "<tr><th align=right>Status:</th><td>-</td></tr>\n";
  } else {
    r += "<tr><th align=right>Status:</th><td>"+
         (m->status||"Disabled")+"</td></tr>\n";
  }
  r += "<tr><th>&nbsp;</th><td>&nbsp;</td></tr>\n";
  foreach(sort(indices(m->log_args)), string arg) {
    r += "<tr><th nowrap align=right>"+arg_title(arg)+":</th><td>"+
         m->log_args[arg]+"</td></tr>\n";
  }
  r += "</table>\n";
  
  return r;
}

void wizard_done(object id, object o)
{
  mapping m = parse_config(id, o);
  if(!sizeof(m->log_file))
    m->log_file = 0;
  
  switch(m->mode) {
  case "this":
    mapping c = o->read_main_config();
    c->this_server_log = m->log_file;
    o->save_main_config(c);
  case "log":
  case "clf":
    if(m->log_file)
      o->save_log_config(m->log_file, m);
    break;
  }
}
