
inherit "wizard";
inherit "wiz.pik";

constant name = "50:<p>Import Common Log Format...";
constant doc = "Import Common Log Format (ASCII) files.";

mapping parse_config(object id, object o)
{
  mapping m = ([]);
  
  array a = ({});
  foreach(o->list_groups(), string s) {
    if(id->variables["group_"+s] != "0")
      a += ({ s });
  }
  if(id->variables->new_group != "0")
    a |= ({ id->variables->new_group_name });
  m->groups = sort(a);

  m->fugroups = ({});
  foreach(o->list_fugroups(), string fugroup)
    if(id->variables["fugroup_"+fugroup] != "0")
      m->fugroups += ({ fugroup });
  
  m->group_args = ([]);
  foreach(glob("group_args_*", indices(id->variables)), string arg)
    m->group_args[arg[sizeof("group_args_")..]] = id->variables[arg];

  m->clf_file = id->variables->clf_file;
  if(m->clf_file)
    m->clf_file = combine_path(getcwd()+"/", m->clf_file);
  m->log_file = id->variables->log_file?id->variables->log_file - "/":0;

  m->log_args = ([]);
  foreach(indices(o->list_log_args(m->log_file)), string s)
    if(string v = id->variables["log_args_"+s])
      m->log_args[s] = v;

  return m;
}

string verify_0(object id, object o)
{
  mapping m = parse_config(id, o);
  if(!m->clf_file)
    return 0;
  if(!sizeof(m->clf_file))
    return "Please fill in the Common Log file name.";
  if(Stdio.file_size(m->clf_file) < 0)
    return "The file does not exist.";
  return 0;
}

string page_0(object id, object o)
{
  string r = "";

#if 0
  array a = o->list_logs();
  mapping clf = o->clf_imports();
  if(sizeof(a)) {
    array v = ({});
    r += "List of logs.<p>\n";
    foreach(sort(a), string name) {
      if(clf[name])
	v += ({ ({ clf[name], name, mb_size(o->log_size(name)) }) });
      else
	v += ({ ({ "-", name, mb_size(o->log_size(name)) }) });
    }
    r += html_table(({ "Common Log Format file", "Binary log file",
		       "Size (Mb)" }), v);
  }
#endif

  //  r += "LogView will convert the Common Log file to a binary log file.\n"
  //       "The binary log is used by the statistics functions.<p>";
  r += "Please enter the Common Log Format file name with full path.<p>\n";
  r += "<b>Common Log Format file name</b> "
       "<font size=-1>(full path)</font><br>"
       "<var name=clf_file type=string size=50><p>\n";

  if(string err = verify_0(id, o))
    r += "<p><font color=red><b>"+err+"</b></font>";
  
  return r;
}

string verify_1(object id, object o)
{
  mapping m = parse_config(id, o);
  if(!m->log_file)
    return 0;
  if(!sizeof(m->log_file))
    return "Please fill in a name.";
  if(mkmultiset(o->list_logs())[m->log_file])
    return "Please enter a <i>unique</i> name.";
}

string page_1(object id, object o)
{
  string r = "";

  r += "Please enter a unique name for the log.<p>\n";
  r += "<b>Log name</b> "
       "<font size=-1>(enter a unique name)</font><br>"
       "<var name=log_file type=string size=50><p>\n";

  if(string err = verify_1(id, o))
    r += "<p><font color=red><b>"+err+"</b></font>";

  return r;
}

string page_2(object id, object o)
{
  return log_args(0, id, o);
}

string verify_3(object id, object o)
{
  catch {
    mapping m = parse_config(id, o);
    foreach(m->groups, string group)
      if(!sizeof(group))
	return "Please fill in the new group name.";
  };
}

string page_3(object id, object o)
{
  string r = "";

  r += "Please select none, one or several groups you want "
       " '"+id->variables->log_file+"' to belong to. To add a new\n"
       "group, enter the new name and select 'Add new group'.<p>";
  
  array a = ({});
  foreach(o->list_groups(), string name) {
    a += ({ ({ "<font size=-1><var name='group_"+name+"' type=checkbox "
	       "value='"+name+"'></font> "+name }) });
  }
  a += ({ ({ "<font size=-1><var name=new_group default=selected "
	     "type=checkbox></font> "
	     "<var name=new_group_name type=string size=30 "
	     "default='"+id->variables->log_file+"'> "
	     "<font size=-1>Add new group.</font>" }) });
  r += html_table(({ "Group" }), a);

  if(string err = verify_3(id, o))
    r += "<p><font color=red><b>"+err+"</b></font>";

  return r;
}

string page_4(object id, object o)
{
  mapping m = parse_config(id, o);
  array a = m->groups - o->list_groups();
  if(!sizeof(a)) return 0;
  return enable_functions(a[0], id, o, mkmultiset(o->list_fugroups() -
					     ({ "Visitor profiles" })));
}

string page_5(object id, object o)
{
  mapping m = parse_config(id, o);
  array a = m->groups - o->list_groups();
  if(!sizeof(a)) return 0;
  return group_args(a[0], id, o);
}

string page_6(object id, object o)
{
  string r = "";

  mapping m = parse_config(id, o);

  int i = 0;
  r += "Confirm the following settings:<p>";
  r += "<table>";
  r += "<tr><th nowrap valign=top align=right>Common Log:</th><td>"+
       m->clf_file;
  if(Stdio.file_size(m->clf_file) < 0)
    r += "<br>(This file does not exist)";
  r += "</td></tr>\n";
  r += "<tr><th nowrap align=right>Log Name:</th><td>"+
       m->log_file+"</td></tr>\n";
  r += "<tr><th nowrap align=right>Groups:</th><td>"+
       String.implode_nicely(m->groups)+"</td></tr>\n";
  r += "<tr><th>&nbsp;</th><td>&nbsp;</td></tr>\n";
  foreach(sort(indices(o->list_log_args())), string arg) {
    r += "<tr><th nowrap align=right>"+arg_title(arg)+":</th><td>"+
         id->variables["log_args_"+arg]+"<td></tr>";
    r += "</td></tr>\n";
  }
  r += "</table>\n";
  
  return r;
}

void wizard_done(object id, object o)
{
  mapping tmp = parse_config(id, o);
  array groups = tmp->groups - o->list_groups();
  string new_group = sizeof(groups)?groups[0]:0;
  string virtual = o->virtual_name();
  foreach(tmp->groups, string group) {
    mapping m = ([]);
    if(group == new_group) {
      m->fugroups = tmp->fugroups;
      m->group_args = tmp->group_args;
    } else
      m = o->read_group_config(group) || ([]);
    
    if(!m->logs) m->logs = ([]);
    if(!m->logs[virtual])
      m->logs[virtual] = ({ tmp->log_file });
    else
      m->logs[virtual] |= ({ tmp->log_file });
    o->save_group_config(group, m);
  }

  mapping m = ([]);
  m->mode = "clf";
  m->status = "Import";
  m->clf_file = tmp->clf_file;
  m->log_args = tmp->log_args;
  o->save_log_config(tmp->log_file, m);
}
