
inherit "wizard";
inherit "wiz.pik";

constant name = "70:<p>Report maker...";
constant doc = "Create and edit reports.";

mapping parse_report(object id)
{
  string report;
  if(id->variables->report == "add_new_report" ||
     sizeof(id->variables->new_report))
    report = (sizeof(id->variables->new_report)?
	      id->variables->new_report:"unknown");
  else
    report = id->variables->report;

  return ([ "email":id->variables["email#"+report],
	    "report":report,
	    "to":id->variables["to#"+report],
	    "from":id->variables["from#"+report],
	    "subject":id->variables["subject#"+report],
	    "description":id->variables["description#"+report],
	    "contents":id->variables["contents#"+report],
	    "frequency":id->variables["frequency#"+report],
	    "mailtest":id->variables["mailtest#"+report] ]);
}

string page_0(object id, object o)
{
  string r = "";

  r += "Please select a report to edit.<p>\n";

  int i = 0;
  array v = ({});
  mapping reports = o->read_main_config()->reports || ([]);
  foreach(sort(indices(reports)), string name)
    v += ({ ({ "<font size=-1><var name=report type=radio "+
	       (i++?"":"default=checked ")+
	       "value='"+name+"'></font> "+name }) });
  v += ({ ({ "<font size=-1><var name=report type=radio "+
	     (i++?"":" default=checked")+" value=add_new_report></font> "
	     "<var name=new_report type=string size=30> "
	     "<font size=-1>Add new report.</font>" }) });
  r += html_table(({ "Report" }), v);

  return r;
}

string page_1(object id, object o)
{
  string r = "";
  string report = parse_report(id)->report;
  mapping e = (o->read_main_config()->email_service||([]))[report];
  string d = (o->read_main_config()->report_descriptions||([]))[report]||""; 
  string s = (o->read_main_config()->reports||([]))[report]||""; 

  r += ("<b>Report by email</b>&nbsp;&nbsp;&nbsp;<var name='email#"+report+"' "
	"type=select default="+(e?"Yes":"No")+" choices=Yes,No><p>\n"
	
	"<p>Report description for ``"+report+"''.<p>\n"
	"<cvar type=text name='description#"+report+"' cols=60 rows=2>"
	""+d+"</cvar>"
	
	"<p>Please edit the report below.<p>\n"
	"<cvar type=text name='contents#"+report+"' cols=60 rows=16>"
	""+s+"</cvar>");

  return r;
}

string verify_2(object id, object o)
{
  mapping m = parse_report(id);
  if(!m->to)
    return 0;
  if(!sizeof(m->to))
    return "Please fill in the 'To' field.";
  return 0;
}

string page_2(object id, object o)
{
  string r = "";
  mapping m = parse_report(id);
  string report = m->report;

  if(m->email != "Yes")
    return 0;

  string s = (o->read_main_config()->reports||([]))[report] || "";
  mapping e = (o->read_main_config()->email_service||([]))[report];

  string to = (e?e->to:0)||"";
  string from = (e?e->from:0)||"logview";
  string subject = (e?e->subject:0)||"LogView report on "+report;
  string frequency = (e?e->frequency:0)||"Mondays";
  
  r += ("Send report ``"+report+"'' by email "
	"(the report will be sent five o'clock in the morning).<p>"
	"<table border=0>"
	"<tr><th align=left>To</th><td><var name='to#"+report+"' "
	"type=string size=40 default='"+to+"'></td></tr>"
	"<tr><th align=left>From</th><td><var name='from#"+report+"' "
	"type=string size=40 default='"+from+"'></td></tr>"
	"<tr><th align=left>Subject</th><td><var name='subject#"+report+"' "
	"type=string size=40 default='"+subject+"'></td></tr>"

	"<tr><th align=left>Frequency</th><td><var name='frequency#"+report+"' "
	"type=select default='"+frequency+"' "
	"choices='Every day,Mondays,Tuesdays,Wednesdays,Thursdays,"
	"Fridays,Saturdays,Sundays,--Never--'></td></tr>\n"
	
	"</table>");

  if(string err = verify_2(id, o))
    r += "<p><font color=red><b>"+err+"</b></font>";

  return r;
}

string page_3(object id, object o)
{
  string r = "";
  mapping m = parse_report(id);
  string report = m->report;

  if(m->email == "Yes") {
    r += ("A report for ``"+report+"'' will be sent by email "
	  "five o'clock in the morning on "+m->frequency+".<p>"
	  "<ul><table>"
	  "<tr><th align=left>To</th><td>"+m->to+"</td></tr>\n"
	  "<tr><th align=left>From</th><td>"+m->from+"</td></tr>\n"
	  "<tr><th align=left>Subject</th><td>"+m->subject+"</td></tr>\n"
	  "<tr><th align=left>Frequency</th><td>"+m->frequency+"</td></tr>\n"
	  "</table></ul><p>"
	  
	  "<b>Send test email now?</b>"
	  "&nbsp;&nbsp;&nbsp;<var name='mailtest#"+report+"' "
	  "type=select default=No choices=Yes,No>");
    
  }
  r += ("Press OK to confirm the report ``"+report+"''.\n");

  return r;
}

void wizard_done(object id, object o)
{
  mapping m = parse_report(id);
  string report = m->report;
  mapping mc = o->read_main_config();

  mc->reports = mc->reports || ([]);
  mc->report_descriptions = mc->report_descriptions || ([]);
  mc->email_service = mc->email_service || ([]);

  mc->reports[report] = m->contents;
  mc->report_descriptions[report] = m->description;
  if(m->email == "Yes") {
    mc->email_service[report] = ([ "to":m->to,
				   "from":m->from,
				   "subject":m->subject,
				   "frequency":m->frequency ]);
  } else
    m_delete(mc->email_service, report);
  o->save_main_config(mc);
  
  if(m->email == "Yes" && m->mailtest == "Yes")
    o->parent->email_service(report);
}
