// Roxen LogView 1.0
// Program Maker: Fredrik Noring
//
// Copyright  1996, 1997, 1998 Idonex AB. All rights reserved.

constant cvs_version = "$Id: logview.pike,v 1.167 1999/07/23 16:36:06 mast Exp $";

#define LOGVIEW_NUMBER "1.0.60"
#define LOGVIEW_VERSION (LOGVIEW_NUMBER"")

#define PIE_CHART_SIZE 350

#define MAX_OPERATIONS 100000

#define BUCKET_SIZE 10000
#define BUCKET_CACHE_SIZE 1024
#define BUCKET_CACHE_MEM 16384

#define QUEUE_TIMEOUT 60
#define QUEUE_SIZE 100

#define EMAIL_SERVICE_TIMEOUT 300
#define DELAYED_ACTION_TIMEOUT 10

#define QUERY_CACHE  0
#define DEBUG_TIMING 0

#if __VERSION__ >= 0.6
import ".";
#endif

// These functions may be called by Roxen. They must check if kludge_start()
// has been called. Do not run kludge_start() in start(), as it may create
// files with the wrong permissions/owner.
//
//    tag_logview(), log() and find_file().
//

void kludge_start(int num, object configuration);
string query_location_low(object id);

// #define UTILIZATION 1

#define OMIT_FIELD 200

#define MKDIR(s) { if((file_stat(s)||({0,0}))[1] != -2 && !mkdirhier(combine_path(s, "bogus"))) eventlog->error("LogView failed to create directory '"+s+"'\nSuggestion: Check for permission problems."); }

#include <module.h>
#include <roxen.h>

inherit "http";
inherit "module";
inherit "wizard";
inherit "roxenlib";

#include "LogView.pmod/fu/country.h"
mapping reverse_country_codes = mkmapping(values(country_codes),
					  indices(country_codes));

constant language = roxen->language;

#ifdef THREADS
object megals = Thread.Mutex();
#define MEGALS object key; catch(key = megals->lock());
#else
object megals;
#define MEGALS
#endif

// #ifdef THREADS
//   object key;
//   catch(key = mega_ls->lock());
// #endif

#define TABLE_PAGE_SIZE 100
#define MIN(a,b) ((a)<(b)?(a):(b))

string logo;
string rpckey = (string) random(2*1024*1024*1024-1);

void check_queue(int|void);

mapping logged = ([]);

object sql, eventlog, gate, rpc, httprpc, fu, roxen_configuration;
string mydisklocation;
object this_server_bucket;
mapping log_defaults, group_defaults;
string virtual_name, virtual_short_name;
int summary_pid, log_this_server, loaded, logview_is_new, update_every_hour;

object update_friendly_lock;
int friendly_process_policy_counter = 0;

#define ERR(msg) throw(({ "(LogView) "+msg+"\n", backtrace() }))

string get_fu_directory()
{
  return combine_path(getcwd(), __FILE__ + "/../LogView.pmod/fu");
}

void remove_friendly_lock()
{
  if(update_friendly_lock)
    destruct(update_friendly_lock);
}

class UpdateLock {
    static private string lock_file;
  static private int have_lock = 0;

  static private int get_locked_pid()
  {
    return (int) (Stdio.read_file(lock_file)||"0");
  }

  void destroy()
  {
    /* Release PID lock when the object is destroyed. */
    if(have_lock && get_locked_pid() == getpid())
      rm(lock_file);
  }

  void create()
  {
    // Check if we've got a /tmp directory at all.
    if(!file_stat("/tmp"))
      return;
    
    lock_file = "/tmp/LogView_update_lockfile";

    for(int tryout = 0; tryout < 2; tryout++) {
      object f = Stdio.File();
      
      if(f->open(lock_file, "cxw")) {
	/* We're being paranoid... */
	mixed err = catch {
	  string s = sprintf("%d", getpid());
	  if(f->write(s) != sizeof(s))
	    ERR("Failed to write lock-file");
	  f->close();
	  if(get_locked_pid() != getpid())
	    ERR("Failed to reread lock-file");
	  
	  /* We now have the lock! */
	  have_lock = 1;
	  return;
	};
	rm(lock_file);
	throw(err);
      }
      
      if(!file_stat(combine_path(lock_file, "../")))
        ERR(sprintf("The database does not exist (use the `c' flag)."));
      
      /* Failed to obtain lock, now trying some obscure techniques... */
      int pid = get_locked_pid();
      if(pid && kill(pid, 0))
	ERR(sprintf("Out-locked by PID %d", pid));
      
      if(!rm(lock_file))
	ERR("Lock removal failure (insufficient permissions?)");
      sleep(1);
    }
    ERR("Lock tryout error (insufficient permissions?)");
  }
}

int module_dependencies(object configuration, array (string) modules)
{
  if(configuration) {
    foreach (modules, string module) {
      if(!configuration->modules[module] ||
	 (!configuration->modules[module]->copies &&
	  !configuration->modules[module]->master))
	configuration->enable_module(module+"#0");
    }
    if(roxen->root)
      roxen->configuration_interface()->build_root(roxen->root);
  }
  _do_call_outs();
  return 1;
}

int duks(string dir)
{
  int n = (file_stat(dir) || ({ 0, -4711 }) )[1];
  if(n >= 0) return (n/1024)+1;
  if(n != -2) return 0;
  n = 0;
  foreach(get_dir(dir) || ({}), string s)
    n += duks(combine_path(dir, s));
  return n;
}

void check_login(object id)
{
  if(logged[id->remoteaddr] + 1000 < time()) {
    eventlog->notice("Administrator logged on from " +
		     roxen->blocking_ip_to_host(id->remoteaddr) + ".");
  }
  logged[id->remoteaddr] = time(1);
}

/*** ***/
class baba_base {
  object parent, eventlog;

  void ws(mixed ... args) { werror(sprintf(@args)); }

  mixed read_safe(string filename)
  {
    if(string s = Stdio.read_bytes(filename)) {
      mixed d;
      if(catch(d = decode_value(s))) {
	eventlog->error("LogView failed to decode '"+filename+"'.\n");
	return 0;
      }
      return d;
    }
    return 0;
  }

  void write_safe(string filename, mixed m)
  {
    object file;
    string s = encode_value(m);
    if(!(file = open(filename+".tmp", "cwt"))) {
      eventlog->error("LogView failed to open (for write) '"+filename+"'.\n"
		      "Suggestion: Check for premission problems.");
      throw(({ "cannot open table file\n", backtrace() }));
    }

    int r = file->write(s) == sizeof(s);
    file->close();

    if(r) {
      rm(filename);
      mv(filename+".tmp", filename);
    } else {
      eventlog->error("LogView failed to write '"+filename+"'.\n"
		      "Suggestion: Check if disk is full.");
      rm(filename+".tmp");
    }
  }

  string cache_dir()
  {
    return combine_path(getcwd(), combine_path(parent->query("logpath"),
					       "cache/"));
  }

  string log_dir(string|void name)
  {
    return combine_path(getcwd(), combine_path(parent->query("logpath"),
					       "logs/"+((name||"")-"/")));
  }

  string group_dir(string|void name)
  {
    return combine_path(getcwd(), combine_path(parent->query("logpath"),
					       "groups/"+((name||"")-"/")));
  }

  string config_dir(string|void name)
  {
    return combine_path(getcwd(), combine_path(parent->query("logpath"),
					       "config/"+((name||"")-"/")));
  }

  string waste_dir(string|void name)
  {
    return combine_path(getcwd(), combine_path(parent->query("logpath"),
					       "waste/"+((name||"")-"/")));
  }

  array|string true_names(array|string a)
  {
    array names;
    string name;
    if(stringp(a)) names = ({ a }); else names = a;
    for(int i = 0; i < sizeof(names); i++)
      if(sscanf(names[i], "%*d:%s", name) == 2)
	names[i] = name;
    return stringp(a)?names[0]:names;
  }

  void create(object parent_in)
  {
    parent = parent_in;
    eventlog = parent->eventlog;
  }
}

class baba_remote {
  static private inherit baba_base;

#define UNKN "Reason unknown."
  string msg = UNKN;

  string exit_msg()
  {
    string s = msg;
    msg = UNKN;
    return s;
  }

  string bonk()
  {
    if(msg != UNKN)
      throw(({ "#bonk", backtrace() }));
    else
      return "#bonk";
  }

  void notice(string msg) { eventlog->notice(msg); }
  void warning(string msg) { eventlog->warning(msg); }
  void error(string msg) { eventlog->error(msg); }

  string virtual_name()
  {
    return parent->virtual_name;
  }

  string virtual_short_name()
  {
    return parent->virtual_short_name;
  }
  
  mapping read_main_config()
  {
    return read_safe(config_dir("config")) || ([]);
  }
  
  mixed read_log_config(string name)
  {
    return read_safe(combine_path(log_dir(name), "log.config")) || ([]);
  }
  
  mixed read_group_config(string name)
  {
    return read_safe(combine_path(group_dir(name), "group.config")) || ([]);
  }
  
  array list_groups()
  {
    return get_dir(group_dir()) || ({});
  }
  
  array list_logs()
  {
    return get_dir(log_dir()) || ({});
  }
  
  int log_size(string|void name)
  {
    return parent->duks(log_dir(name));
  }
  
  int group_size(string|void name)
  {
    return parent->duks(group_dir(name));
  }

  string this_server_log()
  {
    return read_main_config()->this_server_log;
  }
  
  mapping clf_imports()
  {
    mapping m = ([]);
    foreach(list_logs(), string name) {
      mapping c = read_log_config(name);
      if(c->mode == "clf")
	m[name] = c->clf_file;
    }
    return m;
  }

  array(string) list_fugroups()
  {
    array groups = ({});
    object fu = parent->fu;
    foreach(fu->list_fus(), string funame)
      groups |= ({ fu->load_fu(funame)->group });
    return true_names(sort(groups));
  }

  array(string) list_fus(string fugroup)
  {
    mapping m;
    array fus = ({});
    object fu = parent->fu;
    fugroup = lower_case(fugroup);
    foreach(fu->list_fus(), string funame)
      if((lower_case(true_names((m = fu->load_fu(funame))->group)) == fugroup)
	 && !m->invisible)
	fus |= ({ m->title });
    return true_names(sort(fus));
  }

  string fu_title(string funame)
  {
    string title;
    mapping m = parent->fu->load_fu(funame);
    if(lower_case(title = true_names((m->title/"(")[0])))
      return title;
    return 0;
  }

  string find_fu(string title, string vues)
  {
    object fu = parent->fu;
    title = lower_case(title);
    foreach(fu->list_fus(), string funame) {
      mapping m = fu->load_fu(funame);
      if(lower_case(true_names((m->title/"(")[0])) == title &&
	 (!m->vues || mkmultiset(m->vues)[vues]) && !m->invisible)
	return funame;
    }
    return 0;
  }
  
  array(string) list_all_fufilenames(string group)
  {
    return copy_value(parent->fu->list_fus());
  }
  
  array(string) list_fufilenames(string fugroup)
  {
    mapping m;
    array fus = ({}), fufiles = ({});
    object fu = parent->fu;
    fugroup = lower_case(fugroup);
    foreach(fu->list_fus(), string funame)
      if(lower_case(true_names((m=fu->load_fu(funame))->group)) == fugroup) {
	fus += ({ m->title });
	fufiles += ({ funame });
      }
    sort(fus, fufiles);
    return fufiles;
  }

  mapping list_log_args(string name)
  {
    mapping m = ([]);
    mapping c = read_log_config(name) || ([]);
    if(!c->log_args) c->log_args = ([]);
    foreach(parent->fu->list_log_args(), string arg)
      m[arg] = c->log_args[arg] || parent->log_defaults[arg] || "";
    return m;
  }
  
  mapping list_group_args(string name)
  {
    mapping m = ([]);
    mapping c = read_group_config(name) || ([]);
    if(!c->group_args) c->group_args = ([]);
    foreach(parent->fu->list_group_args(), string arg)
      m[arg] = c->group_args[arg] || parent->group_defaults[arg] || "";
    return m;
  }
  
  mapping list_virtual_logs()
  {
    mapping m = ([]);
    foreach(roxen->configurations, object o) {
      string name = o->query("name");
      if(!name || !sizeof(name))
	name = o->name;
      if(o->modules->logview) {
	// Make sure the other LogView is started properly.
	o->modules->logview->enabled->kludge_start(0, parent->
						   roxen_configuration);
	m[name] = o->modules->logview->enabled->gate->list_logs();
      } else
	m[name] = 0;
    }
    return m;
  }

  string version()
  {
    return read_main_config()->version||"1.0.43";
  }

  mixed tag_logview(mixed ... args)
  {
    return parent->tag_logview(@args);
  }

  int get_nice_value()
  {
    return (int) parent->query("update-nice-value");
  }
  
  void remove_friendly_lock()
  {
    parent->remove_friendly_lock();
  }

  string get_fu_directory()
  {
    return parent->get_fu_directory();
  }
  
  void create(mixed ... args) { ::create(@args); }
}

static private mapping PDB_cache = ([]);
object PDB_Manager(string db)
{
#if 0
  return PDB_cache[db] = PDB_cache[db] || LogView.PDBSql.db(db, "wceC", sql,
							    virtual_name);
#endif
  // return PDB_cache[db] = PDB_cache[db] || PDB.db(db, "swceC");
  return PDB_cache[db] = PDB_cache[db] || Yabu.db(db, "cwtC");
}
object PDB_Drop(string db)
{
  if(PDB_Manager(db))
    PDB_cache[db]->purge();
}

static private object cache_table;
void start_cache()
{
#if QUERY_CACHE
  if(!cache_table)
    cache_table = Yabu.lookup(gate->cache_dir()+"db", "wsC")["query"];
#endif
}
string cache_check(string key)
{
  start_cache();
  // werror("### cache check "+(cache_table[key]?"OK":"FAIL")+".\n");
  return cache_table?cache_table[key]:0;
}
void cache_insert(string key, string s)
{
  start_cache();
  // werror("### cache insert.\n");
  if(cache_table)
    cache_table[key] = s;
}
void cache_drop()
{
  if(cache_table)
    cache_table->purge();
  cache_table = 0;
}

class baba_gate {
  inherit baba_base;
  inherit baba_remote;

  // Waste handler
  void empty_wastebin()
  {
    LogView.RmDir.rm_dir(waste_dir());
  }

  // Note! Keep this function behind walls!
  static private void waste(string thing)
  {
    empty_wastebin();
    MKDIR(waste_dir());
    mv(thing, waste_dir((string) random(2*1024*1024-1)));
    empty_wastebin();
  }
  
  // ...

  static private void exit_update_callback(function f, mapping t,
					   mixed ... args)
  {
    if(parent->rpc->num_connections() && ++t->x < DELAYED_ACTION_TIMEOUT) {
      call_out(exit_update_callback, 1, f, t, @args);
      return;
    }
    if(parent->rpc->num_connections())
      kill(parent->summary_pid, signum("SIGTERM"));
    if(f)
      f(@args);
  }

  void exit_update(string s, function f, mixed ... args)
  {
    if(parent->rpc->num_connections()) {
      call_out(exit_update_callback, 1, f, ([]), @args);
      msg = s||UNKN;
    } else
      f(@args);
  }

  object read_group_db(string name)
  {
    return parent->PDB_Manager(combine_path(group_dir(name), "db"));
  }
  
  object write_group_db(string name)
  {
    return parent->PDB_Manager(combine_path(group_dir(name), "db"));
  }

  void update_log_this_server()
  {
    parent->log_this_server = read_log_config(read_main_config()->
					      this_server_log||"")->
      status?1:0;
#if 0
    if(parent->log_this_server)
      ws("LOG ENABLED.\n");
    else
      ws("LOG DISABLED.\n");
#endif
  }
  
  void save_main_config(mapping m)
  {
    MKDIR(config_dir());
    // parent->main_config_data = copy_value(m);
    write_safe(config_dir("config"), m);
    update_log_this_server();
  }

  void save_log_config(string name, mixed m)
  {
    MKDIR(log_dir(name));
    write_safe(combine_path(log_dir(name), "log.config"), m);
    update_log_this_server();
  }

  void save_group_config(string name, mixed m)
  {
    MKDIR(group_dir(name));
    write_safe(combine_path(group_dir(name), "group.config"), m);
  }

  void rm_log(string name)
  {
    waste(log_dir(name));
  }

  void purge_log(string name)
  {
    waste(combine_path(log_dir(name), "import"));
    waste(combine_path(log_dir(name), "log.q"));
    waste(combine_path(log_dir(name), "log.i"));
    waste(combine_path(log_dir(name), "log"));
  }
  
  void purge_group(string name)
  {
    string dbname = combine_path(group_dir(name), "db");
    parent->PDB_Drop(dbname);
    waste(dbname);
  }

  void rm_group(string name)
  {
    purge_group(name);
    waste(group_dir(name));
  }

  object fu()
  {
    return parent->fu;
  }

  int start_summarizing()
  {
    return parent->start_summarizing();
  }

  object vue(string name)
  {
    return parent->PDB_Manager(combine_path(group_dir(name), "db"));
  }
  
  object safe_vue(string name)
  {
    if(sizeof(read_group_config(name)))
      return parent->PDB_Manager(combine_path(group_dir(name), "db"));
  }

  string upgrade_version()
  {
    mapping m = read_main_config();
    m->version = LOGVIEW_NUMBER;
    save_main_config(m);
  }

  object sql()
  {
    string s = read_main_config()->sql_database;
    return s?Sql.sql(s):0;
  }

  void clear_query_cache()
  {
    parent->cache_drop();
  }
  
  void create(mixed ... args) { ::create(@args); }
}

void bundle()
{
  if(this_server_bucket) {
    // werror("toto ");
    string toto = combine_path(query("logpath"), "working");
    this_server_bucket->flush();
    this_server_bucket = 0;
    mv(combine_path(toto, "now"),
       combine_path(toto, sprintf("done.%d", time())));
  }
}

int summary_status()
{
  if(rpc && rpc->num_connections())
    return 1;
  return 0;
}

void check_friendly_lock()
{
  summary_status();
  remove_call_out(check_friendly_lock);
  call_out(check_friendly_lock, 120);
}

void summary_schedule(int|void go, int|void override);
int start_summarizing(int|void be_friendly)
{
  MEGALS
  
  if(summary_status()) {
    summary_schedule();
    return 0;
  }

  if(!query("be-friendly"))
    be_friendly = 0;
  
  if(be_friendly) {
    if(catch { update_friendly_lock = UpdateLock(); }) {
      int delay_in_minutes = (int) (0.8*(update_every_hour?1:24)*60);
      if(friendly_process_policy_counter++ < delay_in_minutes &&
	 random(delay_in_minutes) > 0) {
	summary_schedule(60, 1);  // Try again in one minute.
	return 0;
      }
    }
  } else {
    check_queue(1);
    bundle();
  }
  friendly_process_policy_counter = 0;
  string adr = rpc->query_address();
  string host = (adr/" ")[0];
  int port = (int) (adr/" ")[1];
  // werror(sprintf("PORT: %O\n", port));
  gate->exit_msg();  // Clear exit msg buffer
  mixed spawn_res = spawn_pike(({ combine_path(getcwd(), program_directory()+
					       "/LogView.pmod/summary.pike"),
				  MIME.encode_base64(encode_value(({ host, port,
								     rpckey }))) }),
			       query("logpath"));
  if(objectp(spawn_res))
    summary_pid=spawn_res->pid();
  else if(intp(spawn_res))
    summary_pid=spawn_res;
  
  summary_schedule();
  return 1;
}

void summary_schedule(int|void go, int|void override)
{
  if(!override && !zero_type(find_call_out(start_summarizing))) {
    // werror(sprintf("(LogView) Queue full.\n"));
    return;
  }
  
  if(!go) {
    if(update_every_hour)
      go = (65-(localtime(time())->min%60))*60;
    else {
      int h;
      sscanf(query("update-hour"), "%d", h);
      mapping t = localtime(time());
      go = ((24+h-t->hour)%24)*60 - t->min;
      if(go < 0)
	go += 24*60;
      go = (go+5)*60;
    }
  }
  
  remove_call_out(start_summarizing);
  call_out(start_summarizing, go, 1);
  // werror(sprintf("(LogView) Go in approx. %02d:%02d.\n", go/3600, (go/60)%60));
}

/*** ***/

string html_simple_select(string name, string dft, array options)
{
  string r = "<select name="+name+">\n";
  foreach(options, string option)
    r += "<option name='"+option+"'"+(dft==option?" selected":"")+">"+
         option+"\n";
  return r+"</select>";
}

string|mapping html_eventlog(array(string) args, object id, string base)
{
  string r = "";
  r += "<form method=get>\n";

  string selected = id->variables->event||"Events";
  r += "<b>Log type</b>&nbsp;&nbsp;";
  r += html_simple_select("event", selected, ({"Admin log","Summary log"}));
  r += "&nbsp;&nbsp;<input type=submit name=select value=Select><p>";

  if(selected == "Summary log") {
    string debugfile = combine_path(combine_path(getcwd(), query("logpath")),
				    "summary.out");
    r += "<b>Debug log file at:</b> <i>"+debugfile+"</i><p>\n";
    r += "<pre>"+reverse((Stdio.read_bytes(debugfile)||
			  "<i>&lt;The debug log is empty&gt;</i>\n")/"\n")*
      "\n"+"</pre>";
  } else {
    r += "<input type=submit name=clear value='Clear Admin log'><br>";
    if(id->variables->clear == "Clear Admin log")
      eventlog->clear("Administrator cleared Event Log from "+
		      roxen->blocking_ip_to_host(id->remoteaddr)+".");
    r += eventlog->html()+"</form>\n";
  }
  return r;

#if 0
  r += "A debug log file will be created at:<p>&nbsp;&nbsp;&nbsp;&nbsp;"+
       debugfile+"&nbsp;&nbsp;&nbsp;&nbsp;<p>\n";
  r += "<b>Excerpt from the debug log file:</b><p><font size=-1><pre><ul>";
  array rows = (Stdio.read_bytes(debugfile)||
		"<i>&lt;The debug log is empty&gt;</i>\n")/"\n";
  r += rows[0..9]*"\n";
  rows = rows[10..];
  if(sizeof(rows) > 16)
    r += "\n\n</ul><i>&lt;truncated, last part of the debug log "
         "displayed below&gt;</i><ul>\n";
  r += reverse(reverse(rows)[0..15])*"\n";
  r += "</ul></pre></font>\n";
#endif

}

string html_report(array(string) args, object id, string base)
{
  mapping mc = gate->read_main_config();
  if(!mc->reports || !sizeof(mc->reports))
    return "<p><em>No reports are available.</em>\n";

  if(id->variables->select_report == "Select" ||
     (sizeof(args) && sizeof(args[0]) &&
      (int) args[0] < sizeof(mc->reports) &&
      0 <= (int) args[0])) {
    string r = "<form method=get><b>Report</b>&nbsp;&nbsp;";
    string report = id->variables->report ||
		    sort(indices(mc->reports))[(int)args[0]];
    r += html_simple_select("report", report,
			    sort(indices(mc->reports)));
    r += "<input type=submit name=select_report value=Select>";
    r += "</form>\n";
    
    return r + mc->reports[report];
  }

  return ("<p>Please select a report to view.<p><ul>"+
	  Array.map(sort(indices(mc->reports)),
		    lambda(string name, string base, mapping mc, mapping m)
		    {
		      return ("<dt><font size='+2'>"
			      "<a href="+base+"report/"+(string)(m->c++)+">"
			      +name+"</a>"
			      "</font>"
			      "<dd>"+((mc->report_descriptions &&
				       mc->report_descriptions[name]) || ""));
		    }, base, mc, ([]))*"\n"+
	  "</ul>");
}

string logview_table(array table, array fields, array types, int min)
{
  string r = "";

  r += ("<tablify nice modulo=3 cellseparator='\t' rowseparator='\n' "
	"fgcolor1=#dddddd>"
	"<fields>"+({ "num", @types })*"\t"+"</fields>"+
	({ "#", @fields })*"\t"+"\n");
  array new_table = ({});
  for(int n = 0; n < sizeof(table); n++) {
    array row = copy_value(table[n]);
    for(int i = 0; i < sizeof(table[n]); i++) {
      row[i] = ((string)row[i]-"<")-">";
      switch(lower_case(i<sizeof(types)?types[i]:"text")) {
      case "delta url":
	if(!n || (table[n-1][i] != row[i]))
	  row[i] = "<a href='"+row[i]+"'>"+row[i]+"</a>";
	else
	  row[i] = "";
	break;
      case "page url":
	array a = row[i]/"\0";
	row[i] = "<a href='"+(a*"")+"'>"+(sizeof(a)>1?a[1]:a[0])+"</a>";
	break;
      case "host url":
	row[i] = "<a href='http://"+row[i]+"'>"+row[i]+"</a>";
	break;
      case "url":
	row[i] = "<a href='"+row[i]+"'>"+row[i]+"</a>";
	break;
      case "urlmaybe":
	int c = row[i][0];
	row[i] = row[i][1..];
	if(c == '!')
	  row[i] = "<a href='"+row[i]+"'>"+row[i]+"</a>";
	else
	  row[i] = (string) row[i];
	break;
      }
      row[i] -= "\t";
      row[i] -= "\n";
    }
    new_table += ({ ({ sprintf("%d", min+n+1), @row })*"\t" });
  }
  r += new_table*"\n"+"</tablify>";
  return r;
}

string logview_export(array table, array fields, array types, int min)
{
  string r = fields*"\t"+"\n";
  object t = LogView.Table.table(table, fields);
  
  for(int i = 0; i < sizeof(fields); i++) {
    switch(i<sizeof(types)?types[i]:"text") {
    case "urlmaybe":
      t = t->map_col(lambda(string s) { return s[0]=='!'?s[1..]:s; }, i);
      break;
    }
    t = t->map_col(lambda(mixed x)
		   { return replace((string)x,
				    ({ "\n", "\t", "\0" }),
				    ({ "", "", "" })); }, i);
  }
  r += Array.map(values(t), lambda(array r) { return r*"\t"; })*"\n";
  return r;
}

string logview_ascii(array table, array fields, array types, int min)
{
  string r = "";
  object t = LogView.Table.table(table, fields);
  
  for(int i = 0; i < sizeof(fields); i++) {
    switch(i<sizeof(types)?types[i]:"text") {
    case "urlmaybe":
      t = t->map_col(lambda(string s) { return s[0]=='!'?s[1..]:s; }, i);
      break;
    }
    t = t->map_col(lambda(mixed x)
		   { return replace((string)x,
				    ({ "\n", "\t", "\0" }),
				    ({ "", "", "" })); }, i);
  }

  array s = ({}), a = ({ indices(t) }) + values(t);

  fields = ({ "#" }) + fields;
  types = ({ "num" }) + types;
  for(int j = 0; j < sizeof(a); j++)
    a[j] = (j?({ (string)j }):({ "#" })) + a[j];

  string l = "";
  for(int i = 0; i < sizeof(fields); i++) {
    int max = 0;
    for(int j = 0; j < sizeof(a); j++)
      max = max>sizeof(a[j][i])?max:sizeof(a[j][i]);
    s += ({ max });
    l += "-"+(i?"-":"");
    for(int j = 0; j < max; j++)
      l += "-";
    l += "-";
  }
  l += "\n";
  array q = ({});
  for(int j = 0; j < sizeof(a); j++) {
    string tmp = "";
    for(int i = 0; i < sizeof(s); i++)
      tmp += sprintf((i?" ":"")+" %"+
		     (j&&i<sizeof(types)&&types[i]=="num"?"":"-")+
		     (string)s[i]+"s ", a[j][i]);
    q += ({ tmp });
  }

  r = q[0]+"\n"+l+q[1..]*"\n"+(sizeof(q)>1?"\n":"")+l;

  // r += Array.map(values(t), lambda(array r) { return r*"\t"; })*"\n";
  return r;
}

array(string) html_select_group(string group, object id,
				array(string)|void groups)
{
  string r = "";
  multiset m = mkmultiset(gate->list_groups());
  if(!sizeof(m))
    return ({ "", "<p><em>No statistics groups are available.</em>" });
  if(sizeof(m) == 1)
    return ({ indices(m)[0], indices(m)[0]+"\n" });
  if(id->variables->select_group)
    group = m[id->variables->select_group]?id->variables->select_group:0;
  if(!group)
    group = sort(indices(m))[0];
  if(sizeof(m) > 1) {
    r += "<table border=0 cellspacing=0 cellpadding=0><tr><td>\n";
    r += "<select name=select_group>";
    foreach(sort(indices(m)), string name) {
      if(name == group)
	r += "<option selected>"+name+"\n";
      else
	r += "<option>"+name+"\n";
    }
    r += "</select>";
    r += "<input name=group type=submit value=Select>";
    r += "</td></tr></table>\n";
  }
  return ({ group, r });
}

array(array(int)) list_table_dates(array(string) tables)
{
  // Reduce duplicate dates
  mapping(string:int) has_data_dates = ([]);
  foreach(tables, string s)
    has_data_dates[(s/":")[0]] = 1;
  array(array(int)) has_data_dates = Array.map(sort(indices(has_data_dates)),
			lambda(string s) {
                            array(int) ymd = allocate(3);
			    sscanf(s, "%d-%d-%d", ymd[0], ymd[1], ymd[2]);
			    return ymd; });
  return has_data_dates;
}

string html_period_of_time(string name, array years,
			   int year, int month, int day, int hour)
{
  string r = "";
  r += "<td><select name='"+name+"_year'>\n";
  r += Array.map(years,
		 lambda(int y, int year)
		 { return sprintf("<option%s>%d",
			  y==year?" selected":"", y);}, year)*"\n";
  r += "</select></td>\n";
  r += "<td><select name='"+name+"_month'>\n";
  for(int m = 0; m < 12; m++)
    r += "<option"+(m+1==month?" selected":"")+">"+
         LogView.Calendar.month_names[m]+"\n";
  r += "</select></td>\n";
  r += "<td><select name='"+name+"_day'>\n";
  for(int d = 1; d <= 31; d++)
    r += "<option name="+d+(d==day?" selected":"")+">"+d+"\n";
  r += "</select></td>\n";
#if UTILIZATION
  r += "<td><select name='"+name+"_hour'>\n";
  for(int h = 0; h <= 23; h++)
    r += "<option name="+h+(h==hour?" selected":"")+sprintf(">%02d\n",h);
  r += "</select></td>\n";
#endif
  return r;
}

array get_ymdh(string name, mapping vars)
{
  array ymdh = ({ 0,0,0,0 });
  if(!sscanf(vars[name+"_year"]||"", "%d", ymdh[0]) ||
     !sscanf(vars[name+"_month"]||"", "%s", ymdh[1]) ||
     !sscanf(vars[name+"_day"]||"", "%d", ymdh[2])
#if UTILIZATION
     || !sscanf(vars[name+"_hour"]||"", "%d", ymdh[3])
#endif
     )
    return 0;
  ymdh[1] = mkmapping(LogView.Calendar.month_names,
		      LogView.Calendar.monthsofyear(ymdh[0]))[ymdh[1]];
  return ymdh;
}

string|mapping html_advanced(array(string) args, object id, string base)
{
  string r = "";
  string focus_redirect=combine_path(query_location_low(id),"focus")+base+args*"/";

  array(string) a = html_select_group(0, id);
  if(!a[0] || !sizeof(a[0])) return a[1];
  string group = a[0], groupselect = a[1];

  r += "<form method=get><table border=0>\n";

  r += "<tr><th align=left valign=center>Group</th><td colspan=3>"+
       groupselect+"</td></tr>\n";

  // Group
  object safe_vue = gate->safe_vue(group);
  if(!safe_vue) return r+"<b>Glm det grabben!</b><p>\n";
  mapping gc = gate->read_group_config(group);
  if(!safe_vue || !gc->fugroups || !sizeof(gc->fugroups))
    return r+"</form><p><em>No items are logged.</em><p>\n";

  string per = id->variables->per?id->variables->per:"Day";
  string granul = id->variables->granul?id->variables->granul:"Day";
  string pres = id->variables->pres?id->variables->pres:"Append";

  array(mixed) has_data_list=safe_vue->table("global")->get("calendar")||({});
  if(!has_data_list || !sizeof(has_data_list))
    return r + "</table></form><p><em>No items are logged.</em><p>";
  
  // Report
  string report = id->variables->report||"#unknown";
  r += "<tr><th align=left>Report</th><td colspan=3>";
  r += "<select name=report>\n";
  foreach(gc->fugroups, string fugroup) {
    foreach(gate->list_fus(fugroup), string name) {
      r += ("<option name='"+name+"'"+(report==name?" selected":"")+">"+
	    name+"\n");
    }
  }
  r += "</select><b>&nbsp;&nbsp;as&nbsp;&nbsp;</b>";
  r += html_simple_select("granul", granul, ({"Year","Month","Week",
					      "Day","Hour"
#if UTILIZATION
					      ,"Minute"
#endif
}));
  r += "</select><b>&nbsp;&nbsp;per&nbsp;&nbsp;</b>";
  r += html_simple_select("per", per, ({"Year","Month","Week",
					"Day","Hour"
#if UTILIZATION
					      ,"Minute"
#endif
}));
  r += "</td></tr>\n";

  int max_rows;
  sscanf(id->variables->max_rows||"100", "%d", max_rows);
  r += "<tr><th align=left>Presentation</th><td colspan=3>"+
       html_simple_select("pres", pres, ({"Append","Sum"}))+
       "<b>&nbsp;max&nbsp;&nbsp;</b>"
       "<input type=text name=max_rows size=5 value="+max_rows+">"
       "<b>&nbsp;rows</b>"
       "</td></tr>\n";

  pres = lower_case(pres);
  per = lower_case(per);
  granul = lower_case(granul);

  // Period of time
  r += ("<tr><td></td><th align=left>Year</th>"
	"<th align=left>Month</th><th align=left>Day</th>"
#if UTILIZATION
	"<th>Hour</th>"
#endif
	"</tr>\n");
  has_data_list = glob("day.*", has_data_list);
  multiset years = (<>);
  for(int i = 0; i < sizeof(has_data_list); i++) {
    int y,m,d;
    sscanf(has_data_list[i][4..], "%d_%d_%d", y,m,d);
    // has_data_list[i] = ({ y,m,d });
    if(!years[y]) years[y] = 1;
  }
  array years = sort(indices(years));
  array from_ymdh = get_ymdh("from", id->variables)||({years[-1], 1, 1, 0});
  array to_ymdh = get_ymdh("to", id->variables)||({years[-1], 12, 31, 23});
  r += "<tr><th align=left>From</th>"+
       html_period_of_time("from", years, @from_ymdh)+"</tr>\n";
  r += "<tr><th align=left>To</th>"+
       html_period_of_time("to", years, @to_ymdh)+"</tr>\n";

  r += ("<tr><th align=left>Display</th><td colspan=3>"

	"<table border=0 cellspacing=0 cellpadding=0><tr>"
	"<td><input type=submit name=display value='Table'></td>"
	"<td><input type=submit name=display value='Line chart'></td>"
	"<td><input type=submit name=display value='Bar chart'></td>"
	"<td><input type=submit name=display value='Sum bars'></td>"
	// "</tr></table></td></tr>\n"
	// "<tr><th align=left></th><td colspan=3>"
	// "<table border=0 cellspacing=0 cellpadding=0><tr>"
	"<td><input type=submit name=display value='2D pie'></td>"
	"<td><input type=submit name=display value='3D pie'></td>"
	"<td><input type=submit name=display value='Map'></td>"
	"<td><input type=submit name=display value='Ascii'></td>"
	"<td><input type=submit name=display value='Export'></td>"
	"<td><input type=submit name=display value='Tag'></td>"
	"</tr></table>");
  r += "</td></tr>\n";

  mapping display_options = ([ "Table":"table",
			       "Line chart":"line-chart",
			       "Bar chart":"bar-chart",
			       "Sum bars":"sum-bars",
			       "2D pie":"pie-chart",
			       "3D pie":"3d-pie-chart",
			       "Map":"world-map",
			       "Ascii":"ascii" ]);

  r += ("</table>\n<input type=hidden name=last_display value="+
	(display_options[id->variables->display] ||
	 id->variables->last_display || "table")+"></form>\n<p>\n");

  // The statistics
  r += "<a name=focus>\n";

  if(id->variables->display) {
    string s = "";
    s += ("<logview  group='"+group+"'  report='"+report+"'"
	  "  unit="+granul+"  per="+per+"\n"
	  " op="+pres+"  max="+max_rows+"\n"
	  " from-year="+from_ymdh[0]+"\n"
	  " from-month="+from_ymdh[1]+"\n"
	  " from-day="+from_ymdh[2]+"\n"
	  " to-year="+to_ymdh[0]+"\n"
	  " to-month="+to_ymdh[1]+"\n"
	  " to-day="+to_ymdh[2]+"\n"
#if UTILIZATION
	  " from-hour="+from_ymdh[3]+"\n"
	  " to-hour="+to_ymdh[3]+"\n"
#endif
	  " ");

    string hs = "", he = "";
    switch(id->variables->display) {
    case "Export":
      s += "display=export>\n";
      return ([ "data":parse_rxml(s, id),
		"extra_heads":(["Content-Disposition":
				"attachment; filename="+report]),
		"type":"application/x-excel" ]);
    case "Tag":
      s += "display="+(id->variables->last_display||"table");
      break;
    case "Ascii":
      hs = "<pre>";
      he = "</pre>";
    default:
      s += "display="+(display_options[id->variables->display]||"table");
      break;
    }
    s += ">\n";
    if(id->variables->display == "Tag")
      r += "<pre>"+replace(s, ({ "\n", "<", ">" }), 
			   ({ "<br>", "&lt;", "&gt;" })) + "</pre>";
    else
      r += hs+s+he;
  }

  return r;
}

int within_boundry(array check_ymd, int from_jd, int to_jd)
{
  int check_jd = LogView.Calendar.ymd2jd(@check_ymd);
  return from_jd <= check_jd && check_jd <= to_jd;
}

string granuality(int y, int m, int d, string granul, string|void rel)
{
  switch(granul) {
  case "year":
    return sprintf("%04d", y);
  case "month":
    return rel?sprintf("%02d", m):sprintf("%04d\0%02d", y,m);
  case "week":
    int wk = LogView.Calendar.week(y,m,d);
    return (rel?sprintf("%02d", LogView.Calendar.week(y,m,d)):
	    sprintf("%04d\0%02d", LogView.Calendar.
		    jd2ymd((LogView.Calendar.ymd2jd(y,m,d)/7)*7+
			   (wk==1?6:0))[0], wk));
  case "day":
  case "hour":
  case "minute":
  default:
    if(granul == "day" && rel == "week")
      return (string) LogView.Calendar.weekday(y,m,d);
    return rel?sprintf("%02d", d):sprintf("%04d\0%02d\0%02d", y,m,d);
  }
}

array delta_year(int y, array ymd)
{
  return ({ ymd[0]+y, @ymd[1..] });
}

array delta_month(int m, array ymd)
{
  ymd[0] += (m+ymd[1]-1)/12;
  ymd[1] = ((m+ymd[1]-1+12)%12) + 1;
  return ymd;
}

array delta_day(int d, array ymd)
{
  return LogView.Calendar.jd2ymd(LogView.Calendar.ymd2jd(@ymd)+d);
}

array convert_time(string t, int p, array ymd)
{
  int x;
  string s, u;

  mapping days = ([ "monday":0, "mon":0,
		    "tuesday":1, "tue":1,
		    "wednesday":2, "wed":2,
		    "thursday":3, "thu":3,
		    "friday":4, "fri":4,
		    "saturday":5, "sat":5,
		    "sunday":6, "sun":6 ]);
  mapping months = ([ "january":1, "jan":1,
		      "february":2, "feb":2,
		      "march":3, "mar":3,
		      "april":4, "apr":4,
		      "may":5,
		      "june":6, "jun":6,
		      "july":7, "jul":7,
		      "august":8, "aug":8,
		      "september":9, "sep":9,
		      "october":10, "oct":10,
		      "november":11, "nov":11,
		      "december":12, "dec":12 ]);

  array f = ({ delta_year, delta_month, delta_day });

  t = t?t:"0";
  if(sscanf(t, "today%s", s)) t = "this-day"+s;
  if(sscanf(t, "yesterday%s", s)) t = "last-day"+s;
  if(sscanf(t, "tomorrow%s", s)) t = "next-day"+s;
  if(sscanf(t, "%s-%s%d", s, u, x) >= 2 && sizeof(s)) {
    u = lower_case(u-"+");
    if(s == "next") {
      if(!zero_type(days[u]))
	x += 7 + days[u] - LogView.Calendar.weekday(@ymd);
      else
	x++;
    } else if(s == "last" || s == "previous") {
      if(!zero_type(days[u]))
	x += -7 + days[u] - LogView.Calendar.weekday(@ymd);
      else
	x--;
    } else if(!zero_type(days[u]))
      x += days[u] - LogView.Calendar.weekday(@ymd);
    return f[p](x, ymd);
  }
  
  int x;
  if(sscanf(t, "%s", s) && (s = lower_case(s)) && months[s])
    x = months[s];
  else
    sscanf(t, "%d", x);

  if(x == 0 || t[0] == '-' || t[0] == '+')
    return f[p](x, ymd);
  ymd[p] = x;
  return ymd;
}

array tag_time(string y, string m, string d)
{
  mapping t = localtime(time());
  array ymd = ({ t->year+1900, t->mon+1, t->mday });
  ymd = convert_time(y, 0, ymd);
  ymd = convert_time(m, 1, ymd);
  ymd = convert_time(d, 2, ymd);
  return ymd;
}

#define TAGERR(msg) sprintf("<b>LogView Error</b>: %s<br>", (msg))

int hr_time, hr_acc, hr_na;
void hr_start()
{
#if DEBUG_TIMING
  hr_acc = 0;
  hr_time = gethrtime();
#endif
}

void hr_read(string|void s)
{
#if DEBUG_TIMING
  hr_acc += gethrtime() - hr_time;
  werror(sprintf("# %s %d sec\n", s?s+":":"", hr_acc));
#endif
}

void hr_stop()
{
  // hr_na = gethrtime();
}

void hr_cont()
{
  // hr_acc -= gethrtime() - hr_na;
}

string compose(array a, array|void x, int|void omit)
{
  if(!sizeof(a))
    return "-";
  if(x && equal(a[0..sizeof(a)-2], x[0..sizeof(x)-2]))
    return omit?" ":a[-1];
  string s = sizeof(a)?a[0]:" ";
  array x = ({ "-", "-", ":", ":", ":" });
  foreach(a[1..], string q) {
    s += x[0]+q;
    x = x[1..];
  }
  return s;
}

string tag_logview_doc()
{
  return replace((Stdio.read_file(mydisklocation+
				  "LogView.pmod/logview_tag.doc")),
		 ({ "{", "}" }), ({ "&lt;", "&gt;" }));
}

string manual_ignore(string name, mapping args, string contents, object id)
{
  return contents;
}

string read_manual(array(string) args, object id)
{
  args -= ({ ".." });
  args += !sizeof(args)||args[-1]==""||args[-1]=="."?({ "logview.html" }):({});
  string s = Stdio.read_bytes(mydisklocation+"/manual/"+args*"/");
  if(s && glob("*.html", args[-1]))
    s = replace(parse_html(s, ([]),
			   ([ "html":manual_ignore,
			      "head":lambda() { return ""; },
			      "body":manual_ignore ])),
		"img/", query_location_low(id)+"manual/img/");
  return s||"#";
}

string encode_args_key(mapping m)
{
  array a = indices(m);
  array b = values(m);
  sort(a,b);
  return encode_value(({ a,b }));
}

int world_id = 0, world_time = time();
mapping primary_worlds = ([]), secondary_worlds = ([]);
string add_world(object img)
{
  string key = sprintf("world%08x%08x%x.gif", world_time,
		       random(2*1024*1024-1), world_id++);
  if(sizeof(primary_worlds) >= 20) {
    secondary_worlds = primary_worlds;
    primary_worlds = ([]);
  }

  primary_worlds[key] = Image.GIF.encode( img );
  return key;
}
string get_world(string key)
{
  return primary_worlds[key] || secondary_worlds[key];
}

int correct_number(mixed x)
{
  if(stringp(x))
    sscanf(x, "%d", x);
  return x;
}

mixed tag_logview(string tag_name, mapping args, object id,
		  object file, mapping defines)
{
  kludge_start(0, id->conf);
  MEGALS

  string r = "";
  if(!id->misc->logview_portal && !query("enable_logview_tag"))
    r += "The <b>&lt;logview&gt;</b> tag is not enabled.<p>";

  if(args->manual)
    return r+read_manual(({ "logview.html" }), id);

  if(args->help)
    return r+tag_logview_doc();

  if(sizeof(r))
    return r;

  // Parse arguments
  args->group = args["group"]||
		((sizeof(gate->list_groups()) && gate->list_groups()) ||
		 ({"#undefined"}))[0];
  args->report = lower_case(args["report"]||"#undefined");
  args->per = lower_case(args["per"]||"Month");
  args->unit = lower_case(args["unit"]||args->per);
  args->op = lower_case(args["op"]||"Append");
  
  array from_ymd = tag_time(args["from-year"], 
			    args["from-month"],
			    args["from-day"]);
  array to_ymd = tag_time(args["to-year"],
			  args["to-month"],
			  args["to-day"]);
  array a = ({ "year", "month", "day" });
  for(int i = 0; i < sizeof(a); i++) {
    args["from-"+a[i]] = (string) from_ymd[i];
    args["to-"+a[i]] = (string) to_ymd[i];
  }
  args["max"] = args["max"]||"100";

#if UTILIZATION
  int from_hour = correct_number(args["from-hour"]||"0");
  int to_hour = correct_number(args["to-hour"]||"23");
#endif

  args->display = lower_case(args["display"]||"table");
  if(id->misc->logview_email)
    args->display = "ascii";
  
  if(args["list-groups"])
    return gate->list_groups()*"<br>";

  if(!mkmultiset(gate->list_groups())[args->group])
    return TAGERR("group ``"+(args->group||"#unknown")+"'' not definied");
  mapping gc = gate->read_group_config(args->group);

  if(args["list-reports"]) {
    foreach(gc->fugroups, string fugroup)
      r += gate->list_fus(fugroup)*"<br>"+"<br>";
    return r;
  }

  if(args["list-defaults"])
    return ("<pre>\n"+
	    Array.map(sort(indices(args)) - ({ "list-defaults" }),
		      lambda(string s, mapping args)
		      { return sprintf("%O = %O", s, args[s]); }, args)*"\n"+
		      "\n</pre>\n");

  string args_key = encode_args_key(args);
  {
    string r;
    if(r = cache_check(args_key))
      return r;
  }

  if(args->display == "world-map") {
    args->unit = args->per = "day";
    args->op = "sum";
  }
  
  // Recalculate arguments
  int min_rows, max_rows;
  sscanf(args["min"]||"0", "%d", min_rows);
  sscanf(args["max"]||"100", "%d", max_rows);
  args->report = gate->find_fu(args->report, "day");
  if(!args->report)
    return TAGERR("report not defined");
  int from_jd = LogView.Calendar.ymd2jd(@from_ymd);
  int to_jd = LogView.Calendar.ymd2jd(@to_ymd);
  
  // Compile statistics
  object safe_vue = gate->safe_vue(args->group);
  array calendar = sort(glob("day.*", safe_vue->table("global")->
			     get("calendar")||({})));
  mapping fum = fu->load_fu(args->report);
  if(!fum)
    return TAGERR("report not ``"+args->report+"'' defined");

  mapping date = ([ "year":"Year", "month":"Year\0Month",
		    "week":"Year\0Week", "day":"Year\0Month\0Day",
		    "hour":"Year\0Month\0Day",
		    "minute":"Year\0Month\0Day" ]);
  mapping shortdate = ([ "year":"Year", "month":"Month",
                         "week":"Week", "day":"Day", "hour":"Hour",
			 "minute":"Minute" ]);
  mapping levels = ([ "year":0, "month":1,
		      "week":2, "day":3, "hour":4,
		      "minute":5 ]);
  if(levels[args->unit] < levels[args->per])
    return TAGERR("<i>unit</i> cannot be greater than <i>per</i>");    
  int hour_mode, minute_mode;
  if(lower_case(args->unit) == "hour") {
    if(lower_case(args->per) == "hour") {
      hour_mode = (args->op != "sum");
      args->op = "sum";
      args->unit = args->per = "day";
    } else {
      hour_mode = 1;
      args->unit = args->per;
    }
  }
  if(lower_case(args->unit) == "minute") {
    if(lower_case(args->per) == "minute") {
      minute_mode = (args->op != "sum");
      args->op = "sum";
      args->unit = args->per = "day";
    } else if(lower_case(args->per) == "hour") {
      minute_mode = 1;
      hour_mode = (args->op != "sum");
      args->op = "sum";
      args->unit = args->per = "day";
    } else {
      hour_mode = 1;
      minute_mode = 1;
      args->unit = args->per;
    }
  }
  
  // FIX! Absolute week problems.
  if(lower_case(args->op) != "sum" &&
     lower_case(args->unit) == "week" &&
     lower_case(args->per) == "year")
    args->per = "week";
  //
  
  array fields = ({ date[lower_case(args->per)],
		    shortdate[lower_case(args->unit)] });
  array types = ({ "text", "text" });

  if(args->unit == args->per) {
    fields = fields[0..0] + fields[2..];
    types = types[0..0] + types[2..];
  }
  if(args->op == "sum") {
    fields = fields[1..];
    types = types[1..];
  }
  fields = sizeof(fields)?fields[0]/"\0" + fields[1..]:({});
  types = Array.map(fields, lambda() { return "text";});

  hr_start();

  // Create memory optimal table combination scheme
  string old_tp;
  mapping dm, tp_map = ([]);
  array scheme = ({}), index = ({});
  multiset tablenames = mkmultiset(indices(safe_vue->table(args->report)));
  foreach(calendar, string entry) {
    int y,m,d;
    sscanf(entry, "day.%d_%d_%d", y,m,d);
    if(within_boundry(({ y,m,d }), from_jd, to_jd) && tablenames[entry]) {
      string tp = granuality(y,m,d, args->unit, args->per);
      string per = granuality(y,m,d, args->per);
      array a;
      if(args->op == "sum")
	a = (args->unit == args->per?({ }):({ tp }));
      else
	a = (args->unit == args->per?({ per }):({ per, tp }));
      
      if(tp == old_tp || (args->op == "sum" && args->unit == args->per)) {
	if(sizeof(scheme))
	  scheme[-1] += ({ ({ entry, @a }) });
	else
	  scheme += ({ ({ ({ entry, @a }) }) });
      } else {
	if(args->op == "sum") {
	  if(zero_type(tp_map[tp])) {
	    tp_map[tp] = sizeof(tp_map);
	    scheme += ({ ({ ({ entry, @a }) }) });
	    index += ({ tp });
	  } else
	    scheme[tp_map[tp]] += ({ ({ entry, @a }) });
	  tp = 0;
	} else
	  scheme += ({ ({ ({ entry, @a }) }) });
      }
      old_tp = tp;
    }
  }
  if(sizeof(index))
    sort(index, scheme);

  hr_read("<logview compiling scheme>");
  hr_start();

  // Query Optimizer
  for(int i = 0; i < sizeof(scheme); i++) {
    int y, m;
    mapping q = ([]);
    foreach(scheme[i], array a)
      if(sscanf(a[0], "day.%d_%d_%*s", y, m) == 3)
	q[sprintf("day.%04d_%02d_*", y, m)] = ({ y, m, a[1..] });
    
    foreach(sort(indices(q)), string g) {
      if(sizeof(glob(g, calendar) - column(scheme[i], 0)) == 0) {
	string s = sprintf("month.%04d_%02d", @q[g][0..1]);
	if(tablenames[s]) {
	  // werror(sprintf("#OPTIMIZING "+s+"\n"));
	  mapping r = mkmapping(column(scheme[i], 0), scheme[i]);
	  scheme[i] = (rows(r, indices(r) - glob(g, column(scheme[i], 0))) +
		       ({ ({ s, sizeof(glob(g, calendar)), @q[g][2] }) }));
	} else {
	  // werror(sprintf("#UNABLE TO OPTIMIZE "+s+"\n"));
	}
      }
    }
  }

  // Table rebuild
  object rep = fu->query_fu(fum, gate->read_group_config(args->group)->
			    group_args);
  int init_fields;
  array squeeze = ({}), table = ({});
  function display = rep->display, store = rep->store;
  function get = safe_vue->table(args->report)->get;
  if(!hour_mode)
    squeeze += ({ "hour" });
#if UTILIZATION
  if(!minute_mode)
    squeeze += ({ "minute" });
#endif
  if(!sizeof(squeeze))
    squeeze = 0;
  array table = ({});
  // werror(sprintf("############\n"));
  foreach(scheme, array a) {
    // werror(sprintf("------------ [%d entries]\n", sizeof(a)));
    mixed data;
    array entry;
    int ndays = 0;
    foreach(a, entry) {
      if(glob("month*", entry[0])) {
	ndays += entry[1];
	entry = ({ entry[0] }) + entry[2..];
      } else
	ndays += 1;
      // werror(sprintf("entry: %s\n", entry[0]));
      data = store(data?rep->join(data, get(entry[0])):get(entry[0]));
    }
    // werror(sprintf("------------\n"));
    // werror(sprintf("Parsing table\n"));

    function hf;
#if UTILIZATION
    hf = compile_string(sprintf("int hf(int h) { return h >= %d && h <= %d; }", from_hour, to_hour))()->hf;
#endif

    mapping m = display(([ "ndays":ndays ]), data, squeeze, hf);
    if(!init_fields) {
      fields += m->fields;
      types += m->types;
      init_fields = 1;
    }

    // Change weekday numbers to weekday names
    if(args->unit == "day" && args->per == "week")
      entry[-1] = LogView.Calendar.sday_names[correct_number(entry[-1])];
    // Change month numbers to month names
    if(args->unit == "month" && args->per != "month")
      entry[-1] = LogView.Calendar.smonth_names[correct_number(entry[-1])-1];
    if((args->per == "month") && args->op != "sum") {
      int y, m, i = (args->per==args->unit)?-1:-2;
      sscanf(entry[i], "%d\0%d", y,m);
      entry[i] = sprintf("%04d\0%s", y,LogView.Calendar.smonth_names[m-1]);
    }
    if((args->per == "day") && args->op != "sum") {
      int y, m, d;
      sscanf(entry[-1], "%d\0%d\0%d", y,m,d);
      entry[-1] = sprintf("%04d\0%s\0%02d",
			y,LogView.Calendar.smonth_names[m-1],d);
    }
    array row = Array.map(entry[1..],lambda(string s){return s/"\0";})*({});

    foreach(m->table, array r)
      table += ({ row + r });

    if(sizeof(table) > min_rows+max_rows)
      break;
  }

  hr_read("<logview joining tables>");
  
  if(!sizeof(table))
    return r + "<em>No items are logged.</em><p>";

  int old_size = sizeof(table);
  table = table[min_rows..];
  if(max_rows)
    table = table[..(max_rows-1)];
  if(args->verbose)
    r += sprintf("<font size=-1>Total number of rows: %d</font><br>\n",
		 old_size);

  if(!sizeof(table))
    return r + "<em>No items are logged.</em><p>";

  switch(args->display) {
  case "sql":
  case "ascii":
  case "export":
    break;
  case "table":

    /***
     *** Compatibility
     ***
     ***
     ***/

    if(args->compatibility) {
      string sblock = args["compatibility-sblock"];
      string report = args->report;
      string base = args["compatibility-base"];
      string datebase = args["compatibility-datebase"];

      // Calculate navigation bar
      string nav = "";
      int block, n = 0, table_size = old_size;
      if(args->display == "table") {
	nav += "<font size='-1'>";
	if(table_size > TABLE_PAGE_SIZE) {
	
	  if(sblock == "all") {
	    nav += "&nbsp;Rows "+(block+1)+" - "+table_size+" of "+
	      table_size+"&nbsp;&nbsp;&nbsp;\n<b>"
	      "&nbsp;<a href='"+base+report+"/0/"+
	      datebase+"#focus'>first "+TABLE_PAGE_SIZE+"</a>"
	      "&nbsp;&nbsp;\n</b>";
	  } else {
	    block = (int) sblock;
	  
	    nav += "&nbsp;Rows "+(block+1)+" - "+
	      MIN(block+TABLE_PAGE_SIZE,table_size)+" of "+
	      table_size+"&nbsp;&nbsp;&nbsp;\n<b>&nbsp;";
	    if(block)
	      nav += "<a href='"+base+report+"/"+(block-TABLE_PAGE_SIZE)+"/"+
		datebase+"#focus'>prev "+TABLE_PAGE_SIZE+"</a>"
		"&nbsp;&nbsp;&nbsp;\n";
	    else
	      nav += "<a href='"+base+report+"/"+
		(table_size-((table_size%TABLE_PAGE_SIZE)||
			     TABLE_PAGE_SIZE))+"/"+
		datebase+"#focus'>last "+TABLE_PAGE_SIZE+"</a>"
		"&nbsp;&nbsp;&nbsp;\n";
	    if(block+TABLE_PAGE_SIZE < table_size)
	      nav += "<a href='"+base+report+"/"+(block+TABLE_PAGE_SIZE)+"/"+
		datebase+"#focus'>next "+TABLE_PAGE_SIZE+"</a>";
	    else
	      nav += "<a href='"+base+report+"/0/"+
		datebase+"#focus'>first "+TABLE_PAGE_SIZE+"</a>";
	    nav += "&nbsp;&nbsp;&nbsp;<a href='"+base+report+"/all/"+datebase+
	      "#focus'>all</a></b>\n";
	  }
	} else
	  nav += "&nbsp;Rows "+(block+1)+" - "+table_size+" of "+
	    table_size+"&nbsp;&nbsp;&nbsp;\n";
	nav += "</font><br>";
      }
      r += nav;
    }

    break;

    /***
     *** Compatibility
     ***
     ***
     ***/

    break;
  default:
    // Combine all non-numerical fields to one field only
    array textfields = ({});
    array nontextfields = ({});
    for(int i = 0; i < sizeof(types); i++)
      if(types[i] == "text")
	textfields += ({ i });
      else
	nontextfields += ({ i });
    fields = (({ compose(rows(fields, textfields)) }) +
	      rows(fields, nontextfields));
    
    array x, y;
    for(int i = 0; i < sizeof(table); i++) {
      table[i] = (({ compose(x = rows(table[i], textfields), y,
			     sizeof(table) > OMIT_FIELD) }) +
		  rows(table[i], nontextfields));
      y = x;
    }
  }

#if 0
  if(!(<"table", "line-chart", "bar-chart", "sum-bars">)[args->display])
    return TAGERR("unknown display type");
#endif

  switch(args->display) {
  case "sql":
    if(tag_name == "sql")
      return ([ "fields":fields, "types":types, "table":table ]);
  case "table":
    r += logview_table(table, fields, types, min_rows);
    break;
  case "export":
    r += logview_export(table, fields, types, min_rows);    
    break;
  case "world-map":
    if(fields[0] != "Country") {
      r += "The world map is only available with ``Most active countries''.";
    } else {
      int most_hits = 0;
      mapping m = ([]);
      object e = LogView.Map.Earth();
      object cs = LogView.Map.Legend()->scheme("white-to-red");
      foreach(table, array row) {
	string s = reverse_country_codes[row[0]];
	m[e->domain_to_country[s]] += row[-1];
      }
      foreach(indices(m), string s)
	most_hits = max(most_hits, m[s]);
      foreach(indices(m), string s)
	m[s] = (float)m[s]/(float)(most_hits||1);

      mapping opt = ([ "border":16,
		       "color_fu":lambda(string c, function cs, mapping m)
				  { return cs(m[c]||0.0, "white-to-red"); },
		       "fu_args":({ cs->color_scale, m }) ]);

      object map = e->image(600, 350, opt);
      object lgn = cs->image(175, 175,
			     ([ "border":20,
				"title":replace(fields[-1], " ", "\n"),
				"font":resolve_font("default"),
				"titles":({ max(most_hits, 10)/2,
					    max(most_hits, 10) }) ]));
      map->paste(lgn, 0, map->ysize() - lgn->ysize());
      r += "<img border=0 src="+add_world(map)+">\n";
    }
    break;
  case "ascii":
    r += logview_ascii(table, fields, types, min_rows);    
    break;
  default:
    int w = (sizeof(table)*13)+300;
    if(w < 300) w = 300;
    string extravaganza = "";
    switch(args->display) {
      case "3d-pie-chart":
	extravaganza = "3D=30";
      case "pie-chart":
	r += ("<diagram neng bgcolor=#ffffff vertgrid horgrid\n"
	      "notrans textcolor=#000000\n"
	      "width="+(string)PIE_CHART_SIZE+"\n"
	      "height="+(string)PIE_CHART_SIZE+"\n"
	      "tunedbox=#319cce,#319cce,#ffffff,#ffffff\n"
	      "gridcolor=#319cce linewidth=3.2 "
	      "type=piechart "+
	      (args->title?" name='"+args->title+"'":"")+"\n"
	      " "+extravaganza+">\n"
	      
	      "<colors>#8c5aa5	#009955</colors>\n"
	      "<legend orient=vertical>"+
	      (Array.map(column(table, 0),
			 lambda(mixed s) {return(string)s; })*"\t"+
			 "</legend>\n"));
	break;
    default:
      r += ("<diagram neng bgcolor=#ffffff vertgrid horgrid\n"
	    "notrans textcolor=#000000\n"
	    "width="+(string)w+" height=300\n"
	    "tunedbox=#319cce,#319cce,#ffffff,#ffffff\n"
	    "gridcolor=#319cce linewidth=3.2 "
	    "type="+(args->display-"-")+
	    (args->title?" name='"+args->title+"'":"")+"\n"
	    " "+extravaganza+">\n"
	    
	    "<colors>#8c5aa5	#009955</colors>\n"
	    "<xaxis quantity='"+fields[0]+"'>\n"+
	    "<yaxis quantity='"+(fields[1..]*", ")+"'>\n"
	    "<xnames orient=vertical>"+
	    (Array.map(column(table, 0),
		       lambda(mixed s) {return(string)s; })*"\t"+
		       "</xnames>\n")+
	    "<legend>"+fields[1..]*"\t"+"</legend>");
    }
    array a = ({});
    for(int i = 1; i < sizeof(table[0]); i++)
      a += ({ Array.map(column(table, i),
			lambda(mixed s) { return (string)s; } )*"\t" });
    r += "<data form=row>\n"+a*"\n"+"</data></diagram><br>\n";
    break;
  }

#if 0
  if(old_size != sizeof(table))
    r += "<b>Note! Data is truncated to "+max_rows+" entries.</b>";
#endif

  cache_insert(args_key, r);
  return r;
}

string|mapping html_view_function(array(string) args, object id, string base)
{
  string focus_redirect=combine_path(query_location_low(id),"focus")+base+args*"/";

  string group;
  if(args && sizeof(args) && sizeof(args[0]))
    group = args[0];
  args = args[1..];
  array(string) a = html_select_group(group, id);
  if(group != a[0])
    args = ({});
  group = a[0];
  string groupselect = a[1];
  if(!group || !sizeof(group))
    return groupselect;
  base = combine_path(base, group);
  
  string old_report, report, r = "";
  if(!args || !sizeof(args) || !sizeof(args[0]))
    report = "void";
  else
    old_report = report = args[0];
  args = args[1..];

  string sblock = "0";
  if(args && sizeof(args) && sizeof(args[0]))
    sblock = args[0];
  args = args[1..];

  string display_type = id->variables->display_type;

  base = combine_path(base, "");  // Fix last /
  r += "<form method=get><table border=0><tr><th>Group&nbsp;&nbsp;</th><td>"+
       groupselect+"</td></tr></table></form>\n<p>\n";
  
  object safe_vue = gate->safe_vue(group);
  if(!safe_vue) return r+"<b>Glm det grabben!</b><p>\n";
  mapping gc = gate->read_group_config(group);
  if(!safe_vue || !gc->fugroups || !sizeof(gc->fugroups))
    return r+"<em>No items are logged.</em><p>\n";

  array(mixed) has_data_list = safe_vue->table("global")->get("calendar");
  if(!has_data_list || !sizeof(has_data_list))
    return r + "<em>No items are logged.</em><p>";
  has_data_list = glob("day.*", has_data_list);

  int last_year;
  for(int i = 0; i < sizeof(has_data_list); i++) {
    int y,m,d;
    sscanf(has_data_list[i][4..], "%d_%d_%d", y,m,d);
    has_data_list[i] = ({ y,m,d });
    if(last_year < y)
      last_year = y;
  }
  if(report == "void") {
    report = "null";
    args = ({ "year", (string) last_year });
  }
  string freq = sizeof(args)?args[0]:0;

  string s = LogView.Calendar.html_calendar(args, base+report+"/0/",
					    has_data_list);
  if(!s)
    return r+"<em>No items are logged.</em><p>\n";
  r += s;
  
  // Build function group table
  string this_grp, this_title;
  foreach(gc->fugroups, string s)
    if(mkmultiset(gate->list_fufilenames(s))[report]) {
      this_grp = s;
      break;
    }
  mapping m = fu->load_fu(report);
  // if(m && m->vues && !mkmultiset(m->vues)[freq])
  // report = "null";
  string datebase = combine_path(args*"/", "");

  // Print legend of function names  
  int i = 0;
  string main_title;
  r += "<p><table bgcolor=#dddddd cellpadding=8 cellspacing=5>"
       "<tr align=left valign=top><td>\n";
  foreach(gc->fugroups, string groupname) {
    array groups = gate->list_fufilenames(groupname);
    if(sizeof(groups)) {
      if(i++) r += "<br>";
      r += "<b>"+groupname+"</b><br>\n";
      foreach(groups, string name) {
	mapping m = fu->load_fu(name);
	if(!m->invisible) {
	  // if((!m->vues || mkmultiset(m->vues)[freq]) && !m->invisible) {
	  if(report == "null" && this_grp == groupname && old_report &&
	     fu->load_fu(old_report)->title[0..1] == m->title[0..1])
	    report = name;
	  string title = (gate->true_names(m->title)/"(")[0];
	  if(name == report) {
	    this_title = group+", "+title;
	    main_title = title;
	    r += "&nbsp;<a href='"+base+"null/0/"+datebase+"'>"+
	         "<font color=#338844>"
	         "<nobr>"+title+"</nobr>"
	         "</font></a>&nbsp;&nbsp;&nbsp;<br>\n";
	  } else
	    r += "&nbsp;<a href='"+base+name+"/0/"+datebase+"#focus'>"+
	         "<nobr>"+title+"</nobr>"
	         "</a>&nbsp;&nbsp;&nbsp;<br>\n";
	}
      }
      if(i == (sizeof(gc->fugroups)+1)/2) {
	i = 0;
	r += "</td><td>";
      }
    }
  }
  r += "</td></tr></table><p>\n";

  if(!m)
    return r; //  + "<em>No items are logged.</em><p>";

  // Calculate title and statistics data table name
  string from_year, from_month, from_day;
  string to_year, to_month, to_day;
  string unit, per;
  string op = (mkmultiset(m->display_types)["line chart"]?"append":"sum");
  if(report != "null") {
    string h, tp = "";
    switch(freq) {
    case "year":
      h = args[1];
      tp = sprintf("year.%04d", correct_number(args[1]));
      from_year = sprintf("%04d", correct_number(args[1]));
      from_month = "1";
      from_day = "1";
      to_year = sprintf("%04d", correct_number(args[1]));
      to_month = "12";
      to_day = "31";
      unit = per = "month";
      break;
    case "month":
      h = LogView.Calendar.month_names[correct_number(args[2])-1]+", "+args[1];
      tp = sprintf("month.%04d_%02d", correct_number(args[1]), correct_number(args[2]));
      from_year = sprintf("%04d", correct_number(args[1]));
      from_month = sprintf("%02d", correct_number(args[2]));
      from_day = "1";
      to_year = sprintf("%04d", correct_number(args[1]));
      to_month = sprintf("%02d", correct_number(args[2]));
      to_day = sprintf("%02d",LogView.Calendar.
		       lastdayofmonth(correct_number(args[1]),
				      correct_number(args[2])));
      unit = per = "day";
      break;
    case "week":
      h = "Week "+args[3]+", "+args[1];
      tp = sprintf("week.%04d_%02d", correct_number(args[1]),
		   correct_number(args[3]));
      int jd = LogView.Calendar.week2jd(correct_number(args[1]),
					correct_number(args[3]));
      array from_ymd = LogView.Calendar.jd2ymd(jd);
      array to_ymd = LogView.Calendar.jd2ymd(jd+6);
      from_year = sprintf("%04d", correct_number(from_ymd[0]));
      from_month = sprintf("%02d", correct_number(from_ymd[1]));
      from_day = sprintf("%02d", correct_number(from_ymd[2]));
      to_year = sprintf("%04d", correct_number(to_ymd[0]));
      to_month = sprintf("%02d", correct_number(to_ymd[1]));
      to_day = sprintf("%02d", correct_number(to_ymd[2]));
      unit = (op=="sum"?"week":"day");
      per = "week";
      break;
    case "day":
      h = LogView.Calendar.day_names[LogView.Calendar.
				    weekday(correct_number(args[1]),
					    correct_number(args[2]),
					    correct_number(args[3]))]+
        ", "+LogView.Calendar.month_names[correct_number(args[2])-1]+" "+
        args[3]+", "+args[1];
      tp = sprintf("day.%04d_%02d_%02d",
		   correct_number(args[1]),
		   correct_number(args[2]),
		   correct_number(args[3]));
      from_year = sprintf("%04d", correct_number(args[1]));
      from_month = sprintf("%02d", correct_number(args[2]));
      from_day = sprintf("%02d", correct_number(args[3]));
      to_year = sprintf("%04d", correct_number(args[1]));
      to_month = sprintf("%02d", correct_number(args[2]));
      to_day = sprintf("%02d", correct_number(args[3]));
      unit = "hour"; per = "day";
      break;
    }
    
    // Load statistics function
    mapping m = fu->load_fu(report);

    // Set display type
    string default_type = "table";
    if(m->display_types)
      default_type = m->display_types[0];
    else
      m->display_types = ({});
    multiset displays = mkmultiset(m->display_types);
    if(!displays["table"])
      m->display_types = ({ "table" }) + m->display_types;
    if(!displays["export"])
      m->display_types = m->display_types + ({ "export" });
    if(!mkmultiset(m->display_types)[display_type])
      display_type = default_type;

    r += "<a name=focus>\n";

    // Load statistics data
    int table_size = 0;
    string display_options;
    array table, fields, types;
    mixed data = 0; // safe_vue->table(report)->get(tp);
    if(data) {
      data = fu->query_fu(m)->display(display_type, data);
      table = data->table;
      fields = data->fields;
      types = data->types;
      table_size = sizeof(table);
      if(data->diagram_type)
	display_type = data->diagram_type;
      if(data->diagram_options)
	display_options = data->diagram_options;
    }
#if 0
    if(!table_size)
      return r + "<em>No items are logged.</em><p>";
#endif

    // Print title
    r += "<form method=get action='"+focus_redirect+"'>";
    r += "<font size=+3><b>"+this_title+", "+h+" as </b></font>";
    r += "<select name=display_type>";
    // ({ "table", "bar chart", "line chart", "pie chart", "export" })
    foreach(m->display_types, string type)
      r += "<option"+(type==display_type?" selected":"")+">"+type;
    r += "</select><input type=submit value=Select><br>";

    // Calculate navigation bar
    string nav = "";
    int block, n = 0;
    if(display_type == "table") {
      nav += "<font size='-1'>";
      if(table_size > TABLE_PAGE_SIZE) {
	
	if(sblock == "all") {
	  nav += "&nbsp;Rows "+(block+1)+" - "+table_size+" of "+
	         table_size+"&nbsp;&nbsp;&nbsp;\n<b>"
	         "&nbsp;<a href='"+base+report+"/0/"+
		 datebase+"#focus'>first "+TABLE_PAGE_SIZE+"</a>"
		 "&nbsp;&nbsp;\n</b>";
	} else {
	  block = correct_number(sblock);
	  
	  nav += "&nbsp;Rows "+(block+1)+" - "+
	         MIN(block+TABLE_PAGE_SIZE,table_size)+" of "+
	         table_size+"&nbsp;&nbsp;&nbsp;\n<b>&nbsp;";
	  if(block)
	    nav += "<a href='"+base+report+"/"+(block-TABLE_PAGE_SIZE)+"/"+
	           datebase+"#focus'>prev "+TABLE_PAGE_SIZE+"</a>"
	           "&nbsp;&nbsp;&nbsp;\n";
	  else
	    nav += "<a href='"+base+report+"/"+
	           (table_size-((table_size%TABLE_PAGE_SIZE)||
				TABLE_PAGE_SIZE))+"/"+
	           datebase+"#focus'>last "+TABLE_PAGE_SIZE+"</a>"
	           "&nbsp;&nbsp;&nbsp;\n";
	  if(block+TABLE_PAGE_SIZE < table_size)
	    nav += "<a href='"+base+report+"/"+(block+TABLE_PAGE_SIZE)+"/"+
	           datebase+"#focus'>next "+TABLE_PAGE_SIZE+"</a>";
	  else
	    nav += "<a href='"+base+report+"/0/"+
	           datebase+"#focus'>first "+TABLE_PAGE_SIZE+"</a>";
	  nav += "&nbsp;&nbsp;&nbsp;<a href='"+base+report+"/all/"+datebase+
	         "#focus'>all</a></b>\n";
	}
      } else
	nav += "&nbsp;Rows "+(block+1)+" - "+table_size+" of "+
	       table_size+"&nbsp;&nbsp;&nbsp;\n";
      nav += "</font><br>";
    }

    if(display_type == "export")
      sblock = "all";

    string s = ("<logview  group='"+group+"'  report='"+main_title+"'"
		"  unit="+unit+"  per="+per+"\n"
		" display="+replace(display_type, " ", "-")+"\n"
		" op="+op+"\n"+
		(sblock=="all"?"":" min="+(string)((int)sblock))+"\n"+
		(sblock=="all"?"max=0":" max="+(string)TABLE_PAGE_SIZE)+"\n"
		" from-year="+from_year+"\n"
		" from-month="+from_month+"\n"
		" from-day="+from_day+"\n"
		" to-year="+to_year+"\n"
		" to-month="+to_month+"\n"
		" to-day="+to_day+"\n"
		" title='"+this_title+", "+h+"'\n"
		
		" compatibility\n"
		" compatibility-sblock='"+sblock+"'\n"
		" compatibility-base='"+base+"'\n"
		" compatibility-datebase='"+datebase+"'\n"
		">\n");

    if(display_type == "export")
      return ([ "data":parse_rxml(s, id),
		"extra_heads":(["Content-Disposition":
				"attachment; filename="+report]),
		"type":"application/x-excel" ]);
    r += s;
    
#if 0
    switch(display_type) {
    case "export":
      array a = ({}), b = ({});
      for(int i = 0; i < sizeof(table[0]); i++)
	if(types[i] != "hidden")
	  a += ({ ({ fields[i] }) +
		  Array.map(column(table, i),
			    lambda(mixed s) { return (string)s; } ) });
      for(int i = 0; i < sizeof(a[0]); i++)
	b += ({ column(a, i)*"\t" });
      return ([ "data":b*"\n",
		"extra_heads":(["Content-Disposition":
				"attachment; filename="+report+"."+tp]),
		"type":"application/x-excel" ]);
      break;
    case "pie chart":
    case "3D pie chart":
      r += ("<diagram bgcolor=#ffffff vertgrid horgrid width=550 height=250\n"
	    "notrans textcolor=#000000\n"
	    "type=piechart"+(display_type=="3D pie chart"?" 3D=30":"")+">\n");
      r += "<colors>#8c5aa5	#009955</colors>\n";
      r += "<legend>"+
	   Array.map(column(table, 0),
		     lambda(mixed s) {return (string)s;} )*"\t"+"</legend>\n";
      array a = ({});
      for(int i = 1; i < sizeof(table[0]); i++)
	if(types[i] != "hidden")
	  a += ({ Array.map(column(table, i),
			    lambda(mixed s) { return (string)s; } )*"\t" });
      r += "<data form=row>\n"+a*"\n"+"</data>";
      r += "</diagram>\n";
      break;
    case "bar chart":
    case "line chart":
      r += ("<diagram bgcolor=#ffffff vertgrid horgrid width=550 height=250\n"
	    "notrans textcolor=#000000\n"
	    "tunedbox=#319cce,#319cce,#ffffff,#ffffff\n"
	    "voidsep='[void]' neng gridcolor=#319cce linewidth=3.2 type="+
	    (display_type=="bar chart"?"bars":"linechart")+
	    (display_options?" "+display_options:"")+">\n");
      r += "<colors>#8c5aa5	#009955</colors>\n";
      r += "<xaxis quantity='"+fields[0]+"'>\n";
      r += "<yaxis quantity='"+this_title+", "+h+"'>\n";
      // r += "<yaxis quantity='"+fields[1..]*"/"+"'>\n";
      r += "<legend>"+
	   Array.map(fields[1..],
		     lambda(mixed s) {return (string)s;} )*"\t"+"</legend>\n";
      r += "<xnames>"+
	   Array.map(column(table, 0),
		     lambda(mixed s) {return (string)s;} )*"\t"+"</xnames>\n";
      array a = ({});
      for(int i = 1; i < sizeof(table[0]); i++)
	if(types[i] != "hidden")
	  a += ({ Array.map(column(table, i),
			    lambda(mixed s) { return (string)s; } )*"\t" });
      r += "<data form=row>\n"+a*"\n"+"</data>";
      r += "</diagram>\n";
      break;
    case "table":
      int min = block, max = (sblock=="all"?sizeof(table):TABLE_PAGE_SIZE);
      r += nav + logview_table(table, fields, types, min) + nav;
      break;
    default:
      return r + "<b>Unknown display type.</b><p>";
    }
#endif
  }
  r += "</form>\n";
  return r;
}

string query_location()
{
  return query("mount");
}

string query_location_low(object id)
{
  if(id->misc->sb && id->misc->build_url)
    return id->misc->build_url(id,"LogView/");
  else
    return query_location();
}


array register_module()
{
  string s = "";
  if(loaded) {
    string q = query_location();
    if(sizeof(q) && q[0] == '/') q = q[1..];
    s = "Go to the <a href='"+
        my_configuration()->query("MyWorldLocation")+
        combine_path(q, "admin/")+"'>admininstration page</a>.\n";
  } else
    s = "The Roxen LogView module gathers and displays various\n"
        "statistics for a Roxen WWW server. LogView is also capable\n"
        "of importing Common Log Format files produced by other WWW "
        "servers.\n";
  return ({ MODULE_LOGGER | MODULE_LOCATION | MODULE_PARSER | MODULE_PROVIDER,
	    "LogView", s, 0, 1 });
}

mapping query_tag_callers()
{
  return ([ "logview":tag_logview ]);
}

void create(object configuration)
{
  mydisklocation = combine_path(backtrace()[-1][0], "../");
  master()->add_module_path(mydisklocation);

  defvar("mount", "/logview/", "Mount point", TYPE_LOCATION, "");
  defvar("update-hour", "03", "Update hour", TYPE_MULTIPLE_STRING,
	 "Time when LogView will update the statistics databases.",
	 Array.map(indices(allocate(24)),
		   lambda(int x) { return sprintf("%02d", x); })+
	 ({ "Every hour" }));
  defvar("update-nice-value", 5, "Update nice value",
	 TYPE_INT, "Nice value between 0 and 20 for the LogView update "
	 "process.");
  defvar("be-friendly", 1,
	 "Minimize load", TYPE_FLAG,
	 "Minimize the number of concurrent update processes.");
  defvar("logpath",
	 configuration && combine_path(configuration->query("LogFile"), "../")+
	 "logview.data", "Log path", TYPE_DIR, "");
  defvar("admin", roxen->query("ConfigurationUser"),
	 "Admin user name", TYPE_STRING,
	 "This user name grants full access to the configuration "
	 "part of LogView.");
  defvar("adminpass", roxen->query("ConfigurationPassword"),
	 "Admin password", TYPE_PASSWORD,
	 "This password grants full access to the configuration "
	 "part of LogView.");
  defvar("view", "", "Viewer user name", TYPE_STRING,
	 "This user name grants users access to view statistics.");
  defvar("viewpass", crypt(""), "Viewer password", TYPE_PASSWORD,
	 "This password grants users access to view statistics.");
  defvar("enable_logview_tag", 0, "Enable &lt;logview&gt; tag.", TYPE_FLAG,
	 "Enables the &lt;logview&gt; tag outside LogView.");

  defvar("provider_disable", 0, "Provider disable", TYPE_INT|VAR_EXPERT);
  defvar("provider_backdoor", 0, "Provider backdoor", TYPE_INT|VAR_EXPERT);
}

class LsMekanism {
  inherit RoxenRPC.Server;
  static private object megals;

  mixed do_call(mixed ... args)
  {
    object key;
    catch(key = megals->lock());
    return ::do_call(@args);
  }

  void create(object ls_in, mixed ... args)
  {
    megals = ls_in;
    ::create(@args);
  }
}

string validate_user(object id, array(string) args)
{
  if(id->misc->sb)
    return id->misc->logview_adminp?"admin":"logview";
  string type = "logview";
  if(sizeof(args) && args[0] == "admin")
    type = "admin";
  string user = ((id->realauth||"*:*")/":")[0];
  string key = ((id->realauth||"*:*")/":")[1];
  if (id->misc->content_editor_base) return type;
  catch {
    switch(type) {
    case "logview":
      if(query("view") == "")
	return "logview";
      if(user == query("view"))
	if(stringp(query("viewpass")) && crypt(key, query("viewpass")))
	  return "logview";
      break;
    case "admin":
      if(user == query("admin"))
	if(stringp(query("adminpass")) && crypt(key, query("adminpass"))) {
	  check_login(id);
	  return "admin";
	}
      break;
    }
  };
  return 0;
}

array(array) log_queue = ({});
int queue_timeout = 0, num_log_entries = 0;

void check_queue(int|void go)
{
  int size = sizeof(log_queue);
  // werror(sprintf("(LogView) checking [%O]\n", size));
  if(log_this_server && size && (size > QUEUE_SIZE || go || queue_timeout)) {
    // werror(sprintf("(LogView) PRE FLUSH\n", size));

    MEGALS
    if(!(sizeof(log_queue) && (sizeof(log_queue) >= size || go)))
      return;
    // werror(sprintf("(LogView) FLUSH [%O]\n", size));

    array(array) local_queue;
    array(string) f_name = ({ "host", "referrer", "authuser",
			      "year", "month", "date",
			      "hour", "minute", "second", "timezone",
			      "method", "path", "protocol", "status", "size",
			      "client", "userid" });
    
    mapping t = localtime(time(1));
    array(int) timestamp = ({ t->year+1900, t->mon+1, t->mday,
			      t->hour, t->min, t->sec });
    string tablename;
    string adir = combine_path(query("logpath"), "working");
    string bdir = adir+"/now";

    if(update_every_hour)
      tablename = sprintf("%04d_%02d_%02d:%02d",@timestamp[0..3]);
    else
      tablename = sprintf("%04d_%02d_%02d",@timestamp[0..2]);
    
    string workingbucket = combine_path(bdir, "log");
    if(!this_server_bucket)
      if(!catch(this_server_bucket = LogView.Bucket.bucket(workingbucket)))
	num_log_entries = 0;
    
    if(this_server_bucket &&
       ((this_server_bucket->name() != tablename) ||
	(num_log_entries++ > BUCKET_SIZE)))
      bundle();
    
    if(!this_server_bucket) {
      num_log_entries = 0;
      MKDIR(adir);
      MKDIR(bdir);
      this_server_bucket = LogView.Bucket.bucket(workingbucket, "cw",
						 tablename, f_name);
    }

    size = sizeof(log_queue);
    local_queue = log_queue[0..(size-1)];
    log_queue = log_queue[size..];
    if(this_server_bucket) {
      foreach(local_queue, array row)
	this_server_bucket->add(row);
      
      if(queue_timeout)
	this_server_bucket->flush();
    } else
      eventlog->warning("Bucket lost?");
  }
}

void got_host(string host, array row)
{
  // werror(sprintf("(LogView) logging\n"));
  queue_timeout = 0;

  if(host && stringp(host) && sizeof(host))
    row[0] = host;

  /* FIX: Update arrival time. */
  mapping t = localtime(time(1));
  array(int) timestamp = ({ t->year+1900, t->mon+1, t->mday,
			    t->hour, t->min, t->sec });
  for(int i = 0; i < sizeof(timestamp); i++)
    row[3+i] = timestamp[i];
  
  log_queue += ({ row });
  check_queue();
}

void queue_schedule()
{
  // werror(sprintf("(LogView) schedule [%O]\n", queue_timeout));
  if(queue_timeout)
    check_queue();

  queue_timeout = 1;
  call_out(queue_schedule, QUEUE_TIMEOUT);
}

int log(object id, mapping file)
{
  kludge_start(0, id->conf);

  queue_timeout = 0;
    
  // if(query("provider_disable"))
  //   return 0;

  if(log_this_server) {
    mapping t = localtime(time(1));
    array(int) timestamp = ({ t->year+1900, t->mon+1, t->mday,
			      t->hour, t->min, t->sec });
    
    string host     = (string) id->remoteaddr;
    string referer  = id->referer && sizeof(id->referer) && id->referer[0];
    string authuser = id->realauth && sizeof((id->realauth/":")[0])?
                      (id->realauth/":")[0]:0;
    int timezone    = t->tz;
    string method   = (string) id->method;
    string path     = (string) id->not_query +
		      (id->query ? "?" + (string) id->query : "");
    string protocol = (string) id->prot;
    int    status   = (int) (file->error||200);
    int    size     = file->len;
    
    // Roxen extras!
    string client   = (sizeof(id->client||({}))?id->client*" ":"unknown");
    int    userid   = (int) id->cookies->RoxenUserID;

    // werror("add ");
    roxen->ip_to_host(host, got_host,
		      ({ host, referer, authuser,
			 @timestamp, timezone, method,
			 path, protocol, status, size,
			 client, userid }));
  }
  return 0;
}

int email_service(string report)
{
  mapping mc = gate->read_main_config();
  mapping es = mc->email_service||([]);

  mapping r = es[report];
  string msg = mc->reports[report];
  if(r && msg) {
    msg = parse_html(msg, ([ "logview":tag_logview ]), ([]),
		     ([ "misc":([ "logview_portal":1,
				  "logview_email":1 ]) ]));
    mixed err;
    object o;
    if(err=catch(o = Protocols.SMTP.client())) {
      eventlog->error(err);
    } else if(o->simple_mail) {
      if(err=catch(o->simple_mail(r->to, r->subject, r->from, msg)))
	eventlog->error(err);
      else
	return 1;
    } else
      eventlog->error("This version of Pike is too old. "
		      "Please upgrade to get SMTP functionality.");
  } else
    m_delete(es, report);
  return 0;
}

int email_service_weekday_checked = -1;
void email_service_schedule()
{
  mapping m = localtime(time());
  mapping wdays = ([ "Every day":m->wday,
		     "Mondays":1,
		     "Tuesdays":2,
		     "Wednesdays":3,
		     "Thursdays":4,
		     "Fridays":5,
		     "Saturdays":6,
		     "Sundays":7 ]);

  multiset _sent = (<>);
  if(m->hour == 5 && email_service_weekday_checked != m->wday) {
    mapping mc = gate->read_main_config();
    mapping es = mc->email_service||([]);
    foreach(indices(es), string report)
      if(wdays[es[report]->frequency] == m->wday)
	if(email_service(report))
	  _sent[es[report]->to] = 1;
    email_service_weekday_checked = m->wday;
  }

  array sent = sort(indices(_sent));
  if(sizeof(sent)) {
    for(int i = 0; i < sizeof(sent); i++)
      sent[i] = "<b>"+sent[i]+"</b>";
    eventlog->notice("LogView Email Report Service. Mail sent to "+
		     String.implode_nicely(sent)+".");
  }

  call_out(email_service_schedule, EMAIL_SERVICE_TIMEOUT);
}

int dependant_done;
mapping find_file(string name, object id)
{
  kludge_start(0, id->conf);

  MEGALS
  
  if(query("provider_backdoor"))
    return 0;

  id->misc->logview_portal = 1;

  if(!dependant_done) {
    // LogView module dependencies.
    module_dependencies(id->conf, ({ "graphic_text", "obox", "tablify",
#if UTILIZATION
				     // "lsu",
#endif
				     "htmlparse", "business", "configtablist" }) );
    dependant_done = 1;
  }

  if(id->method != "GET" && id->method != "POST")
    return 0;

  // if(mapping tmp = httprpc->http(name)) return tmp;
  
  string r = "";
  array(string) args = name/"/";
  string base = combine_path(query_location_low(id), "");

  if(args[0] == "focus")
    return http_redirect("/"+args[1..]*"/"+"?"+id->query+"#focus");

  //
  // User validation
  //
  string user = validate_user(id, args);
  if(!user)
    return (["type":"text/html",
		   "error":401,
		   "extra_heads":
		     ([ "WWW-Authenticate":
		        "basic realm=\"LogView\""]),
		   "data":"<title>Access Denied</title>"
		   "<h2 align=center>Access forbidden</h2>"
    ]);
  if(user == "admin") {
    base += "admin/";
    args = args[1..];
  }

  //
  // LogView logo
  //
  if(sizeof(args) && args[-1] == "logview.gif")
    return http_string_answer(logo, "image/gif");

  //
  // World map
  //
  if(sizeof(args) && glob("world*", args[-1]))
    return (get_world(args[-1])?
	    http_string_answer(get_world(args[-1]), "image/gif"):
	    http_string_answer("Please reload.", "text/html"));

  //
  // Menu bar
  //
  array(string) menu;
  menu = ({ "report", "view", "advanced" });
  if(user == "admin")
    menu += ({ "eventlog", "wizard" });
  menu += ({ "manual" });
  mapping(string:string) menunames = ([ "view":"Statistics",
					"wizard":"Configure",
					"advanced":"Advanced",
					"report":"Report",
					"eventlog":"Event log",
					"manual":"Manual" ]);

  if(!args || !sizeof(args) || !sizeof(args[0]))
    args = ({ user=="admin"?"wizard":"report" });

  r += "<table border=0 cellspacing=0 cellpadding=0><tr><td>";
  r += "<config_tablist>\n";
  foreach(menu, string name)
    r += ("<tab"+(menunames[name]==menunames[args[0]]?" selected":"")+
	  " href='"+base+combine_path(name, "")+"'>"+
	  menunames[name]+"</tab>");
  r += "</config_tablist></td></tr></table>\n";

  //
  // Menu handling
  //
  string megatitle = "", megatitle_x = "";
  if(user == "admin") {
    megatitle = ("Administration on "+virtual_name);
    megatitle_x = ("LogView Administration on "
		   "<a href='"+roxen->config_url()+"'>"+virtual_name+"</a>");
    
    switch(args[0]) {
    case "wizard":
      base += "wizard/";
      mixed w;
      if(mappingp(w = wizard_menu(id, mydisklocation+"LogView.pmod/wizards/",
				  base, gate))) return w;
      r += "<ul>"+w+"</ul>\n";
      // r += html_wizard(args[1..], id, base);
      break;
    case "manual":
      if(sizeof(args) > 1 && args[1] == "img")
	return http_string_answer(read_manual(args[1..],id), "image/gif");
      r += read_manual(args[1..],id);
      break;
    case "report":
      r += html_report(args[1..], id, base);
      break;
    case "eventlog":
      base += "eventlog/";
      mixed m = html_eventlog(args[1..], id, base);
      if(mappingp(m)) return m;
      r += m;
      break;
    case "advanced":
      base += "advanced/";
      mixed m = html_advanced(args[1..], id, base);
      if(mappingp(m)) return m;
      r += m;
      break;
    case "view":
      base += "view/";
      mixed m = html_view_function(args[1..], id, base);
      if(mappingp(m)) return m;
      r += m;
      break;
    default:
      return 0;
    }
  } else if(user == "logview") {
    if(!args || !sizeof(args) || !sizeof(args[0]))
      args = ({ "view" });
    
    switch(args[0]) {
    case "manual":
      if(sizeof(args) > 1 && args[1] == "img")
	return http_string_answer(read_manual(args[1..],id), "image/gif");
      r += read_manual(args[1..],id);
      break;
    case "view":
      base += "view/";
      mixed m = html_view_function(args[1..], id, base);
      if(mappingp(m)) return m;
      r += m;
      break;
    case "report":
      r += html_report(args[1..], id, base);
      break;
    case "advanced":
      base += "advanced/";
      mixed m = html_advanced(args[1..], id, base);
      if(mappingp(m)) return m;
      r += m;
      break;
    default:
      return 0;
    }
  }

  // No body etc. for sb plugins.
  if(id->misc->sb)
    return http_string_answer(parse_rxml(r,id));
  
  string s;
  s = "<html><head><title>Roxen LogView "+megatitle+"</title></head>\n"
      "<body bgcolor='#ffffff' text='#000000' "
      "link='#000070' vlink='#000070' alink='#ff000'>\n";

  s += "<table cellpadding=0 cellspacing=0 border=0 width='100%'>"
       "<tr><td valign=bottom align=left>";
  s += "<a \nhref=\"http://www.roxen.com/\">"
       "<img border=0 alt=\"Roxen\" src=logview.gif></a>";
  // "<img border=0 alt=\"Roxen\" src=/internal-roxen-roxen-icon-gray></a>";
  s += "</td><td>&nbsp;</td>\n<td width='100%' height=39>";

  s += ("<table cellpadding=0 cellspacing=0 width='100%' border=0>"
	"<tr width='100%'><td width=100% align=right valigh=center height=28>"
	"<b>"+megatitle_x+"</b>"
	"</td></tr><tr width='100%'>"
	"<td bgcolor='#003366' align=right height=12 width='100%'>"
	"<font color=white size=-2>Roxen LogView "+
	(user=="admin"?LOGVIEW_VERSION:"")+"&nbsp;&nbsp;</font>"
	"</td></tr></table></td>\n");

  s += "</tr></table>\n<br>\n";
  s += r;
  s += "</body></html>";

  return http_string_answer(parse_rxml(s, id));
}



// SiteBuilder plugin stuff

constant sb_plugin_name="LogView";
constant sb_minimal_access = 1;
mapping|string sb_plugin_handle(string sub, string pre, object id, int access)
{
  id->misc->content_editor_base=pre;
  id->misc->logview_adminp = sub[0..4]=="admin" || sub[0..4]=="focus";
  if(id->misc->logview_adminp&&access<2)
    return "No write access.";
  if(!id->misc->logview_adminp && access==2) {
    id->misc->logview_adminp = 1;
    sub="admin/"+sub;
  }

  mixed answer=this_object()->find_file(sub,id);

  if(answer->error || (answer->type && answer->type!="text/html"))
    return answer;
  else
    return answer->data;
}

multiset query_provides()
{
  return (< "sitebuilder_plugin" >);
}


int kludge_start_done = 0;
string old_logpath;

void start(int num, object configuration)
{
  loaded = 1;
  if(!kludge_start_done || old_logpath != query("logpath")) {
    kludge_start_done = 0;
    call_out(kludge_start, 10, num,configuration);
  }

  update_every_hour = (query("update-hour") == "Every hour");
  summary_schedule(0, 1);  // Wise?
}

void stop()
{
  check_queue(1);
  PDB_cache = ([]);

  function f;
  foreach( ({ start_summarizing, queue_schedule,
	      email_service_schedule, check_friendly_lock }), f)
    remove_call_out(f);
  remove_friendly_lock();
}

void kludge_start(int num, object configuration)
{
  if(kludge_start_done)
    return;
  MEGALS
  if(kludge_start_done)
    return;
  kludge_start_done = 1;

  roxen_configuration = configuration;

  eventlog = LogView.Event.event(this_object());
  eventlog->notice("LogView starting up.");

  MKDIR(query("logpath"));
  MKDIR(combine_path(query("logpath"), "config"));
  MKDIR(combine_path(query("logpath"), "logs"));
  MKDIR(combine_path(query("logpath"), "groups"));
  MKDIR(combine_path(query("logpath"), "working"));
  old_logpath = query("logpath");

  logo = Stdio.read_bytes(mydisklocation+"LogView.pmod/logview.gif");

  // Check if fu directory is symlinked (or bad).
  string fudir = combine_path(query("logpath"), "fu");
  rm(fudir);
  string mwd=combine_path(combine_path(getcwd(), backtrace()[-1][0]), "..");
#if 0
#if UTILIZATION
  if(catch(symlink(mwd+"/LogView.pmod/utilization", fudir)))
#else
  if(catch(symlink(mwd+"/LogView.pmod/fu", fudir)))
#endif
    eventlog->error(sprintf("Unable to create symlink! [%O -> %O]\n",
			    mwd+"/LogView.pmod/fu", fudir));
#endif
  
  virtual_name = configuration->query("name");
  if(!sizeof(virtual_name))
     virtual_name = configuration->name;
  virtual_short_name = configuration->short_name(virtual_name);
  log_defaults =
    ([ "log_str_Server_Domain_Name":"."+configuration->query("Domain"),
       "log_str_Server_URL":configuration->query("MyWorldLocation"),
       "log_str_Page_Regexp":"(\\.html$)|(\\.htm$)|(/$)|(\\.cgi$)",
       //"log_str_Non_Page_Regexp": "(^/_internal)|(^/gtext/)|(^/diagram/)|(^/admin/)",
       "log_str_Query_Filter_Regexp":"",
       "log_str_Download_Files":"(\\.tar$)|(\\.gz$)|(\\.lzh$)|(\\.zip$)|(\\.Z$)|(\\.pdf$)" ]);
  group_defaults =
    ([ "group_int_Max_Table_Size":"500",
       "group_str_Ignore_Pages":"(^"+query_location()+")|(^/admin/)" ]);

  fu = LogView.Fu.fu(get_fu_directory());
  // fu = LogView.Fu.fu(combine_path(query("logpath"), "fu/"));
  fu->exception(eventlog->warning);
  gate = baba_gate(this_object());
  rpc = LsMekanism(megals, "127.0.0.1", 0);
  rpc->set_security(lambda() { return !rpc->num_connections(); } );
  rpc->provide(rpckey, gate);

#if !UTILIZATION
  if(!gate->read_safe(gate->config_dir("config"))) {
    gate->save_main_config(([ "this_server_log":virtual_short_name,
			      "version":LOGVIEW_NUMBER ])+
			   LogView.Report.defaults(virtual_name,
						   virtual_short_name));
    gate->save_log_config(virtual_short_name,
			  ([ "status":"Active",
			     "mode":"this",
			     "log_file":virtual_short_name,
			     "log_args":log_defaults
			  ]));
    gate->save_group_config(virtual_short_name,
			    ([ "logs":([virtual_name:({virtual_short_name})]),
			       "group":virtual_short_name,
			       "group_args":group_defaults,
			       "fugroups":gate->list_fugroups()
			    ]));
  }
#endif
  
  switch(gate->version()) {
  case LOGVIEW_NUMBER:
    break;
  default:
    gate->upgrade_version();
    gate->notice("LogView upgraded to "+LOGVIEW_VERSION+".\n");
    break;
#if 0
  default:
    array groups = gate->list_groups();
    if(sizeof(groups)) {
      eventlog->notice("<b>LogView upgrade.</b> "
		       "LogView will rebuild the statistics "
		       "databases "+String.implode_nicely(groups)+". "
		       "<p>Note: This may take a few hours if "
		       "the log is large.");
      foreach(groups, string group)
	gate->purge_group(group);
      if(sizeof(groups) == 1)
	gate->notice("Group "+String.implode_nicely(groups)+" is deleted.");
      else if(sizeof(groups) > 1)
	gate->notice("Groups "+String.implode_nicely(groups)+" are deleted.");
    }
    gate->upgrade_version();
#endif
  }
  
  gate->update_log_this_server();

  // summary_schedule();
  queue_schedule();
  email_service_schedule();
  call_out(check_friendly_lock, 120);
}
