/* This program optimises the log by defragmentating it.
 *
 * Usage: roxen-pike optimise <log>
 */

#if __VERSION__ >= 0.6
import "../";
#endif

#define MAX_ROWS 20000
#define POSTFIX ".tmp"

int main(int argc, array(string) argv)
{
  if(argc != 2) {
    write("Usage: roxen-pike optimise <log>\n"
	  "This program optimises a LogView log by defragmentating it.\n");
    return 1;
  }
  
  string filename = argv[1]+"/log";
  write(sprintf("Optimising %O.\n", filename));
  
  array(string) fields = ({ "host", "referrer", "authuser", "year", "month",
			    "date", "hour", "minute", "second", "timezone",
			    "method", "path", "protocol", "status", "size",
			    "client", "userid" });

  int row_counter;
  string working_entry, current_entry;
  object source = LogView.Bucket.bucket(filename), target;
  foreach(sort(indices(source)), string entry) {
    sscanf(entry, "%s%*[#:]", current_entry);

    if(working_entry != current_entry || row_counter > MAX_ROWS) {
      if(target)
	target->bundle();
      row_counter = 0;
      working_entry = current_entry;
      target = LogView.Bucket.bucket(filename+POSTFIX, "wc", entry, fields);
    }
  
    write(sprintf("  optimising %O...\n", entry));
    foreach(source->query(entry, fields), array row) {
      target->add(row);
      row_counter++;
    }
  }
  if(target)
    target->bundle();

  write(sprintf("Statistics: entries reduced from %d to %d.\n",
		sizeof(indices(source)),
		sizeof(indices(LogView.Bucket.bucket(filename+POSTFIX)))));

  mv(filename+POSTFIX, filename);
  mv(filename+POSTFIX+".i", filename+".i");
  
  return 0;
}
