/* Verify that two logs are identical.
 *
 * Usage: roxen-pike verify <log 1> <log 2>
 */

#if __VERSION__ >= 0.6
import "../";
#endif

class LogReader {
  static private object bucket;
  static private string filename;
  static private array data = ({});
  static private int entry_no = 0, row_no = 0;

  array(string) fields = ({ "host", "referrer", "authuser", "year", "month",
			    "date", "hour", "minute", "second", "timezone",
			    "method", "path", "protocol", "status", "size",
			    "client", "userid" });

  array get_row()
  {
    if(row_no < sizeof(data))
      return data[row_no++];
    
    if(entry_no < sizeof(indices(bucket))) {
      row_no = 0;
      string entry = indices(bucket)[entry_no++];
      data = bucket->query(entry, fields);
      write(sprintf("#bucket: %O [%O] (%d)\n", entry, filename, sizeof(data)));
      return get_row();
    }

    return 0;
  }
  
  void create(string _filename)
  {
    filename = _filename;
    bucket = LogView.Bucket.bucket(filename);
    indices(bucket);
  }
}

int main(int argc, array(string) argv)
{
  if(argc != 3) {
    write("Usage: roxen-pike verify <log 1> <log 2>\n"
	  "Verify that two logs are identical.\n");
    return 1;
  }

  string filename1 = argv[1]+"/log";
  string filename2 = argv[2]+"/log";
  write(sprintf("Verifying %O with %O.\n", filename1, filename2));
  
  object log1 = LogReader(filename1);
  object log2 = LogReader(filename2);

  array row1, row2;
  do {
    row1 = log1->get_row();
    row2 = log2->get_row();
    if(!equal(row1, row2)) {
      write("ERROR: Mismatch.\n");
      return 1;
    }
  } while(row1 || row2);
  
  write("OK: Done.\n");
  return 0;
}
